// -*- C++ -*- std header                                                      |
//-----------------------------------------------------------------------------+
//                                                                             |
// auxx.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// some global macros                                                          |
// some global constants                                                       |
// some global typedefs                                                        |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_auxx_h
#define included_auxx_h

#ifndef included_iostream
#include <iostream>
#define included_iostream
#endif

////////////////////////////////////////////////////////////////////////////////
//                                                                              
// 1. INTERNAL PRECISION                                                        
//                                                                              
////////////////////////////////////////////////////////////////////////////////
//                                                                              
// 1.1 from math.h:                                                             
//                                                                              

extern "C" {
  double sqrt (double);
  double exp  (double);
  double log  (double);
  float  sqrtf(float);
  float  expf (float);
  float  logf (float);
  double pow  (double,double);
//   float  powf (float,float);
}

#ifdef DOUBLEPRECISION
//                                                                              
// 1.2 double precision: 64bit floating point numbers to hold a real            
//                                                                              
namespace nbdy {
  typedef double real;
  inline real SQRT(const real x) { return sqrt(x); }
  inline real EXP (const real x) { return exp (x); }
  inline real LOG (const real x) { return log (x); }
  inline real POW (const real x,
		   const real y) { return pow(x,y); }
}

#else // DOUBLEPRECISION
//                                                                              
// 1.3 single precision: 32bit floating point numbers to hold a real            
//                                                                              
namespace nbdy {
  typedef float real;
#ifdef linux
  inline real SQRT(const real x) { return sqrtf(x); }
  inline real EXP (const real x) { return expf (x); }
  inline real LOG (const real x) { return logf (x); }
#else
  inline real SQRT(const real x) { return sqrt(double(x)); }
  inline real EXP (const real x) { return exp (double(x)); }
  inline real LOG (const real x) { return log (double(x)); }
#endif
  inline real POW (const real x,
		   const real y) { return pow(double(x),double(y)); }
}
#endif // DOUBLEPRECISION

////////////////////////////////////////////////////////////////////////////////
//                                                                              
// 2. DATA TRANSFER VIA BODIES OR ARRAYS & PRECISION USED IN ARRAY ARGS         
//                                                                              
////////////////////////////////////////////////////////////////////////////////

namespace nbdy {
#ifdef DOUBLEINPUT
  typedef double areal;
#else  // DOUBLEINPUT
  typedef float areal;
#endif // DOUBLEINPUT
}

////////////////////////////////////////////////////////////////////////////////
//                                                                              
// 3. USEFUL METHODS                                                            
//                                                                              
////////////////////////////////////////////////////////////////////////////////

#ifndef included_vect_h
#include <util/inln.h>
#endif

namespace nbdy {
  using util::min;
  using util::max;
  using util::abs;
  using util::mod;
  using util::twice;
  using util::square;
  using util::cube;
  using util::pow;
  using util::update_max;
  using util::update_min;
}

////////////////////////////////////////////////////////////////////////////////
//                                                                              
// 4. DIMENSIONALITY                                                            
//                                                                              
////////////////////////////////////////////////////////////////////////////////
//                                                                              
// 4.1 Two dimensions                                                           
//                                                                              
#ifdef TWODIMENSIONAL

#define NDIM 2   // # dimensions
#define NSUB 4   // 2^NDIM

#else // TWODIMENSIONAL
//                                                                              
// 4.2 Three dimensions                                                         
//                                                                              

#define NDIM 3   // # dimensions
#define NSUB 8   // 2^NDIM

#endif // TWODIMENSIONAL

#ifndef included_vect_h
#include <nbdy/vect.h>
#endif

#ifndef included_tens_h
#include <nbdy/tens.h>
#endif

namespace nbdy {
  typedef class vector <real>  vect;
  typedef class evector<real>  evec;
  typedef class sym2<real>     matr;
  typedef class sym3<real>     tens;
  typedef class esym2<real>    emat;
  typedef class esym3<real>    eten;
}

#ifdef TWODIMENSIONAL
namespace nbdy {
  typedef vect                   real;
}
#else // TWODIMENSIONAL
namespace nbdy {
  typedef vect                   amom;
}
#endif // TWODIMENSIONAL

////////////////////////////////////////////////////////////////////////////////
//                                                                              
// 5. USEFUL CONSTANTS AND TYPEDEFS                                             
//                                                                              
////////////////////////////////////////////////////////////////////////////////

namespace nbdy {
  typedef unsigned short   indx;         // use only in non-register memory
  typedef unsigned int     uint;
  typedef int              flag_type;
  //----------------------------------------------------------------------------
  inline void error  (const char*m) {
    std::cerr<<" TREE ERROR: "  <<m<<std::endl;
    exit(1);
  }
  //----------------------------------------------------------------------------
  inline void warning(const char*m) {
    std::cerr<<" TREE WARNING: "<<m<<std::endl;
  }
  //----------------------------------------------------------------------------
  const   real zero    = 0., 
               third   = 0.3333333333333333333,
               half    = 0.5,
               one     = 1.,
               two     = 2.,
               ten     = 10.;
}

////////////////////////////////////////////////////////////////////////////////
#endif  // included_auxx_h
