// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// cell.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class cell                         fully inline                             |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_cell_h
#define included_cell_h 1

#ifndef included_auxx_h
#include <nbdy/auxx.h>
#endif

#ifndef included_bcell_h
#include <nbdy/bcell.h>
#endif

#ifndef included_soul_h
#include <nbdy/soul.h>
#endif

////////////////////////////////////////////////////////////////////////////////
//                                                                              
// macro NGRAVP: number of reals needed in grav-part of cell = 6 + size of matr 
//                                                                              
#if NDIM==3
#define NGRAVP 12
#else
#define NGRAVP  9
#endif
////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
  class TaylorCoeffs;                              // forward declaration       
  //////////////////////////////////////////////////////////////////////////////
  // nbdy::class cell                                                           
  //////////////////////////////////////////////////////////////////////////////
  class cell : 
    // class cell is a element in the linked structure of a tree                
    public flagged_cell<cell,soul>                 // our basic properties      
  {
    cell           (const cell&);                  // not implemented           
    cell& operator=(const cell&);                  // not implemented           
    //--------------------------------------------------------------------------
    // data of class cell                                                       
    //--------------------------------------------------------------------------
  private:
    vect             COFM;                         // center of mass            
    real             RMAX;                         // max distance from CofM    
    real             MASS;                         // mass in cell              
    mutable real     GRAVP[NGRAVP];                // holds grav-part of cell   
    TaylorCoeffs    *COEFFS;                       // pointer to coefficients   
    //--------------------------------------------------------------------------
    // constructor                                                              
    //--------------------------------------------------------------------------
  public:
    cell() {}
    //--------------------------------------------------------------------------
    // simple manipulations                                                     
    //--------------------------------------------------------------------------
  public:
    void set_rcrit(const real ith)    { rcrit()  = ith * rmax();
                                        rcrit2() = rcrit() * rcrit(); }
    //--------------------------------------------------------------------------
    // non-const data access via member functions                               
    //--------------------------------------------------------------------------
    real&         rmax  () { return RMAX; }
    real&         mass  () { return MASS; }
    vect&         cofm  () { return COFM; }
    real&         imass () { return GRAVP[0]; }
    real&         rcrit () { return GRAVP[1]; }
    real&         rcrit2() { return GRAVP[2]; }
    real&         eps   () { return GRAVP[3]; }
    real&         emin  () { return GRAVP[4]; }
    real&         emax  () { return GRAVP[5]; }
    emat          quadru() { return emat(GRAVP+6); }
    TaylorCoeffs*&coeffs() { return COEFFS; }
    //--------------------------------------------------------------------------
    // const data access via friends                                            
    //--------------------------------------------------------------------------
#define CCCC const cell*const&C
    friend real          const &rmax  (CCCC) { return C->RMAX; }
    friend real          const &mass  (CCCC) { return C->MASS; }
    friend vect          const &cofm  (CCCC) { return C->COFM; }
    friend real          const &imass (CCCC) { return C->GRAVP[0]; }
    friend real          const &rcrit (CCCC) { return C->GRAVP[1]; }
    friend real          const &rcrit2(CCCC) { return C->GRAVP[2]; }
    friend real          const &eps   (CCCC) { return C->GRAVP[3]; }
    friend real          const &emin  (CCCC) { return C->GRAVP[4]; }
    friend real          const &emax  (CCCC) { return C->GRAVP[5]; }
    friend emat          const  quadru(CCCC) { return emat(C->GRAVP+6); }
    friend TaylorCoeffs* const &coeffs(CCCC) { return C->COEFFS; }
    //--------------------------------------------------------------------------
    // boolean information via friends                                          
    //--------------------------------------------------------------------------
    friend bool is_source       (CCCC) {return C->MASS!=zero; }
#undef CCCC
    //--------------------------------------------------------------------------
    // other const and static methods                                           
    //--------------------------------------------------------------------------
    static void dump_head(std::ostream&o) {
      o<<"     #";
      flagged_cell<cell,soul>::dump_head(o);
      o<<"              mass              cofm                  rmax";
    }
    //--------------------------------------------------------------------------
    void dump(std::ostream& o,
	      const soul* const &S0,
	      const cell* const &C0) const
    {
      o<<" "<<std::setw(5)<<int(this-C0);
      flagged_cell<cell,soul>::dump(o,S0,C0);
      o<<" "<<setw(8)<<MASS;
      for(register int i=0; i<NDIM; i++)
	o<<" "<<setw(8)<<setprecision(4)<<COFM[i];
      o<<" "<<setw(12)<<RMAX;
    }
  };
}
#undef NGRAVP
////////////////////////////////////////////////////////////////////////////////
#endif // included_cell_h
