// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// grav.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class gravity_kernel                 see file grav.cc                       |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_grav_h
#define included_grav_h

#ifndef included_kern_h
#include <nbdy/kern.h>
#endif

#ifndef included_cell_h
#include <nbdy/cell.h>
#endif

#ifndef included_tayl_h
#include <nbdy/tayl.h>
#endif

////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
  class soul;
  class TreeStats;
  class gravity_kernel {
    // class for the greens function used in gravity. Allows for 12 different   
    // functional forms, see tree.h                                             
  public:
    gravity_kernel           (      TreeStats*,       // statistics             
			      const kern_type,        // type of kernel         
			      const real = zero,      // softening length       
			      const soft_type = global);
    const real& current_eps  () const;                // return eps             
    const real& current_epsq () const;                // return eps^2           
    void        set_eps      (const real) const;      // set eps^i              
    void        set_softening(const soft_type);       // set softening type     
    bool        interact     (cell*) const;           // cell-self interaction  
    bool        interact     (cell*, cell*) const;    // cell-cell interaction  
    bool        interact     (cell*, soul*) const;    // cell-body interaction  
    void        interact     (soul*, soul*) const;    // body-body interaction  
    bool        split_first  (const cell*, const cell*) const;
    bool        continuous_kernel() const;
    void direct_summation    (cell*) const;
    real greens_function     (vect&) const;
  private:
    enum who {
      none = 0,
      some = 1,
      all  = 2
    };
    real& c0      (cell* C) const { return C->coeffs()->c0(); }
    evec  c1      (cell* C) const { return C->coeffs()->c1(); }
    emat  c2      (cell* C) const { return C->coeffs()->c2(); }
    eten  c3      (cell* C) const { return C->coeffs()->c3(); }
    soft_type         SOFT;
    kern_type         KERN;
    bool              CONT;
    TreeStats        *STAT;
    mutable real      EPS, EPQ, IEP, IEQ;

    who who_is_sink       (const cell*) const;
    void many_global      (const bool,soul*,const who,soul*,const soul*) const;
    void many_individual  (const bool,soul*,const who,soul*,const soul*) const;
    void many_global      (cell*, soul*) const;
    void many_individual  (cell*, soul*) const;
    void many_global      (cell*, cell*) const;
    void many_individual  (cell*, cell*) const;
    void many_global      (cell*) const;
    void many_individual  (cell*) const;
    void grav             (vect&, const real, cell*, soul*) const;
    void grav             (vect&, const real, cell*, cell*) const;
    bool mutual_global    (cell*, soul*) const;
    bool mutual_individual(cell*, soul*) const;
    bool mutual_global    (cell*, cell*) const;
    bool mutual_individual(cell*, cell*) const;
    void set_inverse_eps  () const     // pre-sets inverse eps
    { IEP = (EPS>0)? one/EPS : zero;   // 1/eps
      IEQ = IEP*IEP; }                 // 1/eps^2
  };
  //----------------------------------------------------------------------------
  inline gravity_kernel::gravity_kernel(      TreeStats* t,
					const kern_type k,
					const real e,
					const soft_type s)
    : SOFT(s), KERN(k),
      CONT(KERN==newton || KERN==p0 || KERN==p1 || KERN==p2 || KERN==p3),
      STAT(t), EPS(e)
  {
    EPQ = EPS*EPS;
    IEP = (EPS>0)? one/EPS : zero;   // 1/eps
    IEQ = IEP*IEP;                   // 1/eps^2
  }
  //----------------------------------------------------------------------------
  inline bool gravity_kernel::split_first(const cell* A, const cell* B) const
  {
    return is_twig(B) || (!is_twig(A) && rmax(A) > rmax(B));
  }
  //----------------------------------------------------------------------------
  inline const real& gravity_kernel::current_eps  () const { return EPS; }
  inline const real& gravity_kernel::current_epsq () const { return EPQ; }
  inline void        gravity_kernel::set_softening(const soft_type ST) {SOFT=ST;}
  inline bool        gravity_kernel::continuous_kernel() const { return CONT; }
  //============================================================================
  inline const real& eps (const gravity_kernel* K) { return K->current_eps(); }
  inline const real& epsq(const gravity_kernel* K) { return K->current_epsq(); }
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_grav_h
