// -*- C++ -*- std header                                                      |
//-----------------------------------------------------------------------------+
//                                                                             |
// pext.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_pext_h
#define included_pext_h

#ifndef included_auxx_h
#include <nbdy/auxx.h>
#endif

////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::potential                                                      
  //////////////////////////////////////////////////////////////////////////////
  //                                                                            
  // this is an abstract base class for a potential that can be used as external
  // potential in a N-body code, or as stand-alone potential in orbit           
  // integration                                                                
  //                                                                            
  //////////////////////////////////////////////////////////////////////////////
  class potential {
  protected:
    typedef real tensor[NDIM][NDIM];
  public:
    virtual vect force      (                      // R: acceleration           
			     const vect&,          // I: position               
			     const real) const=0;  // I: time                   
    virtual real pot_f      (                      // R: potential              
			     vect&,                // O: acceleration           
			     const vect&,          // I: position               
			     const real)const=0;   // I: time                   
    virtual real rho        (                      // R: density                
			     const vect&,          // I: position               
			     const real) const=0;  // I: time                   
    virtual void energies   (                      // contribution to totals in:
			     real  &u,             // scalar pot energy         
			     tensor&w,             //   pot energy tensor       
			     tensor&t,             // 2*kin energy tensor       
			     vect  &x,             // dipole                    
			     vect  &v,             // momentum                  
			     amom  &l,             // angular momentum          
			     real  &m) const  {}   // mass                      
  };
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::nemo_pot                                                       
  //////////////////////////////////////////////////////////////////////////////
#ifdef ALLOW_NEMO
  class nemo_pot :
    public potential
  {
  private:
    const int ndim;
    typedef void (*potproc_real)(const int*, const real*, real*, real*,
				 const real*);
    const potproc_real MYPOT;
  public:
    nemo_pot(const potproc_real& P) : ndim(NDIM), MYPOT(P) {}
    vect force(const vect& X, const real T) const {
      register vect F;
      register real P;
      MYPOT(&ndim, X.const_pointer(), F.pointer(), &P, &T);
      return F;
    }
    real pot_f(vect&F, const vect&X, const real T) const {
      register real P;
      MYPOT(&ndim, X.const_pointer(), F.pointer(), &P, &T);
      return P;
    }
    real rho  (const vect& X, const real T) const { return zero; }
  };
#endif // ALLOW_NEMO
}
////////////////////////////////////////////////////////////////////////////////

#endif // included_pext_h
