// -*- C++ -*- std header                                                      |
//-----------------------------------------------------------------------------+
//                                                                             |
// step.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class TimeStep            fully inline                                      |
// class BlockStep           fully inline                                      |
// class DaccScheme          fully inline                                      |
// class RhoScheme           fully inline                                      |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_step_h
#define included_step_h

#ifndef included_iostream
#include <iostream>
#define included_iostream
#endif

#ifndef included_iomanip
#include <iomanip>
#define included_iomanip
#endif

#ifndef included_body_h
#include <body.h>
#endif

////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::LeapFrog                                                       
  //////////////////////////////////////////////////////////////////////////////
  class LeapFrog {
    LeapFrog(const LeapFrog&);                     // not implemented           
    LeapFrog operator= (const LeapFrog&);          // not implemented           
    //--------------------------------------------------------------------------
    // data members                                                             
    //--------------------------------------------------------------------------
  private:
    real         TAU, TAUH;                        // time step                 
    mutable real TIME;                             // simulation time           
    //--------------------------------------------------------------------------
    // public methods                                                           
    //--------------------------------------------------------------------------
  public:
    LeapFrog(const int f, const real t)
      : TAU (pow(half,f)), TAUH (half*TAU), TIME(t) {}
    const real& time() const { return TIME; }
    const real& tau () const { return TAU; }
    void predict    (const bodies*) const;         // before force calculation  
    void accelerate (const bodies*) const;         // after force calculation   
  };
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::BlockStep                                                      
  //////////////////////////////////////////////////////////////////////////////
  class BlockStep {
    BlockStep(const BlockStep&);                   // not implemented           
    BlockStep& operator= (const BlockStep&);       // not implemented           
    //--------------------------------------------------------------------------
    // friendships                                                              
    //--------------------------------------------------------------------------
    friend class DaccScheme;
    friend class RhoScheme;
    //--------------------------------------------------------------------------
    // private types                                                            
    //--------------------------------------------------------------------------
  private:
    struct TimeStep {
      mutable uint  N,                             // # bodies in this step     
	            NC;                            // # bodies in this & shorter
      real          TAU, TAUSQ;                    // time step                 
      mutable real  T_OPOS,                        // time of bodies opos       
	            DT, DTH;                       // DT = actual time - T_OPOS 
      void re_set(const int f=0, const real t=zero){ TAU =pow(half,f);
      TAUSQ = TAU*TAU;
      T_OPOS= t; DT=DTH=zero; }
      TimeStep   (const int f=0, const real t=zero){ N=0u; NC=0u; re_set(f,t); }
      void        reset_count ()             const { N=0u; NC=0u; }
      void        set_predict (const real t) const { DT = t-T_OPOS; DTH=half*DT; }
      void        dump   (std::ostream&to)   const { to<<N<<" "<<NC<<" "<<TAU<<" "
						       <<T_OPOS<<" "<<DT<<"\n"; }
      void        add               ()       const { N++; NC++;   }
      void        add_to_shorter    ()       const {      NC++;   }
      void        join_from_shorter ()       const { N++;         }
      void        join_from_longer  ()       const { N++; NC++;   }
      void        leave_to_shorter  ()       const { N--;         }
      void        leave_to_longer   ()       const { N--; NC--;   }
      void        move_time         ()       const { T_OPOS += TAU; }
      const real& time_opos         ()       const { return T_OPOS; }
      const real& tau               ()       const { return TAU;  }
      const real& tausq             ()       const { return TAUSQ;}
      const real& dt                ()       const { return DT;   }
      const real& dt_half           ()       const { return DTH;  }
      const uint& number            ()       const { return N;    }
      const uint& acc_number        ()       const { return NC;   }
    };
    //--------------------------------------------------------------------------
    // public types                                                             
    //--------------------------------------------------------------------------
  public:
    enum too_short {
      go_longer    = 1,                            // step is/was too short     
      forbid_short = 2,                            // step ok, next short enough
      allow_short  = 4                             // step ok, next too long ?  
    };
    //--------------------------------------------------------------------------
    // data members                                                             
    //--------------------------------------------------------------------------
  private:
    int            H0;                             // sets longest time step    
    uint           NSTEPS;                         // # steps                   
    TimeStep      *TS, *SHORTEST;                  // arrays with time steps    
    mutable real   TIME;                           // simulation time           
    real           F, FQ, FQTOL, FQSH, FQSHTOL;    // aux constants             
    //--------------------------------------------------------------------------
    // private methods                                                          
    //--------------------------------------------------------------------------
    void join_from_shorter(const indx l) const { (TS+l)->join_from_shorter(); }
    void join_from_longer (const indx l) const { (TS+l)->join_from_longer();  }
    void leave_to_shorter (const indx l) const { (TS+l)->leave_to_shorter();  }
    void leave_to_longer  (const indx l) const { (TS+l)->leave_to_longer();   }
    //--------------------------------------------------------------------------
    // protected methods                                                        
    //--------------------------------------------------------------------------
  protected:
    void add              (const indx)   const;
    void reset_counts     () const;
    void reset_steps      (const int,  const indx, const real=zero);
    void reset_fac        (const real, const real=two);
    //--------------------------------------------------------------------------
    // construction & destruction                                               
    //--------------------------------------------------------------------------
  public:
    BlockStep                   () : TS(0) {}
    BlockStep                   (const int,        // O: h0, tau_max = 2^-h0    
			         const uint,       // I: #steps                 
			         const real,       // I: actual time            
			         const real,       // I: f                      
			         const real=two);  // I: tol^2                  
    ~BlockStep                  ()              { delete[] TS;}
    //--------------------------------------------------------------------------
    // other public methods                                                     
    //--------------------------------------------------------------------------
    void        clock_on        ()             const;
    void        clock_back      ()             const;
    void        set_predict     ()             const;
    void        dump            (std::ostream&)const;
    void        predict_pos     (body&)        const;
    void        move_times      (const indx)   const;
    uint        longest_moving  (const uint)   const;
    bool        is_leap_frog    ()             const {return NSTEPS == 1; }
    const real& time            ()             const {return TIME; }
    const real& tau             (const indx l) const {return(TS+l)->tau(); }
    const real& tausq           (const indx l) const {return(TS+l)->tausq(); }
    const real& dt              (const indx l) const {return(TS+l)->dt(); }
    const real& dt_half         (const indx l) const {return(TS+l)->dt_half(); }
    const real& time_opos       (const indx l) const {return(TS+l)->time_opos();}
    const uint& number          (const indx l) const {return(TS+l)->number(); }
    const real& tau             (const body B) const {return tau(level(B)); }
    const real& tausq           (const body B) const {return tausq(level(B)); }
    const real& dt              (const body B) const {return dt(level(B)); }
    const real& dt_half         (const body B) const {return dt_half(level(B)); }
    const real& time_opos       (const body B) const {return time_opos(level(B));}
    const real& factor          ()             const {return F; }
    void        move_proper     (body&)        const;
    void        goto_shorter    (body&)        const;
    void        goto_longer     (body&)        const;
    int         shortest_step   ()             const {return NSTEPS-1; }
    const real& shortest_tau    ()             const {return SHORTEST->tau(); }
    const uint& No_in_level     (const indx l) const {return(TS+l)->number(); }
    const uint& No_shorter_level(const indx l) const {return(TS+l)->acc_number();}
    const  int& h0              ()             const {return H0; }
    void        short_stats     (std::ostream&)const;
  };
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::DaccScheme                                                     
  //////////////////////////////////////////////////////////////////////////////
  class DaccScheme {
  private:
    typedef BlockStep::too_short tooshort;
  public:
    static void     assign_to_level(const BlockStep*, body&);
    static bool     step_too_long  (const BlockStep*, const body&);
    static tooshort step_too_short (const BlockStep*, const body&);
  };
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::RhoScheme                                                      
  //////////////////////////////////////////////////////////////////////////////
  class RhoScheme {
  private:
    static bool     too_long       (const BlockStep*, const real&, const indx);
  public:
    static void     assign_to_level(const BlockStep*, body&);
    static bool     step_too_long  (const BlockStep*, const body&);
    static bool     step_too_short (const BlockStep*, const body&);
  };

  //MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM//
  //WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW//
  //                                                                          //
  //                      INLINE FUNCTION DEFINITIONS                         //
  //                                                                          //
  //MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM//
  //WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW//

  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::LeapFrog                                                       
  //////////////////////////////////////////////////////////////////////////////
  inline void LeapFrog::predict(const bodies *BODIES) const {
    LoopBodies(bodies,BODIES,B) {
      B.vel() += TAUH * acc(B);                    // v_1/2 = v_0 +a_0 * dt/2   
      B.pos() += TAU  * vel(B);                    // x_1   = x_0 +v_1/2 * dt   
    }
    TIME += TAU;                                   // t_1   = t_0 + dt          
  }
  //----------------------------------------------------------------------------
  inline void LeapFrog::accelerate(const bodies *BODIES) const {
    LoopBodies(bodies,BODIES,B)
      B.vel() += TAUH * acc(B);                    // v_1   = v_1/2 * a_1 * dt/2
  }
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::BlockStep                                                      
  //////////////////////////////////////////////////////////////////////////////
  inline void BlockStep::reset_steps(const int hz, const indx Ns, const real t)
  {
    if(TS) delete[] TS;
    NSTEPS   = Ns;
    TS       = new TimeStep[NSTEPS];
    SHORTEST = TS+NSTEPS-1;
    TIME     = t;
    H0       = hz;
    register int  h;
    register uint i;
    for(i=0, h=H0; i<NSTEPS; i++,h++) (TS+i)->re_set(h,t);
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::reset_counts() const
  {
    for(register TimeStep* s=TS; s<=SHORTEST; s++) s->reset_count();
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::reset_fac(const real f, const real tq)
  {
    F=f; FQ=F*F; FQTOL=FQ*tq; FQSH=0.0625*FQ; FQSHTOL=FQSH*tq;
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::predict_pos(body& B) const
  {
    register vect newpos(oacc(B));
    newpos   *= dt_half(B);
    newpos   += vel    (B);
    newpos   *= dt     (B);
    newpos   += opos   (B);
    B->pos()  = newpos;
  }
  //----------------------------------------------------------------------------
  inline BlockStep::BlockStep(const int h0, const uint Ns, const real t,
			      const real f1, const real tq)
    : TS(0)
  {
    reset_steps(h0,Ns,t);
    reset_fac  (f1,tq);
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::move_times(const indx l) const
  {
    for(register uint i=l; i<NSTEPS; i++) (TS+i)->move_time();
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::dump(std::ostream&to)  const
  {
    for(register uint i=0; i<NSTEPS; i++) {
      to<<i<<": ";
      (TS+i)->dump(to);
    }
  }
  //----------------------------------------------------------------------------
  inline uint BlockStep::longest_moving(const uint T) const
  {
    register uint  i=NSTEPS-1;
    register uint  t=T;
    for(; !(t&1) && i>0; t>>=1, i--);
    return i;
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::clock_on() const
  {
    TIME += shortest_tau();
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::clock_back() const
  {
    TIME -= shortest_tau();
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::set_predict() const
  {
    for(register TimeStep* s=TS; s<=SHORTEST; s++) s->set_predict(TIME);
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::goto_shorter(body&B) const
  {
    leave_to_shorter((B.level())++);
    join_from_longer(level(B));
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::goto_longer(body&B) const
  {
    leave_to_longer((B.level())--);
    join_from_shorter(level(B));
    B.forbid_longer();
    B.allow_shorter();
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::move_proper(body&B) const
    // note that dt(B) = t_acc - t_oacc = t_pos - t_opos
  {
    register vect dvel(oacc(B));
    dvel    += acc    (B);
    dvel    *= dt_half(B);
    B.vel() += dvel;
    B.opos() = pos(B);
    B.oacc() = acc(B);
    B.allow_longer();
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::add(const indx l) const
  {
    (TS+l)->add();
    for(register TimeStep* s=TS; s<TS+l; s++) s->add_to_shorter();
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::short_stats(std::ostream& to) const
  {
    for(register uint l=0; l<NSTEPS; l++) to<<std::setw(7)<<number(l)<<" ";
  }
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::DaccScheme                                                     
  //////////////////////////////////////////////////////////////////////////////
  inline bool DaccScheme::step_too_long (const BlockStep* s, const body& B)
  {
    return dist_sq(oacc(B),acc(B)) > s-> FQSH * sum_sq(oacc(B),acc(B));
  }
  //----------------------------------------------------------------------------
  inline BlockStep::too_short DaccScheme::step_too_short(const BlockStep* s,
							 const body& B)
  {
    register real dacc=dist_sq(oacc(B),acc(B)), macc=sum_sq(oacc(B),acc(B));
    if(dacc < s->FQSH    * macc) return BlockStep::go_longer;
    if(dacc < s->FQSHTOL * macc) return BlockStep::forbid_short;
    return BlockStep::allow_short;
  }
  //----------------------------------------------------------------------------
  inline void DaccScheme::assign_to_level(const BlockStep* s, body& B)
  {
    register uint lev;
    register real FQSH_x_accm = sum_sq (oacc(B),acc(B)) * s->FQSH,
      dacc = dist_sq(oacc(B),acc(B));
    for(lev = s->shortest_step();
	lev > 0u && dacc < FQSH_x_accm;
	lev--, dacc *= 4.);
    s->add(B.level()=lev);
  }

  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::RhoScheme                                                      
  //////////////////////////////////////////////////////////////////////////////
  inline bool RhoScheme::too_long(const BlockStep*s, const real&rh, const indx l)
  {
    return rh * s->tausq(l) > s->F;
  }
  //----------------------------------------------------------------------------
  inline bool RhoScheme::step_too_long(const BlockStep*s, const body& B)
  {
    return too_long(s,rho(B),level(B));
  }
  //----------------------------------------------------------------------------
  inline bool RhoScheme::step_too_short(const BlockStep*s, const body& B)
  {
    return level(B)>0 && !too_long(s,rho(B),level(B)-1);
  }
  //----------------------------------------------------------------------------
  inline void RhoScheme::assign_to_level(const BlockStep*s, body& B)
  {
    register int lev;
    for(lev = 0;
	lev < s->shortest_step() && too_long(s,rho(B),lev);
	++lev);
    s->add(B.level()=lev);
  }
}
////////////////////////////////////////////////////////////////////////////////
#endif   // included_step_h
