// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// stic.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class sticky_soul                 fully inline                              |
// class sticky_cell                 fully inline                              |
// class basic_finder                see stic.cc                               |
// class sticky_finder               see stic.cc                               |
// class neighbour_finder            see stic.cc                               |
// class neighbour_counter           see stic.cc                               |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_stic_h
#define included_stic_h

#ifndef included_cell_h
#include <nbdy/cell.h>
#endif

#ifndef included_soul_h
#include <nbdy/soul.h>
#endif

#ifndef included_tree_h
#include <nbdy/tree.h>
#endif

////////////////////////////////////////////////////////////////////////////////

namespace nbdy {
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::sticky_soul                                                    
  //////////////////////////////////////////////////////////////////////////////
  class sticky_soul :
    public flagged_soul<bodies>
  {
    sticky_soul           (const sticky_soul&);    // not implemented           
    sticky_soul& operator=(const sticky_soul&);    // not implemented           
    //--------------------------------------------------------------------------
    // data of class sticky_soul                                                
    //--------------------------------------------------------------------------
  private:
    real   SIZE;                                   // physical size             
    real   SIZEQ;                                  // (physical size)^ 2        
    vect   VEL;                                    // velocity                  
    uint   NUM;                                    // # neighbours              
    //--------------------------------------------------------------------------
    // constructor                                                              
    //--------------------------------------------------------------------------
  public:
    sticky_soul() {}
    //--------------------------------------------------------------------------
    // simple manipulations for use with bodies                                 
    //--------------------------------------------------------------------------
    void set_sticky(const bodies*B) {
      SIZE=B->size(mybody());
      VEL =B->vel (mybody());
      NUM =0;
    }
    //--------------------------------------------------------------------------
    void set_sph   (const bodies*B) {
      SIZE =B->size(mybody());
      SIZEQ=SIZE*SIZE;
      NUM  =0;
    }
    //--------------------------------------------------------------------------
    // simple manipulations for use with arrays                                 
    //--------------------------------------------------------------------------
    void set_sticky(const areal*s, const areal*v[NDIM]) {
      SIZE=s[mybody()];
      for(register int d=0; d<NDIM; ++d) VEL[d]=v[d][mybody()];
      NUM=0;
    }
    //--------------------------------------------------------------------------
    void set_sph   (const areal*s) {
      SIZE =s[mybody()];
      SIZEQ=SIZE*SIZE;
      NUM  =0;
    }
    //--------------------------------------------------------------------------
    // non-const data access via members                                        
    //--------------------------------------------------------------------------
    void inc() { NUM++; }
    //--------------------------------------------------------------------------
    // const access via friends                                                 
    //--------------------------------------------------------------------------
#define CSCS const sticky_soul* const &S
    friend uint const &num    (CSCS) { return S->NUM; }
    friend vect const &pos    (CSCS) { return S->cofm(); }
    friend vect const &vel    (CSCS) { return S->VEL; }
    friend real const &size   (CSCS) { return S->SIZE; }
    friend real const &sizeq  (CSCS) { return S->SIZEQ; }
#undef CSCS
    //--------------------------------------------------------------------------
    // other const and static methods                                           
    //--------------------------------------------------------------------------
    void dump_head(std::ostream& o) {
      o<<"     #";
      flagged_soul<body>::dump_head(o);
      o<<"           size [velocity]";
    }
    //--------------------------------------------------------------------------
    void dump(std::ostream&o,
	      const sticky_soul*const &S0) const
    {
      o<<" "<<std::setw(5)<<int(this-S0);
      flagged_soul<bodies>::dump(o);
      o<<" "<<std::setw(5)<<std::setprecision(4)<<SIZE;
      if(is_set(basic_flag::STICKY))
	for(register indx i=0;i<NDIM;i++)
	  o<<" "<<std::setw(7)<<std::setprecision(4)<<VEL[i];
    }
  };
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::sticky_cell                                                    
  //////////////////////////////////////////////////////////////////////////////
  class sticky_cell :
    public flagged_cell<sticky_cell,sticky_soul>
  {
    sticky_cell           (const sticky_cell&);    // not implemented           
    sticky_cell& operator=(const sticky_cell&);    // not implemented           
    //--------------------------------------------------------------------------
    // data of class cell                                                       
    //--------------------------------------------------------------------------
  private:
    vect     POS;                                  // position center           
    vect     VEL;                                  // velocity center           
    real     SIZE;                                 // size                      
    union {
      real  VRAD;                                  // radius of velocity sphere 
      real  RMAX;                                  // radius of position sphere 
    };
    //--------------------------------------------------------------------------
    // constructor                                                              
    //--------------------------------------------------------------------------
  public:
    sticky_cell() {}
    //--------------------------------------------------------------------------
    // non-const data access via member functions                               
    //--------------------------------------------------------------------------
    vect& vel () { return VEL; }
    vect& pos () { return POS; }
    real& size() { return SIZE; }
    real& rmax() { return RMAX; }
    real& vrad() { return VRAD; }
    //--------------------------------------------------------------------------
    // const data access via friends                                            
    //--------------------------------------------------------------------------
#define CCCC const sticky_cell*const&C
    friend vect const &vel (CCCC) { return C->VEL; }
    friend vect const &pos (CCCC) { return C->POS; }
    friend real const &size(CCCC) { return C->SIZE; }
    friend real const &rmax(CCCC) { return C->RMAX; }
    friend real const &vrad(CCCC) { return C->VRAD; }
#undef CCCC
    //--------------------------------------------------------------------------
    // other const and static methods                                           
    //--------------------------------------------------------------------------
    static void dump_head(std::ostream&o)
    {
      o<<"     #";
      flagged_cell<sticky_cell,sticky_soul>::dump_head(o);
      o<<"           size     rmax/ velocity           [vrad]";
    }
    //--------------------------------------------------------------------------
    void dump(std::ostream&o,
	      const sticky_soul*const &S0,
	      const sticky_cell*const &C0) const
    {
      o<<" "<<std::setw(5)<<int(this-C0);
      flagged_cell<sticky_cell,sticky_soul>::dump(o,S0,C0);
      o<<" "<<std::setw(6)<<std::setprecision(4)<<SIZE;
      if(is_set(basic_flag::STICKY)) {
	for(register indx i=0; i<NDIM; i++)
	  o<<" "<<std::setw(9)<<std::setprecision(4)<<VEL[i];
	o<<" "<<std::setw(5)<<std::setprecision(4)<<VRAD;
      }
      else if (is_set(basic_flag::SPH))
	o<<" "<<std::setw(5)<<std::setprecision(4)<<RMAX;
    }
  };
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::sticky_tree                                                    
  //////////////////////////////////////////////////////////////////////////////
  class sticky_tree :
    public sub_tree<sticky_cell,cell>
  {
    sticky_tree           (const sticky_tree&);    // not implemented           
    sticky_tree& operator=(const sticky_tree&);    // not implemented           
    //--------------------------------------------------------------------------
    // data of class sticky_tree                                                
    //--------------------------------------------------------------------------
    const areal *S,**V;                            // arrays with sizes and vels
    //--------------------------------------------------------------------------
    // public methods                                                           
    //--------------------------------------------------------------------------
  public:
    sticky_tree(basic_tree<cell>*parent,
		const areal*s =0, const areal*v[NDIM] =0) :
      sub_tree<sticky_cell,cell>(parent), S(s), V(v) {}
    void prepare_sticky();
    void prepare_sph();
    void set_num(soul*)  const;
    void set_num(areal*) const;
  };
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::basic_iactor                                                   
  //////////////////////////////////////////////////////////////////////////////
  class basic_iactor
  {
    //--------------------------------------------------------------------------
    // types of class basic_finder                                              
    //--------------------------------------------------------------------------
  public:
    typedef sticky_cell                 cell_type;
    typedef sticky_soul                 soul_type;
    typedef        uint                 elem_pair[2];
    //--------------------------------------------------------------------------
    // abstract methods                                                         
    //--------------------------------------------------------------------------
  protected:
    virtual void many       (const bool, soul_type*, soul_type*,
		             const soul_type*)                   const = 0;
    virtual void single     (soul_type*, soul_type*)             const = 0;
    virtual bool discard    (cell_type*, soul_type*)             const = 0;
    virtual bool discard    (cell_type*, cell_type*)             const = 0;
  public:
    virtual bool split_first(const cell_type*, const cell_type*) const = 0;
    //--------------------------------------------------------------------------
    // other methods                                                            
    //--------------------------------------------------------------------------
  protected:
    basic_iactor() {}
    bool many  (cell_type*, soul_type*) const;
    bool many  (cell_type*, cell_type*) const;
    bool many  (cell_type*)             const;
  public:
    bool interact(cell_type*)             const;
    bool interact(cell_type*, cell_type*) const;
    bool interact(cell_type*, soul_type*) const;
    void interact(soul_type*A, soul_type*B) const { single(A,B); }
  };
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::basic_finder                                                   
  //////////////////////////////////////////////////////////////////////////////
  class basic_finder :
    public basic_iactor
  {
    //--------------------------------------------------------------------------
    // data of class basic_finder                                               
    //--------------------------------------------------------------------------
  private:
    const uint    MAX;                             // maximal size of list      
    elem_pair    *BL;                              // list of interaction pairs 
    mutable uint  N;                               // actual size of list       
    //--------------------------------------------------------------------------
    // other methods                                                            
    //--------------------------------------------------------------------------
  protected:
    void add_pair(soul_type*, soul_type*) const;
  public:
    uint const &actual_size_of_list() const { return N; }
    //--------------------------------------------------------------------------
    // constructors                                                             
    //--------------------------------------------------------------------------
    basic_finder(const uint n, elem_pair*l) : MAX(n),BL(l),N(0){}
  };
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::sticky_finder                                                  
  //////////////////////////////////////////////////////////////////////////////
  class sticky_finder :
    public basic_finder
  {
  private:
    const real    TAU;                             // time period               
  protected:
    void many   (const bool, soul_type*, soul_type*, const soul_type*) const;
    void single (soul_type*, soul_type*) const;
    bool discard(cell_type*, soul_type*) const;
    bool discard(cell_type*, cell_type*) const;
  public:
    sticky_finder(const real t, const uint n, elem_pair*l) :
      basic_finder(n,l), TAU(t) {}
    bool split_first(const cell_type*A, const cell_type*B) const {
      return is_twig(B) || size(A)+TAU*vrad(A) > size(B)+TAU*vrad(B);
    }
  };
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::neighbour_finder                                               
  //////////////////////////////////////////////////////////////////////////////
  class neighbour_finder :
    public basic_finder
  {
  protected:
    void many   (const bool, soul_type*, soul_type*, const soul_type*) const;
    void single (soul_type*, soul_type*) const;
    bool discard(cell_type*, soul_type*) const;
    bool discard(cell_type*, cell_type*) const;
  public:
    neighbour_finder(const uint n, elem_pair*l) : basic_finder(n,l) {}
    bool split_first(const cell_type* A, const cell_type* B) const {
      return is_twig(B) || size(A) > size(B);
    }
  };
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::neighbour_counter                                              
  //////////////////////////////////////////////////////////////////////////////
  class neighbour_counter :
    public basic_iactor
  {
  protected:
    void many   (const bool, soul_type*, soul_type*, const soul_type*) const;
   void single (soul_type*, soul_type*) const;
    bool discard(cell_type*, soul_type*) const;
    bool discard(cell_type*, cell_type*) const;
  public:
    neighbour_counter() {}
    bool split_first(const cell_type* A, const cell_type* B) const {
      return is_twig(B) || size(A) > size(B); }
  };
}
//------------------------------------------------------------------------------
#endif // included_stic_h
