// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// tree.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class basic_tree<CELL,SOUL>                                                 |
// class tree<CELL,SOUL>                                                       |
// class sub_tree<CELL,SOUL,PCELL,PSOUL>                                       |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_tree_h
#define included_tree_h

#ifndef included_auxx_h
#include <nbdy/auxx.h>
#endif

#ifndef included_flag_h
#include <nbdy/flag.h>
#endif
////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::basic_tree<CELL>                                               
  //////////////////////////////////////////////////////////////////////////////
  template<class CELL>                             // cell type of tree         
  class basic_tree {
  public:
    basic_tree           (const basic_tree&);      // not implemented           
    basic_tree& operator=(const basic_tree&);      // not implemented           
    //--------------------------------------------------------------------------
    // types of class basic_tree                                                
    //--------------------------------------------------------------------------
    typedef real scalar_type;                      // type of real number       
    typedef vect vector_type;                      // type of vector of reals   
    typedef CELL cell_type;                        // type of tree cell         
    typedef typename CELL::soul_type soul_type;    // type of tree soul         
    typedef typename soul_type::bodies_type bodies_type; // type of bodies      
    typedef typename bodies_type::body_type body_type;   // type of body        
    //--------------------------------------------------------------------------
    // data of class basic_tree                                                 
    //--------------------------------------------------------------------------
  protected:
    const bodies_type*BODIES;                      // bodies                    
    unsigned          Ns,Nc,Nz,Dp;                 // # souls/cells/zombs, depth
    soul_type        *S0,*SN;                      // begin & end of souls      
    cell_type        *C0,*CN;                      // begin & end of cells      
    //--------------------------------------------------------------------------
    // protected methods                                                        
    //--------------------------------------------------------------------------
    basic_tree(const bodies_type* b = 0) : 
      BODIES(b), Ns(0), Nc(0), Nz(0), Dp(0), S0(0), SN(0), C0(0), CN(0) {}
    //--------------------------------------------------------------------------
    // public methods                                                           
    //--------------------------------------------------------------------------
  public:
    void mark_for_subtree(uint &, uint&);
    const bodies_type* const&my_bodies() const { return BODIES; }
    bool      const  use_bodies  () const { return BODIES != 0; }
    unsigned  const& N_souls     () const { return Ns; }
    unsigned  const& N_cells     () const { return Nc; }
    unsigned  const& N_zombies   () const { return Nz; }
    unsigned  const& depth       () const { return Dp; }
    //--------------------------------------------------------------------------
    // access to cells & souls                                                  
    //--------------------------------------------------------------------------
    cell_type*const& root          () const { return C0; }
    cell_type*const& begin_cells   () const { return C0; }
    cell_type*const& end_cells     () const { return CN; }
    cell_type*const  last_cells    () const { return CN-1; }
    cell_type*const  rbegin_cells  () const { return CN-1; }
    cell_type*const  rend_cells    () const { return C0-1; }
    soul_type*const& begin_souls   () const { return S0; }
    soul_type*const& end_souls     () const { return SN; }
    body_type        begin_bodies  () const { return BODIES->begin(); }
    body_type        end_bodies    () const { return BODIES->end(); }
    body_type        end_SPH_bodies() const { return BODIES->end_SPH(); }
    //--------------------------------------------------------------------------
    // non-const access to cells & souls                                        
    //--------------------------------------------------------------------------
    cell_type*     & root        () { return C0; }
    cell_type*       begin_cells () { return C0; }
    cell_type*       end_cells   () { return CN; }
    cell_type*       last_cells  () { return CN-1; }
    cell_type*       rbegin_cells() { return CN-1; }
    cell_type*       rend_cells  () { return C0-1; }
    soul_type*       begin_souls () { return S0; }
    soul_type*       end_souls   () { return SN; }
    //--------------------------------------------------------------------------
    // destructor                                                               
    //--------------------------------------------------------------------------
    ~basic_tree() {
      if(S0) delete[] S0;
      if(C0) delete[] C0;
    }
  };
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::tree<CELL>                                                     
  //////////////////////////////////////////////////////////////////////////////
  template<class CELL>                             // cell type of tree         
  class tree :
    public basic_tree<CELL>
  {
    tree           (const tree&);                  // not implemented           
    tree& operator=(const tree&);                  // not implemented           
    //--------------------------------------------------------------------------
    // construction from list of bodies -> build new tree from scratch          
    //--------------------------------------------------------------------------
  public:
    tree(const bodies_type*,                       // I: bodies                 
	 const vect     * = 0,                     // [I: root center]          
	 const unsigned   = 6,                     // [I: N_crit]               
	 const unsigned   = 50);                   // [I: max tree depth]       
    //--------------------------------------------------------------------------
    // construction from arrays of areal for positions                          
    //--------------------------------------------------------------------------
    tree(const flag_type*,                         // I: array with flags       
	 const areal    *[NDIM],                   // I: arrays with x,y,z      
	 const unsigned,                           // I: size of arrays         
	 const vect     * = 0,                     // [I: root center]          
	 const unsigned   = 6,                     // [I: N_crit]               
	 const unsigned   = 50);                   // [I: max tree depth]       
//     //--------------------------------------------------------------------------
//     // re-construction after souls have slightly changed their positions        
//     //--------------------------------------------------------------------------
//     void rebuild(const unsigned   = 6,             // [I: N_crit]               
// 		 const unsigned   = 50);           // [I: max tree depth]       
  };
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::sub_tree<SUB_CELL,PARENT_CELL>                                 
  //////////////////////////////////////////////////////////////////////////////
  const flag_type SUBTREE = 128;                   // node = node of subtree    
  inline void flag_for_subtree(basic_flag*F) { F->add_flag(SUBTREE); }
  inline void unflag_subtree  (basic_flag*F) { F->un_flag (SUBTREE); }
  //////////////////////////////////////////////////////////////////////////////
  template<class SUB_CELL, class PARENT_CELL>      // cell types                
  class sub_tree :
    public basic_tree<SUB_CELL>
  {
    sub_tree           (const sub_tree&);          // not implemented           
    sub_tree& operator=(const sub_tree&);          // not implemented           
    //--------------------------------------------------------------------------
    // construction as sub-tree from another tree                               
    //--------------------------------------------------------------------------
  public:
    sub_tree(basic_tree<PARENT_CELL>*);            // I: parent tree            
  };
}
////////////////////////////////////////////////////////////////////////////////
// useful macros                                                                
////////////////////////////////////////////////////////////////////////////////
//------------------------------------------------------------------------------
// loop the cells down, root first; using non-const cell pointers               
//------------------------------------------------------------------------------
#define LoopCellsDown(TREE_TYPE,                   /* type of parent tree    */\
		      TREE_PTER,                   /* pointer to parent tree */\
		      NAME)                        /* name for cell* s       */\
  for(register TREE_TYPE::cell_type	           /* type of cell           */\
     *NAME  = TREE_PTER->begin_cells();            /* from root cell         */\
      NAME != TREE_PTER->end_cells();              /* until beyond last cell */\
    ++NAME)                                        // get next cell             
//------------------------------------------------------------------------------
// loop the cells down, root first; using const cell pointers                   
//------------------------------------------------------------------------------
#define LoopConstCellsDown(TREE_TYPE,              /* type of parent tree    */\
		           TREE_PTER,              /* pointer to parent tree */\
		           NAME)                   /* name for cell* s       */\
  for(register TREE_TYPE::cell_type const          /* type of cell           */\
     *NAME  = TREE_PTER->begin_cells();            /* from root cell         */\
      NAME != TREE_PTER->end_cells();              /* until beyond last cell */\
    ++NAME)                                        // get next cell             
//------------------------------------------------------------------------------
// loop the cells up, root last; using non-const cell pointers                  
//------------------------------------------------------------------------------
#define LoopCellsUp(TREE_TYPE,                     /* type of parent tree    */\
		    TREE_PTER,                     /* pointer to parent tree */\
	            NAME)                          /* name for cell* s       */\
  for(register TREE_TYPE::cell_type	           /* type of cell           */\
     *NAME  = TREE_PTER->rbegin_cells();           /* from last cell         */\
      NAME != TREE_PTER->rend_cells();             /* until beyond root cell */\
    --NAME)                                        // get previous cell         
//------------------------------------------------------------------------------
// loop the cells up, root last; using const cell pointers                      
//------------------------------------------------------------------------------
#define LoopConstCellsUp(TREE_TYPE,                /* type of parent tree    */\
		         TREE_PTER,                /* pointer to parent tree */\
		         NAME)                     /* name for cell* s       */\
  for(register TREE_TYPE::cell_type const          /* type of cell           */\
     *NAME  = TREE_PTER->rbegin_cells();           /* from last cell         */\
      NAME != TREE_PTER->rend_cells();             /* until beyond root cell */\
    --NAME)                                        // get previous cell         
//------------------------------------------------------------------------------
// loop the soulls up, root last; using non-const soul pointers                  
//------------------------------------------------------------------------------
#define LoopSouls(TREE_TYPE,                       /* type of parent tree    */\
		  TREE_PTER,                       /* pointer to parent tree */\
	          NAME)                            /* name for cell* s       */\
  for(register TREE_TYPE::soul_type	           /* type of cell           */\
     *NAME  = TREE_PTER->begin_souls();            /* from last cell         */\
      NAME != TREE_PTER->end_souls();              /* until beyond root cell */\
    ++NAME)                                        // get previous cell         
//------------------------------------------------------------------------------
// loop the souls up, root last; using const soul pointers                      
//------------------------------------------------------------------------------
#define LoopConstSouls(TREE_TYPE,                  /* type of parent tree    */\
		       TREE_PTER,                  /* pointer to parent tree */\
		       NAME)                       /* name for cell* s       */\
  for(register TREE_TYPE::soul_type const          /* type of cell           */\
     *NAME  = TREE_PTER->begin_souls();            /* from last cell         */\
      NAME != TREE_PTER->end_souls();              /* until beyond root cell */\
    ++NAME)                                        // get previous cell         
//------------------------------------------------------------------------------
// loop the bodies                                                              
//------------------------------------------------------------------------------
#ifndef LoopBodies
#define LoopBodies(TREE_TYPE,                     /* type of bodies         */	\
		   TREE_PTER,                     /* pointer to bodies      */	\
		   NAME)                          /* name for body          */	\
  for(register TREE_TYPE::body_type               /* type of body           */	\
      NAME  = TREE_PTER->begin_bodies();          /* from first body        */	\
      NAME != TREE_PTER->end_bodies();            /* until beyond last body */	\
    ++NAME)                                       /* get next body          */
#endif
////////////////////////////////////////////////////////////////////////////////
#endif // included_tree_h
