// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// vect.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 1999-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_vect_h
#define included_vect_h

#ifndef included_inln_h
#include <util/inln.h>
#endif

#ifndef NDIM
#error "NDIM not #defined in vect.h"
#endif
////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
#if NDIM == 2

#define CI(op,b,ox) A[0] op b[0] ox  A[1] op b[1]
#define CX(op,b,ox) A[0] op b    ox  A[1] op b
#define FI(op,b)    A[0] op b[0];    A[1] op b[1]
#define FX(op,b)    A[0] op b;       A[1] op b

#else

#define CI(op,b,ox) A[0] op b[0] ox  A[1] op b[1] ox A[2] op b[2]
#define CX(op,b,ox) A[0] op b    ox  A[1] op b    ox A[2] op b
#define FI(op,b)    A[0] op b[0];    A[1] op b[1];   A[2] op b[2]
#define FX(op,b)    A[0] op b;       A[1] op b;      A[2] op b

#endif

#define OP      operator
  //////////////////////////////////////////////////////////////////////////////
  // using directives                                                           
  //////////////////////////////////////////////////////////////////////////////
  using util::max;
  using util::abs;
  using util::square;
  //////////////////////////////////////////////////////////////////////////////
  // some forward declarations                                                  
  //////////////////////////////////////////////////////////////////////////////
  template<typename REAL> class sym2;
  template<typename REAL> class sym3;
  template<typename REAL> class evector;
  template<typename REAL> class esym2;
  template<typename REAL> class esym3;
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::vector                                                         
  //////////////////////////////////////////////////////////////////////////////
  template<typename REAL>                      // scalar type                   
  class vector {                               // still over templated          
    //--------------------------------------------------------------------------
    // friendships                                                              
    //--------------------------------------------------------------------------
    friend class sym2<REAL>;
    friend class sym3<REAL>;
    friend class evector<REAL>;
    friend class esym2<REAL>;
    friend class esym3<REAL>;
    //--------------------------------------------------------------------------
    // types                                                                    
    //--------------------------------------------------------------------------
  private:
    typedef REAL                    real;      // type for const returns        
    typedef vector                  vect;      // type for const returns        
    typedef const int             c_indx;      // type for const args           
    typedef const REAL           &c_real;      // type for const args           
    typedef const vector         &c_vect;      // type for const args           
    typedef const evector<REAL>  &c_evec;      // type for const args           
    //--------------------------------------------------------------------------
    // data                                                                     
    //--------------------------------------------------------------------------
  private:
    REAL A[NDIM];
    //--------------------------------------------------------------------------
    // construction                                                             
    //--------------------------------------------------------------------------
  public:
    vector ()         {}
    vector (c_real x) { FX(=,x); }
    vector (c_vect x) { FI(=,x.A); }
    vector (c_evec x) { FI(=,x.A); }
    //--------------------------------------------------------------------------
    // non-const methods                                                        
    //--------------------------------------------------------------------------
    real* pointer()                   { return A; }
    real &OP[]   (c_indx i)           { return A[i]; }
    vect &OP=    (c_real x)           { FX(=,x); return *this; }
    vect &OP=    (const real* x)      { FI(=,x); return *this; }
    vect &OP=    (c_vect x)           { FI(=,x.A); return *this; }
    vect &OP=    (c_evec x)           { FI(=,x.A); return *this; }
    vect &OP+=   (c_vect x)           { FI(+=,x.A); return *this; }
    vect &OP+=   (c_evec x)           { FI(+=,x.A); return *this; }
    vect &OP-=   (c_vect x)           { FI(-=,x.A); return *this; }
    vect &OP-=   (c_evec x)           { FI(-=,x.A); return *this; }
    vect &OP*=   (c_real x)           { FX(*=,x); return *this; }
    vect &OP/=   (c_real x)           { return OP*=(real(1)/x); }
    vect &negate ()                   { FI(=-,A); return *this; }
    vect &add_mul(c_vect y, c_real x) { FI(+=,x*y.A); return *this; }
    vect &add_mul(c_evec y, c_real x) { FI(+=,x*y.A); return *this; }
    vect &sub_mul(c_vect y, c_real x) { FI(-=,x*y.A); return *this; }
    vect &sub_mul(c_evec y, c_real x) { FI(-=,x*y.A); return *this; }
    //--------------------------------------------------------------------------
    vect &apply  (real(*f)(c_real))   { 
      A[0]=f(A[0]); A[1]=f(A[1]); 
#if NDIM==3
      A[2]=f(A[2]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    vect &connect(c_vect x, real(*f)(c_real, c_real)) {
      A[0]=f(A[0],x.A[0]); A[1]=f(A[1],x.A[1]);
#if NDIM==3
      A[2]=f(A[2],x.A[2]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    vect &connect(c_evec x, real(*f)(c_real, c_real)) {
      A[0]=f(A[0],x.A[0]); A[1]=f(A[1],x.A[1]);
#if NDIM==3
      A[2]=f(A[2],x.A[2]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    // const methods                                                            
    //--------------------------------------------------------------------------
    const real* const_pointer() const { return A; }
    real const&OP[](c_indx i) const { return A[i]; }
    vect OP*       (c_real x) const { register vect y=*this; y*=x; return y; }
    vect OP/       (c_real x) const { register vect y(*this); y*=1/x;return y; }
    vect OP+       (c_vect x) const { register vect y(*this); y+=x; return y; }
    vect OP+       (c_evec x) const { register vect y(*this); y+=x; return y; }
    vect OP-       (c_vect x) const { register vect y(*this); y-=x; return y; }
    vect OP-       (c_evec x) const { register vect y(*this); y-=x; return y; }
    vect OP-       ()         const { register vect y(0); y-= *this; return y; }
    real norm      ()         const { return CI(*,A,+); }
    real OP*       (c_vect x) const { return CI(*,x.A,+); }
    real OP*       (c_evec x) const { return CI(*,x.A,+); }
    //--------------------------------------------------------------------------
#if NDIM == 2
    real maxnorm   ()         const { max(abs(A[0]),abs(A[1])); }
    real dist_sq   (c_vect x) const { return square(A[0]-x.A[0])
				           + square(A[1]-x.A[1]); }
    real dist_sq   (c_evec x) const { return square(A[0]-x.A[0])
				           + square(A[1]-x.A[1]); }
    real sum_sq    (c_vect x) const { return square(A[0]+x.A[0])
				           + square(A[1]+x.A[1]); }
    real sum_sq    (c_evec x) const { return square(A[0]+x.A[0])
				           + square(A[1]+x.A[1]); }
    real OP^       (c_vect x) const { return = A[0]*x.A[1] - A[1]*x.A[0]; }
    real OP^       (c_evec x) const { return = A[0]*x.A[1] - A[1]*x.A[0]; }
#else
    //==========================================================================
    real maxnorm   ()         const { return max(abs(A[0]),
						 max(abs(A[1]),abs(A[2]))); }
    //--------------------------------------------------------------------------
    real dist_sq   (c_vect x) const { return square(A[0]-x.A[0])
				           + square(A[1]-x.A[1])
				           + square(A[2]-x.A[2]); }
    real dist_sq   (c_evec x) const { return square(A[0]-x.A[0])
				           + square(A[1]-x.A[1])
				           + square(A[2]-x.A[2]); }
    //--------------------------------------------------------------------------
    real sum_sq    (c_vect x) const { return square(A[0]+x.A[0])
				           + square(A[1]+x.A[1])
				           + square(A[2]+x.A[2]); }
    real sum_sq    (c_evec x) const { return square(A[0]+x.A[0])
				           + square(A[1]+x.A[1])
				           + square(A[2]+x.A[2]); }
    //--------------------------------------------------------------------------
    vect OP^       (c_vect x) const {
      register vect z;
      z.A[0] = A[1]*x.A[2] - A[2]*x.A[1];
      z.A[1] = A[2]*x.A[0] - A[0]*x.A[2];
      z.A[2] = A[0]*x.A[1] - A[1]*x.A[0];
      return z;
    }
    vect OP^       (c_evec x) const {
      register vect z;
      z.A[0] = A[1]*x.A[2] - A[2]*x.A[1];
      z.A[1] = A[2]*x.A[0] - A[0]*x.A[2];
      z.A[2] = A[0]*x.A[1] - A[1]*x.A[0];
      return z;
    }
#endif
    //--------------------------------------------------------------------------
    bool OP==   (c_real x) const { return CX(==,x,&&); }
    bool OP==   (c_vect x) const { return CI(==,x.A,&&); }
    bool OP==   (c_evec x) const { return CI(==,x.A,&&); }
    bool OP!=   (c_real x) const { return CX(!=,x,||); }
    bool OP!=   (c_vect x) const { return CI(!=,x.A,||); }
    bool OP!=   (c_evec x) const { return CI(!=,x.A,||); }
  };
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::evector                                                        
  //////////////////////////////////////////////////////////////////////////////
  template<typename REAL>                      // scalar type                   
  class evector {
    //--------------------------------------------------------------------------
    // friendships                                                              
    //--------------------------------------------------------------------------
    friend class vector<REAL>;
    friend class sym2;
    friend class sym3;
    friend class esym2;
    friend class esym3;
    //--------------------------------------------------------------------------
    // types                                                                    
    //--------------------------------------------------------------------------
  private:
    typedef REAL              real;            // type for const returns        
    typedef evector           evec;            // type for non-const returns    
    typedef vector<REAL>      vect;            // type for const returns        
    typedef const int       c_indx;            // type for const args           
    typedef const REAL     &c_real;            // type for const args           
    typedef const vect     &c_vect;            // type for const args           
    typedef const evector  &c_evec;            // type for const args           
    //--------------------------------------------------------------------------
    // data                                                                     
    //--------------------------------------------------------------------------
  private:
    REAL *A;
    //--------------------------------------------------------------------------
    // construction                                                             
    //--------------------------------------------------------------------------
  public:
    evector (real* x) : A(x) {}
    //--------------------------------------------------------------------------
    // non-const methods                                                        
    //--------------------------------------------------------------------------
    real* pointer()                   { return A; }
    real &OP[]   (c_indx i)           { return A[i]; }
    evec &OP=    (c_real x)           { FX(=,x); return *this; }
    evec &OP=    (const real* x)      { FI(=,x); return *this; }
    evec &OP=    (c_vect x)           { FI(=,x.A); return *this; }
    evec &OP=    (c_evec x)           { FI(=,x.A); return *this; }
    evec &OP+=   (c_vect x)           { FI(+=,x.A); return *this; }
    evec &OP+=   (c_evec x)           { FI(+=,x.A); return *this; }
    evec &OP-=   (c_vect x)           { FI(-=,x.A); return *this; }
    evec &OP-=   (c_evec x)           { FI(-=,x.A); return *this; }
    evec &OP*=   (c_real x)           { FX(*=,x); return *this; }
    evec &OP/=   (c_real x)           { return OP*=(real(1)/x); }
    evec &negate ()                   { FI(=-,A); return *this; }
    evec &add_mul(c_vect y, c_real x) { FI(+=,x*y.A); return *this; }
    evec &add_mul(c_evec y, c_real x) { FI(+=,x*y.A); return *this; }
    evec &sub_mul(c_vect y, c_real x) { FI(-=,x*y.A); return *this; }
    evec &sub_mul(c_evec y, c_real x) { FI(-=,x*y.A); return *this; }
    //--------------------------------------------------------------------------
    evec &apply  (real(*f)(c_real))   { 
      A[0]=f(A[0]); A[1]=f(A[1]); 
#if NDIM==3
      A[2]=f(A[2]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    evec &connect(c_vect x, real(*f)(c_real, c_real)) {
      A[0]=f(A[0],x.A[0]); A[1]=f(A[1],x.A[1]);
#if NDIM==3
      A[2]=f(A[2],x.A[2]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    evec &connect(c_evec x, real(*f)(c_real, c_real)) {
      A[0]=f(A[0],x.A[0]); A[1]=f(A[1],x.A[1]);
#if NDIM==3
      A[2]=f(A[2],x.A[2]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    // const methods                                                            
    //--------------------------------------------------------------------------
    const real* const_pointer()     const { return A; }
    real const&OP[](c_indx i) const { return A[i]; }
    vect OP*       (c_real x) const { register vect y=*this; y*=x; return y; }
    vect OP/       (c_real x) const { register vect y(*this); y*=1/x;return y; }
    vect OP+       (c_vect x) const { register vect y(*this); y+=x; return y; }
    vect OP+       (c_evec x) const { register vect y(*this); y+=x; return y; }
    vect OP-       (c_vect x) const { register vect y(*this); y-=x; return y; }
    vect OP-       (c_evec x) const { register vect y(*this); y-=x; return y; }
    vect OP-       ()         const { register vect y(0); y-= *this; return y; }
    real norm      ()         const { return CI(*,A,+); }
    real OP*       (c_vect x) const { return CI(*,x.A,+); }
    real OP*       (c_evec x) const { return CI(*,x.A,+); }
    //--------------------------------------------------------------------------
#if NDIM == 2
    real maxnorm   ()         const { max(abs(A[0]),abs(A[1])); }
    real dist_sq   (c_vect x) const { return square(A[0]-x.A[0])
				           + square(A[1]-x.A[1]); }
    real dist_sq   (c_evec x) const { return square(A[0]-x.A[0])
				           + square(A[1]-x.A[1]); }
    real sum_sq    (c_vect x) const { return square(A[0]+x.A[0])
				           + square(A[1]+x.A[1]); }
    real sum_sq    (c_evec x) const { return square(A[0]+x.A[0])
				           + square(A[1]+x.A[1]); }
    real OP^       (c_vect x) const { return = A[0]*x.A[1] - A[1]*x.A[0]; }
    real OP^       (c_evec x) const { return = A[0]*x.A[1] - A[1]*x.A[0]; }
#else
    //==========================================================================
    real maxnorm   ()         const { return max(abs(A[0]),
						 max(abs(A[1]),abs(A[2]))); }
    //--------------------------------------------------------------------------
    real dist_sq   (c_vect x) const { return square(A[0]-x.A[0])
				           + square(A[1]-x.A[1])
				           + square(A[2]-x.A[2]); }
    real dist_sq   (c_evec x) const { return square(A[0]-x.A[0])
				           + square(A[1]-x.A[1])
				           + square(A[2]-x.A[2]); }
    //--------------------------------------------------------------------------
    real sum_sq    (c_vect x) const { return square(A[0]+x.A[0])
				           + square(A[1]+x.A[1])
				           + square(A[2]+x.A[2]); }
    real sum_sq    (c_evec x) const { return square(A[0]+x.A[0])
				           + square(A[1]+x.A[1])
				           + square(A[2]+x.A[2]); }
    //--------------------------------------------------------------------------
    vect OP^       (c_vect x) const {
      register vect z;
      z.A[0] = A[1]*x.A[2] - A[2]*x.A[1];
      z.A[1] = A[2]*x.A[0] - A[0]*x.A[2];
      z.A[2] = A[0]*x.A[1] - A[1]*x.A[0];
      return z;
    }
    vect OP^       (c_evec x) const {
      register vect z;
      z.A[0] = A[1]*x.A[2] - A[2]*x.A[1];
      z.A[1] = A[2]*x.A[0] - A[0]*x.A[2];
      z.A[2] = A[0]*x.A[1] - A[1]*x.A[0];
      return z;
    }
#endif
    //--------------------------------------------------------------------------
    bool OP==   (c_real x) const { return CX(==,x,&&); }
    bool OP==   (c_vect x) const { return CI(==,x.A,&&); }
    bool OP==   (c_evec x) const { return CI(==,x.A,&&); }
    bool OP!=   (c_real x) const { return CX(!=,x,||); }
    bool OP!=   (c_vect x) const { return CI(!=,x.A,||); }
    bool OP!=   (c_evec x) const { return CI(!=,x.A,||); }
  };
  //////////////////////////////////////////////////////////////////////////////
  // related functions (when declared as friends, compiler problems occur)      
  //////////////////////////////////////////////////////////////////////////////
  template<typename REAL> inline
  vector<REAL> OP*(const REAL&x, const vector<REAL>& y) { return y*x; }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  vector<REAL> OP*(const REAL&x, const evector<REAL>& y) { return y*x; }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  REAL norm(const vector<REAL>& x) { return x.norm(); }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  REAL norm(const evector<REAL>& x) { return x.norm(); }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  REAL dist_sq(const vector<REAL>&x,const vector<REAL>&y){ return x.dist_sq(y); }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  REAL dist_sq(const vector<REAL>&x,const evector<REAL>&y){ return x.dist_sq(y);}
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  REAL dist_sq(const evector<REAL>&x,const vector<REAL>&y){ return x.dist_sq(y);}
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  REAL dist_sq(const evector<REAL>&x,const evector<REAL>&y){return x.dist_sq(y);}
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  REAL sum_sq(const vector<REAL>&x,const vector<REAL>&y){ return x.sum_sq(y); }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  REAL sum_sq(const vector<REAL>&x,const evector<REAL>&y){ return x.sum_sq(y);}
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  REAL sum_sq(const evector<REAL>&x,const vector<REAL>&y){ return x.sum_sq(y);}
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  REAL sum_sq(const evector<REAL>&x,const evector<REAL>&y){return x.sum_sq(y);}
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  std::ostream& OP<<(std::ostream&s, const vector<REAL>& x) {
    return util::write_array<REAL,NDIM>(s,x.const_pointer());
  }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  std::ostream& OP<<(std::ostream&s, const evector<REAL>& x) {
    return util::write_array<REAL,NDIM>(s,x.const_pointer());
  }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  std::istream& OP>>(std::istream&s, vector<REAL>& x) {
    return util::read_array<REAL,NDIM>(s,x.pointer());
  }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  std::istream& OP>>(std::istream&s, evector<REAL>& x) {
    return util::read_array<REAL,NDIM>(s,x.pointer());
  }
  //----------------------------------------------------------------------------
}
#undef CI
#undef CX
#undef FI
#undef FX
#undef OP
////////////////////////////////////////////////////////////////////////////////
#endif	// included_vect_h
