// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// nums.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 1994-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// version 0.0    WD  12/09/2001   first stuff needed by king.cc               |
// version 0.1    WD  19/09/2001   added stuff needed by GalPot.cc             |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// CONTENTS                                                                    |
//                                                                             |
// error & warning messages                             v0.0                   |
// find position in ordered table                       v0.0                   |
// polynomial interpolation on grids                    v0.0                   |
// root finding                                         v0.0                   |
// Burlisch-Stoer integration of 1D real integrals      v0.0                   |
// Runge-Kutte 4th order integrator                     v0.0                   |
// Legendre polynomials and their derivatives           v0.1                   |
// cubic spline (#included from util/spln.h)            v0.1                   |
// Gauss-Legendre integration: points & weights         v0.1                   |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_nums_h
#define included_nums_h

#ifndef included_iostream
#  include <iostream>
#  define included_iostream
#endif
#ifndef included_inln_h
#  include <util/inln.h>
#endif
#ifndef included_tupl_h
#  include <util/tupl.h>
#endif

extern "C" void exit(int);

////////////////////////////////////////////////////////////////////////////////
namespace util {
  //----------------------------------------------------------------------------
  // error & warning messages                                                   
  //----------------------------------------------------------------------------
  inline void numerics_error(const char* msgs) {
    cerr << " ERROR in numerics: " << msgs << '\n';
    exit(1);
  }
  //----------------------------------------------------------------------------
  inline void numerics_warning(const char* msgs) {
    cerr << " WARNING in numerics: " << msgs << '\n';
  }
  //----------------------------------------------------------------------------
  // find position in ordered table                                             
  //----------------------------------------------------------------------------
  template<class scalar_type>
  int hunt(const scalar_type*xarr, const int n, const scalar_type x,
	   const int j) {
    // hunts the ordered table xarr for jlo such that xarr[jlo]<=x<xarr[jlo+1]  
    // on input j provides a guess for the final value of jlo.                  
    // for an ascendingly ordered array, we return                              
    //  -1 for         x < x[0]                                                 
    //   i for x[i] <= x < x[i+1]  if  0<=i<n                                   
    // n-1 for         x == x[n-1]                                              
    // n   for         x >  x[n-1]                                              
    int jm,jlo=j,jhi,l=n-1;
    int ascnd=(xarr[l]>xarr[0]);
    if(!ascnd && xarr[l]==xarr[0] ) return -1;	    // x_0 = x_l                
    if( ascnd && x<xarr[0] || !ascnd && x>xarr[0] ) return -1;
    if( ascnd && x>xarr[l] || !ascnd && x<xarr[l] ) return  n;

    if(jlo<0 || jlo>l) {                            // input guess not useful,  
      jlo = -1;                                     //   go to bisection below  
      jhi = n;
    } else {
      int inc = 1;
      if(x>=xarr[jlo] == ascnd) {                   // hunt upward              
	if(jlo == l) return (x==xarr[l])? l : n;
	jhi = jlo+1;
	while(x>=xarr[jhi] == ascnd) {              // not done hunting         
	  jlo =jhi;
	  inc+=inc;                                 // so double the increment  
	  jhi =jlo+inc;
	  if(jhi>l) {                               // off end of table         
	    jhi=n;
	    break;
	  }
	}
      } else {                                      // hunt downward            
	if(jlo == 0) return -1;
	jhi = jlo;
	jlo-= 1;
	while(x<xarr[jlo] == ascnd) {               // not done hunting         
	  jhi = jlo;
	  inc+= inc;                                // so double the increment  
	  jlo = jhi-inc;
	  if(jlo < 0) {                             // off end of table         
	    jlo = 0;
	    break;
	  }
	}
      }
    }
    while (jhi-jlo != 1) {                          // bisection phase          
      jm=(jhi+jlo) >> 1;
      if(x>=xarr[jm] == ascnd) jlo=jm;
      else jhi=jm;
    }
    return jlo;
  }
  //----------------------------------------------------------------------------
  template<class scalar_type>
  inline void find(int& klo, const int n, scalar_type const *x,
		   const scalar_type xi)
  {
    if(klo<0 || klo>=n-1 || x[klo]>xi || x[klo+1]<xi) {
      klo = int( (xi-x[0]) / (x[n-1]-x[0]) * (n-1) );
      klo = hunt(x,n,xi,klo);
      if(klo<0 || klo>=n) numerics_error("find(): x out of range");
    }
  }
  //----------------------------------------------------------------------------
  // polynomial interpolation on grids                                          
  //----------------------------------------------------------------------------
  template<class scalar_type>
  inline int find_for_polev(int& j, const int n, const int m,
			    const scalar_type *x, const scalar_type xi)
  {
    register int M=m;
    j = int( (xi-x[0]) / (x[n-1]-x[0]) * (n-1) );
    j = hunt(x,n,xi,j) - (m+1)/2 + 1;
    if(j>=0 && j<n && x[j]==xi)	 M = 1; 	    // no interpolation required
    else if(j<0)		 j = 0;
    else if(j>n-M)		 j = n-M;
    return M;
  }
  //----------------------------------------------------------------------------
  template<class scalar_type, class num_type>
  inline num_type polint(const scalar_type *xa,
			 const num_type *ya,
			 const int n,
			 const scalar_type x) {
    // polynom interpolation using n values
    register int i,m;
    register num_type   y, *P=new num_type[n];
    for(i=0;i!=n;++i) P[i]=ya[i];
    for(m=1;m!=n;++m)
      for(i=0;i<n-m;++i) {
	if(xa[i]==xa[i+m]) numerics_error("polev(): x's not distinct");
	P[i]= ( (x-xa[i+m])*P[i] + (xa[i]-x)*P[i+1] ) / (xa[i] - xa[i+m]);
      }
    y = P[0];    
    delete[] P;
    return y;
  }
  //----------------------------------------------------------------------------
  template<class scalar_type, class num_type>
  num_type polev(scalar_type x,
		 scalar_type*xarr,
		 num_type   *yarr,
		 const int   n,
		 const int   m=4) {
    // given the arrays xarr, yarr, polev returns y(x) using m of n values.
    int j, M=find_for_polev(j,n,m,xarr,x);
    return polint(xarr+j, yarr+j, M, x);
  }
  //----------------------------------------------------------------------------
  // root finding                                                               
  //----------------------------------------------------------------------------
  template <class scalar_type>
  scalar_type rtsafe(void(*func)(scalar_type,scalar_type&,scalar_type&),
		     const scalar_type x1,
		     const scalar_type x2,
		     const scalar_type xacc)
  {
    const int            maxit=100;
    register scalar_type xh,xl,dx,dxo,f,df,fh,fl,rts,swap,temp;
    (*func)(x1,fl,df);
    (*func)(x2,fh,df);
    if(fl*fh >= 0.) numerics_error("rtsafe(): root must be bracketed");
    if(fl<0.) {
      xl  = x1;
      xh  = x2;
    } else {
      xh  = x1;
      xl  = x2;
      swap= fl;
      fl  = fh;
      fh  = swap;
    }
    rts = 0.5*(x1+x2);
    dxo = abs(x2-x1);
    dx  = dxo;
    (*func)(rts,f,df);
    for (register int j=0; j!=maxit; ++j) {
      if((((rts-xh)*df-f)*((rts-xl)*df-f)>= 0.) || (abs(2.*f)>abs(dxo*df))) {
	dxo = dx;
	dx  = 0.5*(xh-xl);
	rts = xl+dx;
	if(xl==rts) return rts;
      } else {
	dxo = dx;
	dx  = f/df;
	temp=rts;
	rts-= dx;
	if(temp==rts) return rts;
      }
      if(abs(dx)<xacc) return rts;
      (*func)(rts,f,df);
      if(f<0.) {
	xl  = rts;
	fl  = f;
      } else {
	xh  = rts;
	fh  = f;
      }
    }
    numerics_warning("rtsafe(): max number of iterations exceeded");
    return rts;
  }
  //----------------------------------------------------------------------------
  // Burlisch-Stoer integration of 1D real integrals                            
  //----------------------------------------------------------------------------
  double qbulir(                                    // R:  int(f(x),x=a...b)    
		double(*)(double),                  // I:  f(x)                 
		const double,                       // I:  a                    
		const double,                       // I:  b                    
		const double,                       // I:  eps = error limit    
		double* = 0);                       //[O]: actual error estimate
  //----------------------------------------------------------------------------
  // Runge-Kutta 4th order integrator for ODEs                                  
  //----------------------------------------------------------------------------
  template<class scalar_type, class vector_type>
  inline
  vector_type rk4(const vector_type y,
		  const vector_type dy0,
		  const scalar_type x,
		  const scalar_type h,
		  vector_type(*derivs)(const scalar_type, const vector_type))
  {
    const    scalar_type hh=0.5*h, h6=h/6., xh=x+hh;
    register vector_type yt,dym,dyt;
    dyt = derivs(xh,y+hh*dy0);
    dym = derivs(xh,y+hh*dyt);
    yt  = y+h*dym;
    dym+= dyt;
    dyt = derivs(x+h,yt);
    return y+h6*(dy0+dyt+dym+dym);
  }
  //----------------------------------------------------------------------------
  template<class scalar_type, class vector_type>
  inline
  vector_type rk4(const vector_type y,
		  const scalar_type x,
		  const scalar_type h,
		  vector_type(*derivs)(const scalar_type, const vector_type))
  {
    return rk4(y,derivs(x,y),x,h,derivs);
  }
  //----------------------------------------------------------------------------
  // Legendre polynomials and their derivatives                                 
  //----------------------------------------------------------------------------
  template<class S, int N>
  void LegendrePeven(S*p, const double x) {
    // based on a routine from J.J. Binney                                      
    // evaluates even Legendre Polys up to l=2*(N-1) at x                       
    register int    n,l,l2;
    register double x2=x*x;
    p[0] = 1.;
    p[1] = 1.5*x2-0.5;
    for(n=2; n<N; n++) {
      l    = 2*(n-1);
      l2   = 2*l;
      p[n] = - p[n-2] * l*(l-1)/double((l2+1)*(l2-1))
	+ p[n-1] * (x2 - (l2*l+l2-1)/double((l2-1)*(l2+3)) );
      p[n]*= (l2+1)*(l2+3) / double((l+1)*(l+2));
    }
  }
  //----------------------------------------------------------------------------
  template<class S, int N>
  void dLegendrePeven(S*p, S*d, const double x)
    // based on a routine from J.J. Binney                                      
    // evaluates even Legendre Polys and its derivs up to l=2*(N-1) at x        
  {
    register int    n,l,l2;
    register double x2=x*x;
    p[0] = 1.;
    d[0] = 0.;
    p[1] = 1.5*x2-0.5;
    d[1] = 1.5;
    for(n=2; n<N; n++) {
      l    = 2*(n-1);
      l2   = 2*l;
      p[n] = - p[n-2] * l*(l-1)/double((l2+1)*(l2-1))
	+ p[n-1] * (x2 - (l2*l+l2-1)/double((l2-1)*(l2+3)) );
      p[n]*= (l2+1)*(l2+3) / double((l+1)*(l+2));
      d[n] = - d[n-2] * l*(l-1)/double((l2+1)*(l2-1))
	+ d[n-1] * (x2 - (l2*l+l2-1)/double((l2-1)*(l2+3)) )
	+ p[n-1];
      d[n]*= (l2+1)*(l2+3) / double((l+1)*(l+2));
    }
    x2 = 2*x;
    for(n=0; n<N; n++)
      d[n] *= x2;
  }
  //----------------------------------------------------------------------------
  template<class S, int N> inline
  void LegendrePeven(tupel<S,N>& p, const double x) {
    return LegendrePeven<S,N>((S*)(p),x);
  }
  //----------------------------------------------------------------------------
  template<class S, int N> inline
  void dLegendrePeven(tupel<S,N>& p, tupel<S,N>& d, const double x) {
    return dLegendrePeven<S,N>((S*)(p),(S*)(d),x);
  }
  //----------------------------------------------------------------------------
  // Gauss-Legendre integration: points & weights                               
  //----------------------------------------------------------------------------
  void GaussLegendre(double*, double*, const unsigned);
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_nums_h
