// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// body.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
#include <body.h>
#include <nbdy/nmio.h>
#include <util/ionl.h>
#include <iomanip>

using std::cerr;
using std::setw;
using std::ostream;
using std::istream;
using namespace nbdy;
////////////////////////////////////////////////////////////////////////////////
// class bodies: NEMO I/O                                                       
////////////////////////////////////////////////////////////////////////////////
#ifdef ALLOW_NEMO
void bodies::read_nemo_particles(const nemo_input&NEMO_I, real*TIME)
{
  register uint i;                                 // index                     
  register basic_body*Bi;                          // current basic_body        
  // 1. open parameter set; read N & time; close parameter set; allocate bodies 
  NEMO_I.open_set(nemo_io::param);                 // open nemo parameter set   
  register uint NN = NEMO_I.read_N();              //   read N                  
  if(TIME) {                                       //   IF time wanted         >
    if(NEMO_I.is_present(nemo_io::time))           //     IF time available    >
      NEMO_I.read(nemo_io::time,TIME);             //       read time           
    else {                                         //     < ELSE >              
      cerr<<"WARNING: no time found in snapshot; defaulting to 0"<<endl;
      *TIME = zero;                                //       warning, TIME=0     
    }                                              //     <                     
  }                                                //   <                       
  NEMO_I.close_set(nemo_io::param);                // close nemo parameter set  
  if(NN != N) {                                    // IF N differs from ours >  
    N    = NN;                                     //   re-set our N            
    NSPH = 0;                                      //   re-set our N_SPH        
    if(BB)   delete[] BB;                          //   de-alloc basic_body     
    if(BS) { delete[] BS; BS = 0; }                //   re-set step support     
    if(SS) { delete[] SS; SS = 0; }                //   re-set SPH support      
    BB = new basic_body[N];                        //   allocate N basic_body   
  }                                                // <                         
  // 2. open bodies set                                                         
  NEMO_I.open_set(nemo_io::bodies);                // open nemo particle set    
  // 2.1 read body masses                                                       
  if(NEMO_I.is_present(nemo_io::mass)) {           //   IF(masses to be read) > 
    NEMO_I.read(nemo_io::mass);                    //     read masses           
    for(i=0,Bi=BB; i!=N; ++i,++Bi)                 //     loop basic bodies   > 
      Bi->mass() = NEMO_I.bodies_scl(i);           //       copy mass         < 
  } else {                                         //   < ELSE (no masses)    > 
    cerr<<"WARNING: no masses in snapshot; defaulting to 1/N"<<endl;
    register real ms = one/real(N);                //     default mass          
    for(i=0,Bi=BB; i!=N; ++i,++Bi)                 //     loop basic bodies   > 
      Bi->mass() = ms;                             //       set mass          < 
  }                                                //   <                       
  NEMO_I.reset();                                  //   de-alloc I/O memory     
  // 2.2 read body positions and velocities                                     
  if(NEMO_I.is_present(nemo_io::posvel)) {         //   IF(phases to be read) > 
    NEMO_I.read(nemo_io::posvel);                  //     read phases           
    for(i=0,Bi=BB; i!=N; ++i,++Bi) {               //     loop basic bodies   > 
      Bi->pos()[0] = NEMO_I.bodies_phs(i)[0];      //       copy position       
      Bi->pos()[1] = NEMO_I.bodies_phs(i)[1];
      Bi->vel()[0] = NEMO_I.bodies_phs(i)[NDIM+0]; //       copy velocity       
      Bi->vel()[1] = NEMO_I.bodies_phs(i)[NDIM+1];
#if NDIM==3
      Bi->pos()[2] = NEMO_I.bodies_phs(i)[2];
      Bi->vel()[2] = NEMO_I.bodies_phs(i)[NDIM+2];
#endif
    }                                              //     <                     
  } else if(NEMO_I.is_present(nemo_io::pos) &&     //   < ELSE IF(positions and 
	    NEMO_I.is_present(nemo_io::vel)) {     //     velocities to read)  >
    NEMO_I.read(nemo_io::pos);
    NEMO_I.read(nemo_io::vel);
    for(i=0,Bi=BB; i!=N; ++i,++Bi) {               //     loop basic bodies   > 
      Bi->pos()[0] = NEMO_I.bodies_vec(i)[0];      //       copy position       
      Bi->pos()[1] = NEMO_I.bodies_vec(i)[1];
      Bi->vel()[0] = NEMO_I.bodies_vel(i)[0];      //       copy velocity       
      Bi->vel()[1] = NEMO_I.bodies_vel(i)[1];
#if NDIM==3
      Bi->pos()[2] = NEMO_I.bodies_vec(i)[2];
      Bi->vel()[2] = NEMO_I.bodies_vel(i)[2];
#endif
    }
  } else {                                         //   < ELSE (no pos & vel) > 
    cerr<<"ERROR: no phase-space coordinates found in snapshot"<<endl;
    exit(1);                                       //     fatal error           
  }                                                //   <                       
  NEMO_I.close_set(nemo_io::bodies);               // close nemo particle set   
  NEMO_I.reset();                                  // reset nemo_input          
}
//------------------------------------------------------------------------------
void bodies::read_nemo_snapshot(const nemo_input&NEMO_I, real* TIME)
{
  if(!NEMO_I.is_present(nemo_io::snap)) {          // IF no snapshot present  > 
    cerr<<"ERROR: no snapshot found"<<endl;        //   issue error message     
    exit(1);                                       //   and abort               
  }                                                // <                         
  NEMO_I.open_set(nemo_io::snap);                  //   open nemo snapshot      
  read_nemo_particles(NEMO_I,TIME);                //     read particles        
  NEMO_I.close_set(nemo_io::snap);                 //   close nemo snapshot     
}
//------------------------------------------------------------------------------
void bodies::write_nemo_particles(nemo_output&NEMO_O,
				  const add_out AO,
				  const real   *TIME) const
{
  register uint i;                                 // index                     
  register basic_body*Bi;                          // current basic_body        
  // 1. open parameter set; write out N & time; close parameter set;            
  NEMO_O.open_set(nemo_io::param);                 // open a parameter set      
  NEMO_O.write_N(N);                               //   write N                 
  if(TIME)                                         //   IF(time to be written)  
    NEMO_O.write(nemo_io::time,*TIME);             //     write out time        
  NEMO_O.close_set(nemo_io::param);                // close parameter set       
  // 2. allocate memory for output; open particle set; output data;             
  NEMO_O.allocarrays();                            // allocate memory for x,v   
  NEMO_O.allocscalar();                            // allocate memory for m...  
  NEMO_O.open_set(nemo_io::bodies);                // open particles set        
  NEMO_O.write(nemo_io::cart);                     //   put out coord system    
  // 2.1 output masses                                                          
  for(i=0,Bi=BB; i!=N; ++i,++Bi)                   //   loop basic bodies      >
    NEMO_O.bodies_scl(i) = Bi->mass();             //     copy mass into aux   <
  NEMO_O.write(nemo_io::mass);                     //   write out masses        
  // 2.2 output phases                                                          
  for(i=0,Bi=BB; i!=N; ++i,++Bi) {                 //   loop basic bodies      >
    NEMO_O.bodies_phs(i)[0]      = Bi->pos()[0];   //     copy pos into phases  
    NEMO_O.bodies_phs(i)[1]      = Bi->pos()[1];
    NEMO_O.bodies_phs(i)[NDIM+0] = Bi->vel()[0];   //     copy vel into phases  
    NEMO_O.bodies_phs(i)[NDIM+1] = Bi->vel()[1];
#if NDIM==3
    NEMO_O.bodies_phs(i)[2]      = Bi->pos()[2];
    NEMO_O.bodies_phs(i)[NDIM+2] = Bi->vel()[2];
#endif
  }                                                //   <                       
  NEMO_O.write(nemo_io::posvel);                   //   write out phases        
  // 2.3 output accelerations                                                   
  if(AO & accs) {                                  //   IF pots for output  >   
    for(i=0,Bi=BB; i!=N; ++i,++Bi) {               //     loop basic bodies    >
      NEMO_O.bodies_vec(i)[0] = nbdy::acc(Bi->SOUL)[0]; //  copy accelerations  
      NEMO_O.bodies_vec(i)[1] = nbdy::acc(Bi->SOUL)[1]; 
#if NDIM==3
      NEMO_O.bodies_vec(i)[2] = nbdy::acc(Bi->SOUL)[2]; 
#endif
    }                                              //     <                     
    NEMO_O.write(nemo_io::acc);                    //     write out accs        
  }                                                //   <                       
  // 2.4 output potential                                                       
  if(AO & pots) {                                  //   IF pots for output  >   
    for(i=0,Bi=BB; i!=N; ++i,++Bi)                 //     loop basic bodies    >
      NEMO_O.bodies_scl(i) = nbdy::pot(Bi->SOUL);  //       copy potentials    <
    NEMO_O.write(nemo_io::pot);                    //     write out potential   
  }                                                //   <                       
  // 2.5 output density                                                         
  if(AO & rhos || AO & surf) {                     //   IF rhos for output  >   
    for(i=0,Bi=BB; i!=N; ++i,++Bi)                 //     loop basic bodies    >
      NEMO_O.bodies_scl(i) = nbdy::rho(Bi->SOUL);  //       copy potentials    <
    NEMO_O.write(nemo_io::aux);                    //     write out potential   
  }                                                //   <                       
  // 2.6 output eps_i                                                           
  if(AO & epsi) {                                  //   IF rhos for output  >   
    for(i=0,Bi=BB; i!=N; ++i,++Bi)                 //     loop basic bodies    >
      NEMO_O.bodies_scl(i) = Bi->eps();            //       copy eps_i         <
    NEMO_O.write(nemo_io::eps);                    //     write out potential   
  }                                                //   <                       
  // 3. close particle set & reset nemo output                                  
  NEMO_O.close_set(nemo_io::bodies);               // close particles set       
  NEMO_O.reset();                                  // de-allocate arrays        
}
//------------------------------------------------------------------------------
void bodies::write_nemo_snapshot(nemo_output&NEMO_O, add_out AO,
				 const real*TIME) const
{
  NEMO_O.open_set(nemo_io::snap);                  // open a new nemo snapshot  
  write_nemo_particles(NEMO_O,AO,TIME);            //   write particle set      
  NEMO_O.close_set(nemo_io::snap);                 // close nemo snapshot       
}
////////////////////////////////////////////////////////////////////////////////
#endif // ALLLOW_NEMO
////////////////////////////////////////////////////////////////////////////////
// class bodies: ASCII I/O                                                      
////////////////////////////////////////////////////////////////////////////////
void bodies::write_ascii(ostream&out, add_out ao) const
{
  if(ao & epsi)
    if(ao & pots)
      if(ao & rhos)
	for(register basic_body*Bi=BB; Bi!=BB+N; ++Bi)
	  out<<Bi->mass()<<" "<<Bi->pos()<<" "<<Bi->vel()<<" "
	     <<Bi->eps() <<" "
	     <<nbdy::pot(Bi->SOUL)<<" "
	     <<nbdy::rho(Bi->SOUL)<<"\n";
      else
	for(register basic_body*Bi=BB; Bi!=BB+N; ++Bi)
	  out<<Bi->mass()<<" "<<Bi->pos()<<" "<<Bi->vel()<<" "
	     <<Bi->eps() <<" "
	     <<nbdy::pot(Bi->SOUL)<<"\n";
    else
      if(ao & rhos)
	for(register basic_body*Bi=BB; Bi!=BB+N; ++Bi)
	  out<<Bi->mass()<<" "<<Bi->pos()<<" "<<Bi->vel()<<" "
	     <<Bi->eps() <<" "
	     <<nbdy::rho(Bi->SOUL)<<"\n";
      else
	for(register basic_body*Bi=BB; Bi!=BB+N; ++Bi)
	  out<<Bi->mass()<<" "<<Bi->pos()<<" "<<Bi->vel()<<" "
	     <<Bi->eps() <<"\n";
  else
    if(ao & pots)
      if(ao & rhos)
	for(register basic_body*Bi=BB; Bi!=BB+N; ++Bi)
	  out<<Bi->mass()<<" "<<Bi->pos()<<" "<<Bi->vel()<<" "
	     <<nbdy::pot(Bi->SOUL)<<" "
	     <<nbdy::rho(Bi->SOUL)<<"\n";
      else
	for(register basic_body*Bi=BB; Bi!=BB+N; ++Bi)
	  out<<Bi->mass()<<" "<<Bi->pos()<<" "<<Bi->vel()<<" "
	     <<nbdy::pot(Bi->SOUL)<<"\n";
    else
      if(ao & rhos)
	for(register basic_body*Bi=BB; Bi!=BB+N; ++Bi)
	  out<<Bi->mass()<<" "<<Bi->pos()<<" "<<Bi->vel()<<" "
	     <<nbdy::rho(Bi->SOUL)<<"\n";
      else
	for(register basic_body*Bi=BB; Bi!=BB+N; ++Bi)
	  out<<Bi->mass()<<" "<<Bi->pos()<<" "<<Bi->vel()<<"\n";
  out.flush();
}
//------------------------------------------------------------------------------
void bodies::read_ascii(istream& in, const bool READ_EPS)
{
  if(READ_EPS)
    for(register basic_body*Bi=BB; Bi!=BB+N; ++Bi)
      if(in) {
	in>>Bi->mass()>>Bi->pos()>>Bi->vel()>>Bi->eps();
	util::SwallowRestofLine(in);
      } else return;
  else 
    for(register basic_body*Bi=BB; Bi!=BB+N; ++Bi)
      if(in) {
	in>>Bi->mass()>>Bi->pos()>>Bi->vel();
	util::SwallowRestofLine(in);
      } else return;
}
//------------------------------------------------------------------------------
void bodies::write_binary(ostream&out) const
{
  for(register basic_body*Bi=BB; Bi!=BB+N; ++Bi)
    out.write((char*)(&(Bi->IOD)),sizeof(iodata));
  out.flush();
}
//------------------------------------------------------------------------------
void bodies::read_binary(istream&in)
{
  for(register basic_body*Bi=BB; Bi!=BB+N; ++Bi)
    if(in) in.read((char*)(&(Bi->IOD)),sizeof(iodata));
    else return;
}
////////////////////////////////////////////////////////////////////////////////
