//-----------------------------------------------------------------------------+
//                                                                             |
// grav.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+

#include <nbdy/grav.h>
#include <nbdy/stat.h>
#include <nbdy/gren.h>
#include <iomanip>

////////////////////////////////////////////////////////////////////////////////
using namespace nbdy;

////////////////////////////////////////////////////////////////////////////////
// For individual softening:                                                    
// The little inline function eps_mutual determines the mutual softening        
// between two bodies. If softened cell interactions are enabled, then          
// there are also routines for eps_mutual between cells and cell and body.      
// Otherwise, there are routines for the minimum and maximum eps between        
// two cells and cell and body.                                                 

namespace nbdy {
#ifdef SOFT_MIN      // eps_mutual = min (eps_i, eps_j)
  inline real eps_mutual(soul*A, soul*B) { return min(eps(A),eps(B)); }
  inline real eps_mutual(cell*A, soul*B) { return min(emin(A),eps(B)); }
  inline real eps_mutual(cell*A, cell*B) { return min(emin(A),emin(B)); }
#else  // SOFT_MIN      eps_mutual = mass-weighted mean of eps_i and eps_j
  inline real eps_mutual(soul*A, soul*B)
  {
    if(eps(A)  == eps(B))  return eps(A);
    if(mass(A) == mass(B)) return half*(eps(A)+eps(B));
    return (mass(A)*eps(A)+mass(B)*eps(B))/(mass(A)+mass(B));
  }
  //----------------------------------------------------------------------------
  inline real eps_mutual(cell*A, soul*B)
  {
    if(eps(A)==eps(B)) return eps(A);
    else               return (mass(A)*eps(A)+mass(B)*eps(B))/(mass(A)+mass(B));
  }
  //----------------------------------------------------------------------------
  inline real eps_mutual(cell*A, cell*B)
  {
    if(eps(A)==eps(B)) return eps(A);
    else               return (mass(A)*eps(A)+mass(B)*eps(B))/(mass(A)+mass(B));
  }
  //----------------------------------------------------------------------------
#endif // SOFT_MIN
  inline real emin(const cell *a, const cell *b) { return min(emin(a),emin(b));}
  inline real emin(const cell *a, const soul *b) { return min(emin(a),eps(b)); }
  inline real emax(const cell *a, const cell *b) { return max(emax(a),emax(b));}
  inline real emax(const cell *a, const soul *b) { return max(emax(a),eps(b)); }
}
////////////////////////////////////////////////////////////////////////////////
// constants that determine the behaviour of the gravity routines               
////////////////////////////////////////////////////////////////////////////////
#if NDIM==3
const nbdy::uint N_PRE [3] = {  3, 0,64};            // N_PRE[1]  is not checked
const nbdy::uint N_POST[3] = {128,16, 0};            // N_POST[2] is not checked
#else
const nbdy::uint N_PRE [3] = { 3, 0,12};             // N_PRE[1]  is not checked
const nbdy::uint N_POST[3] = {64, 8, 0};             // N_POST[2] is not checked
#endif
////////////////////////////////////////////////////////////////////////////////
// class gravity_kernel                                                         
////////////////////////////////////////////////////////////////////////////////
//------------------------------------------------------------------------------
// body-many bodies interaction for global and individual softening             
//------------------------------------------------------------------------------
#define LOOP_B for(register soul*B=B0; B!=BN; B++)
#define LOOPMANY								\
if(A_is_sink)									\
  switch(B_sink) {								\
  case all:  LOOP_B { SET_RAD MAMB SET_AB } break;				\
  case some: LOOP_B { if(is_sink(B)) { SET_RAD MAMB SET_AB }			\
                      else           { SET_RAD MAMB SET_A  } } break;		\
  case none: LOOP_B { SET_RAD MAMB SET_A } break;				\
  }										\
else										\
  switch(B_sink) {								\
  case all:  LOOP_B { SET_RAD MAMB SET_B } break;				\
  case some: LOOP_B if(is_sink(B)) { SET_RAD MAMB SET_B } break;		\
  case none: break;								\
  }
//------------------------------------------------------------------------------
real gravity_kernel::greens_function(vect&R) const
{
  register real Rq=norm(R),x,D0,ME;
  BODYBODY_GLOB(1);
  return D0;
}
//------------------------------------------------------------------------------
// define the following macros, such that macro LOOPMANY above results in the   
// computation of body-body interactions with global softening length.          
#define SET_RAD   R  = cofm(A)-cofm(B); Rq = norm(R);
#define MAMB      MM = mass(A)*mass(B);
#define MMA       MM = mass(A);
#define MMB       MM = mass(B);
#define GRAV_A    A->pot()-= D0; A->acc()+= R;
#define GRAV_B    B->pot()-= D0; B->acc()-= R;
#define SET_AB    { BODYBODY_GLOB(MM) GRAV_A GRAV_B }
#define SET_A     { BODYBODY_GLOB(MM) GRAV_A }
#define SET_B     { BODYBODY_GLOB(MM) GRAV_B }

void gravity_kernel::many_global(const bool A_is_sink, soul*A,
				 const who B_sink, soul*B0, const soul*BN) const
  // This routine performs body-body interactions between body A and the        
  // bodies from B0 to BN. Using a fixed softening length.                      
{
  register real Rq,x,D0,MM,ME;
  register vect R;
  LOOPMANY
}
//------------------------------------------------------------------------------
#undef SET_AB
#undef SET_A
#undef SET_B
// define the following macros, such that macro LOOPMANY above results in the   
// computation of body-body interactions with individual softening lengths.     
#define SET_AB    { BODYBODY_INDI(MM) GRAV_A GRAV_B }
#define SET_A     { BODYBODY_INDI(MM) GRAV_A }
#define SET_B     { BODYBODY_INDI(MM) GRAV_B }
void gravity_kernel::many_individual(const bool A_is_sink, soul*A,
				     const who B_sink, soul*B0, const soul*BN)
  const
  // This routine performs body-body interactions between body A and the        
  // bodies from B0 to BN. Using individual softening lengths.                  
{
  register real Rq,x,D0,MM,ME;
  register vect R;
  LOOPMANY
}
#undef  LOOPMANY
//------------------------------------------------------------------------------
inline gravity_kernel::who gravity_kernel::who_is_sink(const cell* C) const
  // This routines figures out whether all, none, or some of the bodies in cell 
  // C are sinks.                                                               
{
  if(!is_sink(C)) return none;
  if( al_sink(C)) return all;
                  return some;
}
//------------------------------------------------------------------------------
inline void gravity_kernel::many_global(cell* A, soul* B) const
{
  many_global(is_sink(B),B,who_is_sink(A),
	      A->begin_soul_descendants(),A->end_soul_descendants());
}
inline void gravity_kernel::many_individual(cell* A, soul* B) const
{
  many_individual(is_sink(B),B,who_is_sink(A),
		  A->begin_soul_descendants(),A->end_soul_descendants());
}
//------------------------------------------------------------------------------
// replacing single cell-cell interaction with body-many body interactions      
//------------------------------------------------------------------------------
#define LoopSoulDesc(C,S) for(register cell::soul_descendant		\
			      S=C->begin_soul_descendants();		\
			      S!=C->end_soul_descendants(); ++S)
//------------------------------------------------------------------------------
inline void gravity_kernel::many_global(cell* A, cell* B) const
{
  if(al_sink(A))
    LoopSoulDesc(A,Ai)
      many_global(true,Ai,who_is_sink(B),
		  B->begin_soul_descendants(),B->end_soul_descendants());
  else if(is_sink(A))
    LoopSoulDesc(A,Ai)
      many_global(is_sink(Ai),Ai,who_is_sink(B),
		  B->begin_soul_descendants(),B->end_soul_descendants());
  else if(is_sink(B))
    LoopSoulDesc(A,Ai)
      many_global(false,Ai,who_is_sink(B),
		  B->begin_soul_descendants(),B->end_soul_descendants());
}
//------------------------------------------------------------------------------
inline void gravity_kernel::many_individual(cell* A, cell* B) const
{
  if(al_sink(A)) {
    LoopSoulDesc(A,Ai)
      many_individual(true,Ai,who_is_sink(B),
		      B->begin_soul_descendants(),B->end_soul_descendants());
  } else if(is_sink(A)) {
    LoopSoulDesc(A,Ai)
      many_individual(is_sink(Ai),Ai,who_is_sink(B),
		      B->begin_soul_descendants(),B->end_soul_descendants());
  } else if(is_sink(B)) {
    LoopSoulDesc(A,Ai) 
      many_individual(false,Ai,who_is_sink(B),
		      B->begin_soul_descendants(),B->end_soul_descendants());
  }
}
//------------------------------------------------------------------------------
// replacing single cell self-interaction with body-many body interactions      
//------------------------------------------------------------------------------
#undef  LoopSoulDesc
#define LoopSoulDesc(C,S) for(register cell::soul_descendant		\
			      S=C->begin_soul_descendants();		\
			      S!=C->last_soul_descendant (); ++S)
//------------------------------------------------------------------------------
inline void gravity_kernel::many_global(cell* A) const
{
  if(al_sink(A)) {
    LoopSoulDesc(A,Ai)
      many_global(true,Ai,all,Ai+1,A->end_soul_descendants());
  } else {
    LoopSoulDesc(A,Ai)
      many_global(is_sink(Ai),Ai,some,Ai+1,A->end_soul_descendants());
  }
}
//------------------------------------------------------------------------------
inline void gravity_kernel::many_individual(cell* A) const
{
  if(al_sink(A)) {
    LoopSoulDesc(A,Ai)
      many_individual(true,Ai,all,Ai+1,A->end_soul_descendants());
  } else {
    LoopSoulDesc(A,Ai)
      many_individual(is_sink(Ai),Ai,some,Ai+1,A->end_soul_descendants());
  }
}
//------------------------------------------------------------------------------
template<class NA, class NB>
inline real mutual_mass(const NA*A, const NB*B) { return mass(A)*mass(B); }    
//------------------------------------------------------------------------------
template<class NA, class NB>
inline bool interacting(const NA*A, const NB*B) { return is_sink(A); }
//------------------------------------------------------------------------------
inline void gravity_kernel::grav(vect &R, const real Rq, cell *A, soul *B) const
  // This routines performs a mutual cell-body interaction.                     
{
  GREENS_FUNCTION                                  // define Dn (see kern.h)    
  if(interacting(A,B)) {                           // IF(interaction B->A) >    
    c0(A)+= D0;                                    //   add to C0 of A          
    c1(A).add_mul(R, D1);                          //   add to C1 of A          
    c2(A).add_quadrup(R,D1,D2);                    //   add to C2 of A          
    c3(A).add_quadrup(R,D2,D3);                    //   add to C3 of A          
  }                                                // <                         
  if(interacting(B,A)) {                           // IF(interaction A->B) >    
    register vect RQ = R * quadru(A);		   //   R*Q_A                   
    register real tQ = trace(quadru(A)), RQR=R*RQ; //   tr(Q_A), R*Q*R          
    B->pot()-=     D0+half*(D1*tQ + D2*RQR);	   //   pot A->B                
    B->acc().sub_mul(R, D1+half*(D2*tQ + D3*RQR)); //   acc A->B (term in R)    
    B->acc().sub_mul(RQ,D2);                       //   acc A->B (term in R*Q_A)
  }                                                // <                         
}
//------------------------------------------------------------------------------
inline void gravity_kernel::grav(vect &R,const real Rq, cell* A,cell *B) const
  // This routines performs a mutual cell-cell interaction.                     
  // It's at the root of the whole code; most of the CPU time is spent in this  
  // routine and its (inlined) calls.                                           
{
  GREENS_FUNCTION                                  // define Dn (see kern.h)    
  register matr C2; C2.quadrup(R,D1,D2);           // pre-compute C2            
  register tens C3; C3.quadrup(R,D2,D3);           // pre-compute C3            
  register vect RQ;                                // to hold R*Q               
  register real tQ, RQR;                           // to hold tr(Q), R*Q*R      
  if(interacting(A,B)) {                           // IF(interaction B->A) >    
    RQ    = quadru(B)*R;                           //   R*Q_B                   
    tQ    = trace(quadru(B));                      //   tr(Q_B)                 
    RQR   = R*RQ;                                  //   R*Q_B*R                 
    c0(A)+= D0 +half*(D1*tQ + D2*RQR);             //   add to C0 of A          
    c1(A).add_mul(R, D1 +half*(D2*tQ + D3*RQR));   //   add to C1 of A (R)      
    c1(A).add_mul(RQ,D2);                          //   add to C1 of A (R*Q_B)  
    c2(A)+= C2;                                    //   add to C2 of A          
    c3(A)+= C3;                                    //   add to C3 of A          
  }                                                // <                         
  if(interacting(B,A)) {                           // IF(interaction A->B) >    
    R.negate();                                    //   set R to -R             
    RQ    = quadru(A)*R;                           //   R*Q_A                   
    tQ    = trace(quadru(A));                      //   tr(Q_A)                 
    RQR   = R*RQ;                                  //   R*Q_A*R                 
    c0(B)+= D0 +half*(D1*tQ + D2*RQR);             //   add to C0 of B          
    c1(B).add_mul(R, D1 +half*(D2*tQ + D3*RQR));   //   add to C1 of A (R)      
    c1(B).add_mul(RQ,D2);                          //   add to C1 of A (R*Q_B)  
    c2(B)+= C2;                                    //   add to C2 of B          
    c3(B)-= C3;                                    //   add to C3 of B          
  }                                                // <                         
}
//------------------------------------------------------------------------------
namespace nbdy {
  inline bool self_softened(const cell   *a,       // I:  sink = source         
			    const real    eps)     // I:  softening length      
    // This routine figures out if a cell-self interaction is fully-softened.   
  {
    return eps > (rmax(a)+rmax(a));
  }
  //----------------------------------------------------------------------------
  inline bool fully_softened(const cell   *a,      // I:  cell A                
			     const cell   *b,      // I:  cell B                
			     const real    Rq,     // I:  |x-y|^2               
			     const real    eps)    // I:  softening length      
    // This routine figures out if a cell-cell interaction is fully softened.   
  {
    register real x = eps - rmax(a) - rmax(b);
    return x>zero && Rq<(x*x);
  }
  //----------------------------------------------------------------------------
  inline bool fully_softened(const cell   *b,      // I:  cell B                
			     const real    Rq,     // I:  |x-y|^2               
			     const real    eps)    // I:  softening length      
    // This routine figures out if a cell-body interaction is fully softened.   
  {
    return eps>rmax(b) && Rq<square(eps-rmax(b));
  }
  //----------------------------------------------------------------------------
  inline bool fully_newtonian(const cell   *a,     // I:  cell A                
			      const cell   *b,     // I:  cell B                
			      const real    Rq,    // I:  |x-y|^2               
			      const real    eps)   // I:  softening length      
    // This routine figures out if a cell-cell interaction is fully Newtonian.  
  {
    return square(eps+rmax(a)+rmax(b)) < Rq;
  }
  //----------------------------------------------------------------------------
  inline bool fully_newtonian(const cell   *b,     // I:  cell B                
			      const real    Rq,    // I:  |x-y|^2               
			      const real    eps)   // I:  softening length      
    // This routine figures out if a cell-body interaction is fully Newtonian.  
  {
    return square(eps+rmax(b)) < Rq;
  }
  //----------------------------------------------------------------------------
  inline bool split_into_many(const cell   *a,     // I:  cell A                
			      const cell   *b,     // I:  cell B                
			      const uint    Ni)    // I:  min # body-body ias   
    // This routine figures out if a cell-cell interaction can be done directly.
  {
    return 4u < Ni && number(a)*number(b) < Ni;
  }
  //----------------------------------------------------------------------------
  inline bool split_into_many(const cell   *b,     // I:  cell                  
			      const uint    Ni)    // I:  min # body-body ias   
    // This routine figures out if a cell-self or a cell-body interaction shall 
    // be done directly.                                                        
  {
    return 2u < Ni && number(b) < Ni;
  }
  //----------------------------------------------------------------------------
  inline bool well_separated(const cell   *a,      // I:  cell A                
			     const cell   *b,      // I:  cell B                
			     const real    Rq)     // I:  |x-y|^2               
    // This routine finds out whether a cell-cell interaction is well-separated.
  {
    return Rq > square(rcrit(a)+rcrit(b));
  }
  //----------------------------------------------------------------------------
  inline bool well_separated(const cell   *b,      // I:  cell                  
			     const real    Rq)     // I:  |x-y|^2               
    // This routine finds out whether a cell-body interaction is well-separated.
  {
    return Rq > rcrit2(b);
  }
  //----------------------------------------------------------------------------
  inline bool are_twigs(const cell *a,             // I:   cell A               
			const cell *b)             // I:   cell B               
    // This routine figures out whether both cells are twig cells.              
  {
    return is_twig(a) && is_twig(b);
  }
  //----------------------------------------------------------------------------
  template<class NA, class NB>
  inline bool have_interaction(const NA*A, const NB*B)
    // This routines figures out whether any interaction is needed between two  
    // nodes A and B.                                                           
  { 
    return interacting(A,B) || interacting(B,A);
  }
}
//------------------------------------------------------------------------------
void gravity_kernel::interact(soul*A, soul*B) const
  // This routines performs an elementary body-body interaction.                
{
  if(!have_interaction(A,B)) return;               // IF no interaction > DONE <
  register vect R  = cofm(A)-cofm(B);              // compute R = x_A-x_B       
  register real Rq = norm(R),x,D0,ME;              // compute R^2               
  switch(SOFT) {                                   // what kind of softening? > 
  case individual:                                 // - individual softening >  
    BODYBODY_INDI(mutual_mass(A,B)) break;         //   compute (see kern.h) <  
  default:                                         // - global softening     >  
    BODYBODY_GLOB(mutual_mass(A,B)) break;         //   compute (see kern.h) <  
  }                                                // <                         
  if(interacting(A,B)) { A->pot()-=D0; A->acc()+=R; }// IF(B->A) > add to A <   
  if(interacting(B,A)) { B->pot()-=D0; B->acc()-=R; }// IF(A->B) > add to B <   
  STAT->record_BB();                               // record statistics         
}  
//==============================================================================
inline bool gravity_kernel::mutual_global(cell*A, soul*B) const
{
  if(split_into_many(A,N_PRE[0])) {                // IF(make many BBs) >       
    many_global(A,B);                              //   perform BB iactions     
    STAT->record_direct_CB();                      //   record statistics       
    return true;                                   //   DONE                    
  }                                                // <                         
  register vect dX = cofm(A)-cofm(B);              // compute R = x_A-x_B       
  register real Rq = norm(dX);                     // compute R^2               
  if(well_separated(A,Rq) &&                       // IF(well separated         
     (continuous_kernel()  ||                      // AND(continuous g(r)       
      fully_newtonian(A,Rq,EPS))) {                //     OR fully Newtonian)) >
    grav(dX,Rq,A,B);                               //   interact                
    STAT->record_taylor_CB();                      //   record statistics       
    return true;                                   //   DONE                    
  }                                                // <                         
  if(is_twig(A) ||                                 // IF(twig cell              
     split_into_many(A,N_POST[0]) ||               // OR make many BBs          
     (!continuous_kernel() &&                      // OR(non-continuous g(r)    
      fully_softened(A,Rq,EPS))) {                 //    AND fully softened)) > 
    many_global(A,B);                              //   perform BB iactions     
    STAT->record_direct_CB();                      //   record statistics       
    return true;                                   //   DONE                    
  }                                                // <                         
  return false;                                    // SPLIT                     
}
//------------------------------------------------------------------------------
inline bool gravity_kernel::mutual_individual(cell*A, soul*B) const
{
  if(split_into_many(A,N_PRE[0])) {                // IF(make many BBs) >       
    many_individual(A,B);                          //   perform BB iactions     
    STAT->record_direct_CB();                      //   record statistics       
    return true;                                   //   DONE                    
  }                                                // <                         
  register vect dX = cofm(A)-cofm(B);              // compute R = x_A-x_B       
  register real Rq = norm(dX);                     // compute R^2               
  if(well_separated(A,Rq) &&                       // IF(well separated         
     (continuous_kernel()  ||                      // AND (continuous g(r)      
      fully_newtonian(A,Rq,emax(A,B)))) {          //     OR fully Newtonian)) >
    if(continuous_kernel()) {                      //   IF(continuous g(r)) >   
      EPS = eps_mutual(A,B);                       //     set eps               
      EPQ = EPS*EPS;                               //     and eps^2             
    }                                              //   <                       
    grav(dX,Rq,A,B);                               //   interact                
    STAT->record_taylor_CB();                      //   record statistics       
    return true;                                   //   DONE                    
  }                                                // <                         
  if(is_twig(A) ||                                 // IF(twig cell              
     split_into_many(A,N_POST[0]) ||               // OR make many BBs          
     (!continuous_kernel() &&                      // OR(non-continuous g(r)    
      fully_softened(A,Rq,emin(A,B)))) {           //    AND fully softened)) > 
    many_individual(A,B);                          //   perform BB iactions     
    STAT->record_direct_CB();                      //   record stats            
    return true;                                   //   DONE                    
  }                                                // <                         
  return false;                                    // SPLIT                     
}
//------------------------------------------------------------------------------
bool gravity_kernel::interact(cell*A, soul*B) const
{
  if(!have_interaction(A,B)) return true;
  switch(SOFT) {
  case individual: return mutual_individual(A,B);
  default:               return mutual_global(A,B);
  }
}
//==============================================================================
inline bool gravity_kernel::mutual_global(cell*A, cell*B) const
{
  register vect dX = cofm(A)-cofm(B);              // compute dX = X_A - X_B    
  register real Rq = norm(dX);                     // and dX^2                  
  if(well_separated (A,B,Rq) &&                    // IF well separated         
     (continuous_kernel() ||                       // AND    continuous g(r)    
      fully_newtonian(A,B,Rq,EPS))) {              //     OR fully Newtonian  > 
    grav(dX,Rq,A,B);                               //    interact               
    STAT->record_taylor_CC();                      //    record stats           
    return true;                                   //    DONE                   
  }                                                // <                         
  if(are_twigs(A,B) ||                             // IF twig cells             
     split_into_many(A,B,N_POST[1]) ||             // OR make many BBs          
     (!continuous_kernel() &&                      // OR(non-continuous g(r)    
      fully_softened(A,B,Rq,EPS))) {               //    AND fully softened)) > 
    many_global(A,B);                              //    perform BB iactions    
    STAT->record_direct_CC();                      //    record stats           
    return true;                                   //    DONE                   
  }                                                // < ELSE  >                 
  return false;                                    //   SPLIT <                 
}
//------------------------------------------------------------------------------
inline bool gravity_kernel::mutual_individual(cell*A, cell*B) const
{
  register vect dX = cofm(A)-cofm(B);              // compute dX = X_A - X_B    
  register real Rq = norm(dX);                     // and dX^2                  
  if(well_separated (A,B,Rq) &&                    // IF  well separated        
     (continuous_kernel() ||                       // AND    continuous g(r)    
      fully_newtonian(A,B,Rq,emax(A,B)))) {        //     OR fully Newtonian >  
    if(continuous_kernel()) {                      //    IF continuous g(r)    >
      EPS = eps_mutual(A,B);                       //       set eps             
      EPQ = EPS*EPS;                               //       and eps^2           
    }                                              //    <                      
    grav(dX,Rq,A,B);                               //    interact hard          
    STAT->record_taylor_CC();                      //    record stats           
    return true;                                   //    DONE                   
  }                                                // <                         
  if(are_twigs(A,B) ||                             // IF twig cells             
     split_into_many(A,B,N_POST[1]) ||             // OR make many BBs          
     (!continuous_kernel() &&                      // OR(non-continuous g(r)    
      fully_softened(A,B,Rq,emin(A,B)))) {         //    AND fully softened)) > 
    many_individual(A,B);                          //    perform BB iactions    
    STAT->record_taylor_CC();                      //    record stats           
    return true;                                   //    DONE                   
  }                                                // < ELSE  >                 
  return false;                                    //   SPLIT <                 
}
//------------------------------------------------------------------------------
bool gravity_kernel::interact(cell*A, cell*B) const
{
  if(!have_interaction(A,B)) return true;
  switch(SOFT) {
  case individual: return mutual_individual(A,B);
  default:               return mutual_global(A,B);
  }
}
//==============================================================================
bool gravity_kernel::interact(cell*A) const
{
  if(!is_sink(A)) return true;
  switch(SOFT){
  case individual:
    if(is_twig(A) ||
       split_into_many(A,N_PRE[2]) ||
       !continuous_kernel() && self_softened(A,emin(A))) { 
      many_individual(A);
      STAT->record_direct_CS();
      return true;
    }
    return false;
  default:
    if(is_twig(A) ||
       split_into_many(A,N_PRE[2]) ||
       !continuous_kernel() && self_softened(A,EPS)) {
      many_global(A);
      STAT->record_direct_CS();
      return true;
    }
    return false;
  }
}
//------------------------------------------------------------------------------
void gravity_kernel::direct_summation(cell*A) const
{
  switch(SOFT){
  case individual: return many_individual(A);
  default:         return many_global(A);
  }
}
//------------------------------------------------------------------------------
