//-----------------------------------------------------------------------------+
//                                                                             |
// nbdy.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// written by Walter Dehnen, 2000/2001                                         |
// e-mail:   dehnen@mpia-hd.mpg.de                                             |
// address:  Max-Planck Institut fr Astronomie, Knigstuhl 69117 Heidelberg   |
//           Germany                                                           |
//                                                                             |
//-----------------------------------------------------------------------------+

#include <nbdy.h>
extern "C" {
#include <fstream>
#include <iomanip>
#include <time.h>
#include <math.h>
}

const double Pi   = 3.14159265358979323846264338328;

using std::ostream;
using std::endl;
using std::ios;
using std::setw;
using std::setprecision;
////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
  inline void record_cpu(clock_t& c0, real& CPU) {
    register clock_t c1 = clock();
    CPU += (c1-c0)/real(CLOCKS_PER_SEC);
    c0 = c1;
  }
}
using namespace nbdy;
#define LoopMyBodies    LoopBodies(bodies,BODIES,Bi)
////////////////////////////////////////////////////////////////////////////////
// class fasstree::basic_nbody                                                  
////////////////////////////////////////////////////////////////////////////////
inline void basic_nbody::reset_cpus() const
{
  CPU_BUILD   = zero;
  CPU_GRAV    = zero;
  CPU_PEX     = zero;
  CPU_PREDICT = zero;
  CPU_DENS    = zero;
  CPU_EPSI    = zero;
  CPU_STEP    = zero;
}
//------------------------------------------------------------------------------
inline void basic_nbody::set_tree(const vect *x, const how_to_make_tree how)
{
  register clock_t cpu = clock();                  // get current CPU time      
  if(how == grow ||                                // IF must falcON::grow()    
     (how == free && N_REUSED >= NREUSE) ) {       // OR may and need to     >  
    falcON::grow(NCRIT,x);                         //   grow() new tree         
    N_REUSED=0u;                                   //   reset reuse() counter   
  } else {                                         // < ELSE >               >  
    falcON::reuse();                               //   reuse() old tree        
    ++N_REUSED;                                    //   count falcON::reuse()s  
  }                                                // <                         
  record_cpu(cpu,CPU_BUILD);                       // record CPU time required  
}
//------------------------------------------------------------------------------
void basic_nbody::set_forces()
{
  register clock_t cpu = clock();
  falcON::approximate_gravity();                   //  compute sink forces      
  record_cpu(cpu,CPU_GRAV);                        //  record CPU               
  if(PEX) {                                        //  if external potential    
    register vect A;
    LoopMyBodies if(is_sink(Bi)) {                 //    loop body sinks        
      Bi.pot()+=PEX->pot_f(A,pos(Bi),time());      //      add ext potential    
      Bi.acc()+=A;                                 //      add ext force        
    }
    record_cpu(cpu,CPU_PEX);                       //  record CPU               
  }
}
//------------------------------------------------------------------------------
void basic_nbody::reset_softening(const soft_type sf,
				  const uint ns, const real e,
				  const kern_type ker)
{
  SOFTENING = sf;
  NSOFT     = ns;
  switch(SOFTENING) {
  case individual_adaptive:
  case individual_fixed:
    EMAX = e;
    falcON::reset_softening(-one,ker);
    break;
  default:
    EMAX = zero;
    falcON::reset_softening(e,ker);
  }
}
//------------------------------------------------------------------------------
inline void basic_nbody::setup_eps(body& Bi) const
{
  const double Pi = 3.14159265358979323846264338328;
#if NDIM==3
  const real fac = 0.75*NSOFT/Pi;
  Bi.eps() = min(EMAX,POW(fac/rho(Bi),third));
#else
  const real fac = NSOFT/Pi;
  Bi.eps() = min(EMAX,SQRT(fac/rho(Bi)));
#endif	
}
//------------------------------------------------------------------------------
void basic_nbody::set_softening_lengths()
{
  register clock_t  cpu = clock();                 // cpu time                  
  LoopMyBodies Bi.flag_as_sink();                  // flag every body           
  estimate_n(NSOFT);                               // estimate n_i              
  LoopMyBodies setup_eps(Bi);                      // set eps_i                 
  record_cpu(cpu,CPU_EPSI);                        // record CPU                
}
//------------------------------------------------------------------------------
basic_nbody::basic_nbody(const bodies   *b,        // I: bodies                 
			 const uint      ns,       // I: N_soft                 
			 const uint      nr,       // I: N_ref                  
			 const real      e,        // I: eps/eps_max            
			 const real      ti,       // I: t_initial              
			 const soft_type sf,       //[I: softening type]        
			 const real      th,       //[I: tolerance parameter]   
			 const uint      nu,       //[I: N_reuse]               
			 const uint      nc,       //[I: N_crit]                
			 const kern_type ker,      //[I: softening kernel]      
			 const potential*p) :      //[I: P_ex]                  
  falcON(b, (ns==0 && e>=zero)? e:-abs(e), th, ker),
  BODIES(b), PEX(p), TINI(ti), 
  NSOFT(ns), NREF(nr), NREUSE(nu), N_REUSED(0), NCRIT(max(1u,nc)),
  EMAX(e), DIAG(false)
{
  reset_softening(sf,ns,e,ker);
}
//------------------------------------------------------------------------------
inline void basic_nbody::estimate_mass_density()
{
  register clock_t  cpu = clock();                 // cpu time                  
  falcON::estimate_rho(NREF);                      // estimate rho in tree      
  if(PEX)                                          // if external potential     
    LoopMyBodies if(is_sink(Bi))                   //   loop body sinks         
      Bi.rho()+=PEX->rho(pos(Bi),time());          //     add ext density       
  record_cpu(cpu,CPU_DENS);                        // record CPU                
}
//------------------------------------------------------------------------------
void basic_nbody::estimate_mass_densities()
{
  LoopMyBodies Bi.flag_as_sink();
  estimate_mass_density();
}
//------------------------------------------------------------------------------
void basic_nbody::estimate_surf_densities()
{
  LoopMyBodies Bi.flag_as_sink();
  falcON::estimate_sd(NREF);
}
//------------------------------------------------------------------------------
inline real basic_nbody::trace(const tensor&T) const
{
  register real tr=zero;
  for(register int i=0; i<NDIM; i++) tr += T[i][i];
  return tr;
}
//------------------------------------------------------------------------------
#define LoopTensor for(i=0; i<NDIM; i++) for(j=0; j<NDIM; j++)
//------------------------------------------------------------------------------
void basic_nbody::diagnose() const
{
  register indx   i,j;
  register real   u=zero, m=zero;
  register vect   x=zero, v=zero;
  register amom   l=zero;
  register tensor t,w;
  LoopTensor t[i][j] = w[i][j] = zero;
  if(PEX) PEX->energies(u,w,t,x,v,l,m);
  LoopMyBodies {
    m += mass(Bi);                                 // total mass                
    u += mass(Bi) * pot(Bi);                       // potential energy          
    LoopTensor {
      t[i][j] += mass(Bi) * vel(Bi)[i]*vel(Bi)[j]; // kin energy tensor         
      w[i][j] += mass(Bi) * pos(Bi)[i]*acc(Bi)[j]; // pot energy tensor         
    }
    x += mass(Bi) * pos(Bi);                       // dipole                    
    v += mass(Bi) * vel(Bi);                       // momentum                  
    l += mass(Bi) * pos(Bi) ^ vel(Bi);             // total angular momentum    
  }
  M    = m;                                        // total mass                
  Ktot = half*trace(t);                            // total kin energy          
  Utot = half*u;                                   // total pot energy          
  TU   =-half*trace(t)/trace(w);                   // virial ratio              
  L    = l;                                        // total angular momentum    
  CMX  = x/m;                                      // center of mass            
  CMV  = v/m;                                      // center of mass velocity   
  LoopTensor {
    KT[i][j] = half * t[i][j];                     // kin energy tensor         
    WT[i][j] =        w[i][j];                     // pos energy tensor         
  }
}
#undef LoopTensor
//------------------------------------------------------------------------------
void basic_nbody::update_diagnostics() const
{
  if(DIAG) return;
  diagnose();
  DIAG = true;                                     // diagnose is up-to-date   
}
//------------------------------------------------------------------------------
const real& basic_nbody::kin_energy    () const
{
  update_diagnostics();
  return Ktot;
}
//------------------------------------------------------------------------------
const real& basic_nbody::pot_energy    () const
{
  update_diagnostics();
  return Utot;
}
//------------------------------------------------------------------------------
const real  basic_nbody::total_energy    () const
{
  update_diagnostics();
  return Ktot+Utot;
}
//------------------------------------------------------------------------------
const real& basic_nbody::virial_ratio   () const
{
  update_diagnostics();
  return TU;
}
//------------------------------------------------------------------------------
const amom& basic_nbody::total_angmom    () const
{
  update_diagnostics();
  return L;
}
//------------------------------------------------------------------------------
const vect& basic_nbody::total_momentum  () const
{
  update_diagnostics();
  return CMV;
}
//------------------------------------------------------------------------------
const vect& basic_nbody::center_of_mass  () const
{
  update_diagnostics();
  return CMX;
}
//------------------------------------------------------------------------------
const real& basic_nbody::kin_energy      (const indx i, const indx j) const
{
  update_diagnostics();
  return KT[i][j];
}
//------------------------------------------------------------------------------
const real& basic_nbody::pot_energy      (const indx i, const indx j) const 
{
  update_diagnostics();
  return WT[i][j];
}
//------------------------------------------------------------------------------
const real  basic_nbody::as_angmom    (const indx i, const indx j) const 
{
  update_diagnostics();
#if NDIM==3
  switch(i) {
  case 0:
    switch(j) {
    case 0:  return  zero;
    case 1:  return  L[2];
    default: return -L[1];
    }  
  case 1:
    switch(j) {
    case 0:  return -L[2];
    case 1:  return  zero;
    default: return  L[0];
    }  
  default:
    switch(j) {
    case 0:  return  L[1];
    case 1:  return -L[0];
    default: return  zero;
    }  
  }
#else
  return (i==j) ? zero : real(L);
#endif
}
////////////////////////////////////////////////////////////////////////////////
// class fasstree::LeapFrogCode                                                 
////////////////////////////////////////////////////////////////////////////////
LeapFrogCode::LeapFrogCode(const bodies    *b,     // I: bodies                 
			   const uint       ns,    // I: N_soft                 
			   const real       e,     // I: eps/eps_max            
			   const real       ti,    // I: initial time           
			   const int        h0,    // I: h0                     
			   const uint       nu,    //[I: N_reuse]               
			   const uint       nr,    //[I: N_ref]                 
			   const vect     * pc,    //[I: center of root]        
			   const soft_type  sf,    //[I: softening type]        
			   const real       th,    //[I: tolerance parameter]   
			   const uint       nc,    //[I: N_crit]                
			   const kern_type  ker,   //[I: softening kernel]      
			   const potential* pp) :  //[I: P_ex]                  
  basic_nbody(b,ns,nr,e,ti,sf,th,nu,nc,ker,pp),
  LeapFrog   (h0,ti)
{
  register clock_t  cpu = clock();                 // cpu time                  
  reset_cpus();                                    // reset CPU time counters   
  if(SOFTENING==individual_adaptive)               // if individual adaptive:   
    set_softening_lengths();                       //   then set eps_i          
  LoopMyBodies Bi->flag_as_sink();                 // flag everybody as sink    
  set_tree  (pc);                                  // create tree               
  set_forces();                                    // compute forces            
  record_cpu(cpu,CPU_STEP);                        // record CPU time           
  CPU_TOTAL = CPU_STEP;                            // record accumulated time   
}
//------------------------------------------------------------------------------
void LeapFrogCode::full_step(const vect*pc)
{
  register clock_t  cpu = clock();                 // cpu time                  
  reset_cpus();                                    // reset CPU time counters   
  predict(BODIES);                                 // move = drift              
  set_tree  (pc);                                  // create tree               
  set_forces();                                    // approximate forces        
  accelerate(BODIES);                              // accelerate = kick         
  DIAG = false;                                    // diagnose out-of-date      
  record_cpu(cpu,CPU_STEP);                        // record CPU time           
  CPU_TOTAL += CPU_STEP;                           // record accumulated time   
}
//------------------------------------------------------------------------------
void LeapFrogCode::long_stats(ostream& to) const
{
  to<<" cpu times [sec]: "
    <<"predict build tree/ext grav  eps_i  total   accum\n                  "
    <<setw(6)<<setprecision(4)<<cpu_predict()    <<" "
    <<setw(6)<<setprecision(4)<<cpu_build()      <<" "
    <<setw(6)<<setprecision(4)<<cpu_tree_force() <<" "
    <<setw(6)<<setprecision(4)<<cpu_ext_force()  <<" "
    <<setw(6)<<setprecision(4)<<cpu_softening()  <<" "
    <<setw(6)<<setprecision(4)<<cpu_longstep()   <<" "
    <<setw(8)<<setprecision(6)<<cpu_total()      <<"\n";
  if(time() != TINI) update_diagnostics();
  to<<" E="<<setw(8)<<setprecision(5)<<total_energy()
    <<"; -T/U="<<setw(7)<<setprecision(5)<<virial_ratio()
    <<"; L ="  
#if NDIM==3
    <<setw(7)<<setprecision(5)<<total_angmom()[0]<<" "
    <<setw(7)<<setprecision(5)<<total_angmom()[1]<<" "
    <<setw(7)<<setprecision(5)<<total_angmom()[2]
#else
    <<setw(7)<<setprecision(5)<<total_angmom()
#endif
    <<"; Vcm= "<<setw(7)<<setprecision(4)<<total_momentum()[0]<<" "
    <<setw(7)<<setprecision(4)<<total_momentum()[1]
#if NDIM==3
    <<" "<<setw(7)<<setprecision(4)<<total_momentum()[2]
#endif
    <<endl;
}
//------------------------------------------------------------------------------
void LeapFrogCode::short_stats(ostream& to) const
{
  to.setf(ios::left, ios::adjustfield);
  if(time() != TINI) update_diagnostics();
  to<<setprecision(5)<<setw(10)<<time()         <<" "
    <<setprecision(7)<<setw(11)<<total_energy() <<" "
    <<setprecision(min(5,max(1,5-int(-log10(virial_ratio())))))
    <<setw(7)<<virial_ratio()<<" ";
#if NDIM==3
  to<<setprecision(5)<<setw(10)<<SQRT(norm(total_angmom()))  <<" ";
#else
  to<<setprecision(5)<<setw(10)<<total_angmom()              <<" ";
#endif
  to<<setprecision(2)<<setw(8)<<SQRT(norm(total_momentum()))  <<" ";
  to.setf(ios::right, ios::adjustfield);
  to<<setprecision(4)<<setw(6)<<cpu_build()    <<" "
    <<setprecision(4)<<setw(7)<<cpu_force()    <<" "
    <<setprecision(4)<<setw(7)<<cpu_longstep() <<" "
    <<setprecision(6)<<setw(9)<<cpu_total()    <<endl;
}
//------------------------------------------------------------------------------
void LeapFrogCode::short_stat_head(ostream& to) const
{
  to<<"    time   "
    <<"  energy    "
    <<"  -T/U  "
#if NDIM==3
    <<"   |L|     "
#else
    <<"   L_z     "
#endif
    <<"  |v_cm| "
    <<" build   force    step     accum"<<endl;
}
//------------------------------------------------------------------------------
void LeapFrogCode::short_stat_line(ostream &to) const
{
  to<<" --------------------------------------------------"
    "--------------------------------"<<endl;
}
////////////////////////////////////////////////////////////////////////////////
// class fasstree::BlockStepCode                                                
////////////////////////////////////////////////////////////////////////////////
void BlockStepCode::reset_stepping(const real fac)
{
  MYSCHEME = (fac > 0)? rhos : dacc;
  FAC = fac;
  BlockStep::reset_fac(fac);
}
//------------------------------------------------------------------------------
inline void BlockStepCode::set_pposs()
{
  register clock_t cpu = clock();
  set_predict();
  LoopMyBodies predict_pos(Bi);
  record_cpu(cpu,CPU_PREDICT);
}
//==============================================================================
// routines to use with dacc scheme                                             
//------------------------------------------------------------------------------
inline bool BlockStepCode::step_too_long(const body&B) const
{
  return nbdy::may_go_shorter(B) && DaccScheme::step_too_long(this,B);
}
//------------------------------------------------------------------------------
inline bool BlockStepCode::may_go_longer(const body&B) const
{
  return  level(B)>0 && nbdy::may_go_longer(B);
}
//------------------------------------------------------------------------------
inline BlockStep::too_short BlockStepCode::step_too_short(const body&B) const
{
  if(may_go_longer(B)) return DaccScheme::step_too_short(this,B);
  else                 return allow_short;
}
//------------------------------------------------------------------------------
inline void BlockStepCode::set_sink_flags(const indx low)
{
  LoopMyBodies if(is_in_tree(Bi))                  // loop bodies in tree       
    if(low == 0 || level(Bi) >= low                //   if level >= low         
       || level(Bi)+1==low && may_go_shorter(Bi))  //      or may go shorter    
         Bi.flag_as_sink();                        //     flag as sink          
    else Bi.unflag_sink ();                        //   else    : no sink       
}
//------------------------------------------------------------------------------
void BlockStepCode::
smallest_step_dacc(const indx low,                 // I: lowest level moving    
		   const vect*x,                   // [I]: center of root       
		   const how_to_make_tree make)    // [I]: how to make tree     
{
  if(low && !No_shorter_level(low-1)) {            // if nobody moves:          
    clock_on();                                    //  increase actual clock    
    move_times(low);                               //  increase steps' clocks   
    return;                                        //  and we are DONE          
  }
  clock_on ();                                     // increase actual clock     
  set_pposs();                                     // predict positions         
  set_tree (x,make);                               // initialize tree           
  set_sink_flags(low);                             // set flags                 
  set_forces();                                    // set accelerations         
  LoopMyBodies                                     // loop bodies               
    if(level(Bi) > low)                            // l>low                     
      switch(step_too_short(Bi)) {
      case go_longer:    move_proper(Bi); goto_longer(Bi);      break;
      case forbid_short: move_proper(Bi); Bi.forbid_shorter(); break;
      default:           move_proper(Bi); Bi.allow_shorter();  break;
      }
    else if(level(Bi) == low)                      // l=low                     
      switch(step_too_short(Bi)) {
      case go_longer:                     goto_longer(Bi);      break;
      case forbid_short: move_proper(Bi); Bi.forbid_shorter(); break;
      default:           move_proper(Bi); Bi.allow_shorter();  break;
      }
    else if(level(Bi)+1==low && step_too_long(Bi)) {  // l=low-1 & step too long
                         move_proper(Bi); goto_shorter(Bi);
    }
  move_times(low);                                 // increase steps' clocks    
}
//==============================================================================
// routines used with the rhos scheme                                           
//------------------------------------------------------------------------------
inline bool BlockStepCode::test_for_longer(const indx low, const body&B) const
{
  return low>0 && may_go_longer(B) && level(B) == low;
}
//------------------------------------------------------------------------------
inline bool BlockStepCode::test_for_shorter(const indx low, const body&B) const
{
  return level(B)+1 == low;
}
//------------------------------------------------------------------------------
void BlockStepCode::
smallest_step_rhos(const indx low,                 // I: lowest level moving    
		   const vect*x,                   // [I]: center of root       
		   const how_to_make_tree make)    // [I]: must grow() a tree   
{
  if(low && !No_shorter_level(low-1)) {            // if nobody moves:          
    clock_on();                                    //   increase actual clock   
    move_times(low);                               //   increase steps clocks   
    return;                                        //   and we are DONE         
  }
  clock_on ();                                     // increase actual clock     
  set_pposs();                                     // predict positions         
  set_tree (x,make);                               // initialize tree           
  if(SOFTENING==individual_adaptive)               // if individual adaptive:   
    set_softening_lengths();                       //   then set eps_i          
  if(low>0) {                                      // low>0: check time steps   
    LoopMyBodies                                   //   loop bodies in tree     
      if(test_for_longer (low,Bi) ||               //     if may go slower      
	 test_for_shorter(low,Bi) )                //     OR faster             
	   Bi.flag_as_sink();                      //       need guessed rho    
      else Bi.unflag_sink ();                      //     else unflag           
    estimate_mass_density();                       //   guess rho               
    LoopMyBodies {                                 //   loop bodies in tree     
      if     (test_for_longer (low,Bi) &&          //     if may go slower      
	      RhoScheme::step_too_short(this,Bi) ) //       AND too fast        
	goto_longer(Bi);                           //       change to l-1       
      else if(test_for_shorter(low,Bi) &&          //     else if may go fast   
	      RhoScheme::step_too_long(this,Bi) )  //       AND too slow        
	goto_shorter(Bi);                          //       change to l+1       
      if(level(Bi) >= low) Bi.flag_as_sink();      //     if l>=low: need acc   
      else                 Bi.unflag_sink ();      //     else     : unflag     
    }
  } else                                           // low=0: move all           
    LoopMyBodies Bi.flag_as_sink();                //   loop bodies in tree     
  if(No_shorter_level(low)) {                      // if anybody moves:         
    set_forces();                                  //   get accelerations       
    LoopMyBodies if(is_in_tree(Bi)){               //   loop bodies in tree     
      if(level(Bi) >= low) {                       //     forall l >= low:      
	move_proper(Bi);                           //       move                
	Bi.allow_longer();                         //       allow change l-1    
      }
    }
  }
  move_times(low);                                 // increase steps' clocks    
}
//==============================================================================
// general routines                                                             
//==============================================================================
void BlockStepCode::full_step(const vect* x)
{
  register clock_t  cpu = clock();                 // cpu time                  
  const    unsigned nn  = 1 << shortest_step();    // number of small steps     
  register unsigned tn;                            // index of small step       
  reset_cpus();                                    // reset CPU time counters   
  switch(MYSCHEME) {
  case rhos:
    if(nn>1)                                       // IF multiple steps   >     
      for(tn=1; tn<=nn; tn++)                      //   loop over all           
	smallest_step_rhos(longest_moving(tn),x,grow);//  smallest steps        
    else                                           // < ELSE      >             
      smallest_step_rhos(0,x);                     //   one step  <             
    break;
  default:
    for(tn=1; tn<nn; tn++)                         // loop over all but last    
      smallest_step_dacc(longest_moving(tn),x);    //   smallest steps          
    smallest_step_dacc(0,x,grow);                  // last step: grow tree      
  }
  DIAG = false;                                    // diagnose out-of-date      
  record_cpu(cpu,CPU_STEP);                        // record CPU time           
  CPU_TOTAL += CPU_STEP;                           // record accumulated time   
}
//------------------------------------------------------------------------------
void BlockStepCode::prepare(const int  h0,         // I:   h0                   
			    const indx nt,         // I:   N_stps               
			    const vect*x)          //[I: center of root]        
{
  if(!BODIES->support_timesteps())                 // IF timesteps not sup      
    BODIES->enable_time_steps();                   //   add timestep support    
  register clock_t cpu = clock();                  // local CPU counter         
  reset_cpus();                                    // reset CPU counters        
  BlockStep::reset_steps(h0,nt,TINI);              // set time steps            
  LoopMyBodies Bi.opos()=pos(Bi);                  // set x=true position       
  set_tree(x,grow);                                // initialize tree           
  if(SOFTENING != global_fixed)                    // if individual softening   
    set_softening_lengths();                       //   set eps_i               
  LoopMyBodies Bi.flag_as_sink();                  // flag every body as sink   
  set_forces();                                    // get accelerations         
  LoopMyBodies Bi.oacc()=acc(Bi);                  // store accelerations       
  diagnose();                                      // energy et al. at t=t0     
  DIAG = true;
  BlockStep::reset_fac(FAC,TQ);                    // reset factors             
  if(nt==1)                                        // only one time-step        
    LoopMyBodies add(Bi.level()=0);                //   all bodies in level 0   
  else {                                           // else: assign levels       
    BlockStep::reset_counts();                     //   reset counts            
    switch(MYSCHEME) {                             //   which scheme ?          
    case rhos:                                     //   rho scheme:             
      estimate_mass_density();                     //     estim mass density    
      LoopMyBodies {                               //     loop bodies           
	RhoScheme::assign_to_level(this,Bi);       //       assign proper lev   
	Bi.allow_longer();                         //       allow longer step   
      }
      break;
    default:                                       //   delta(acc) scheme:      
      clock_on();                                  //     adv t by tau_min      
      set_pposs();                                 //     predict positions     
      clock_back();                                //     set t back            
      set_tree(x,reuse);                           //     initialize tree       
      set_forces();                                //     get accelerations     
      LoopMyBodies {                               //     loop bodies           
	DaccScheme::assign_to_level(this,Bi);      //       assign proper lev   
	Bi.allow_longer();                         //       allow longer step   
      }
    }
  }
  record_cpu(cpu,CPU_STEP);                        // record total CPU          
  CPU_TOTAL = CPU_STEP;                            // record accumulated time   
}
//------------------------------------------------------------------------------
BlockStepCode::BlockStepCode(const bodies    *b,   // I: bodies                 
			     const uint       ns,  // I: N_soft                 
			     const real       e,   // I: eps/eps_max            
			     const real       ti,  // I: initial time           
			     const int        h0,  // I: h0                     
			     const indx       nl,  // I: N_steps                
			     const uint       nr,  // I: N_ref                  
			     const real       fac, // I: factor for stepping    
			     const vect     * pc,  //[I: center of root]        
			     const real       tq,  //[I: tol^2]                 
			     const soft_type  sf,  //[I: softening type]        
			     const real       th,  //[I: tolerance parameter]   
			     const uint       nc,  //[I: N_crit]                
			     const kern_type  ker, //[I: softening kernel]      
			     const potential* pp) ://[I: P_ex]                  
  basic_nbody(b,ns,nr,e,ti,sf,th,0,nc,ker,pp),
  BlockStep  (),
  MYSCHEME   ((fac>zero)? rhos : dacc),
  FAC        (abs(fac)),
  TQ         (tq)
{
  BlockStep::reset_fac(FAC);
  if (FAC==zero) error("time step-controlling factor = 0");
  prepare(h0,nl,pc);
}
//------------------------------------------------------------------------------
void BlockStepCode::long_stats(ostream& to) const
{
  to<<" cpu times [sec]: "
    <<"predict build tree/ext grav  eps_i  total   accum\n                  "
    <<setw(6)<<setprecision(4)<<cpu_predict()    <<" "
    <<setw(6)<<setprecision(4)<<cpu_build()      <<" "
    <<setw(6)<<setprecision(4)<<cpu_tree_force() <<" "
    <<setw(6)<<setprecision(4)<<cpu_ext_force()  <<" "
    <<setw(6)<<setprecision(4)<<cpu_softening()  <<" "
    <<setw(6)<<setprecision(4)<<cpu_longstep()   <<" "
    <<setw(8)<<setprecision(6)<<cpu_total()      <<"\n";
  BlockStep::short_stats(to); to<<"\n";
  if(time() != TINI) update_diagnostics();
  to<<" E="<<setw(8)<<setprecision(5)<<total_energy()
    <<"; -T/U="<<setw(7)<<setprecision(5)<<virial_ratio()
    <<"; L ="  
#if NDIM==3
    <<setw(7)<<setprecision(5)<<total_angmom()[0]<<" "
    <<setw(7)<<setprecision(5)<<total_angmom()[1]<<" "
    <<setw(7)<<setprecision(5)<<total_angmom()[2]
#else
    <<setw(7)<<setprecision(5)<<total_angmom()
#endif
    <<"; Vcm= "<<setw(7)<<setprecision(4)<<total_momentum()[0]<<" "
    <<setw(7)<<setprecision(4)<<total_momentum()[1]
#if NDIM==3
    <<" "<<setw(7)<<setprecision(4)<<total_momentum()[2]
#endif
    <<endl;
}
//------------------------------------------------------------------------------
void BlockStepCode::short_stats(ostream& to) const
{
  to.setf(ios::left, ios::adjustfield);
  if(time() != TINI) update_diagnostics();
  to<<setprecision(5)<<setw(10)<<time()         <<" "
    <<setprecision(7)<<setw(11)<<total_energy() <<" "
    <<setprecision(min(5,max(1,5-int(-log10(virial_ratio())))))
    <<setw(7)<<virial_ratio()<<" ";
#if NDIM==3
  to<<setprecision(5)<<setw(10)<<SQRT(norm(total_angmom()))  <<" ";
#else
  to<<setprecision(5)<<setw(10)<<total_angmom()              <<" ";
#endif
  to<<setprecision(2)<<setw(8)<<SQRT(norm(total_momentum()))  <<" ";
  to.setf(ios::right, ios::adjustfield);
  if(shortest_step()) BlockStep::short_stats(to);
  if(MYSCHEME == rhos)
    to<<setprecision(4)<<setw(6)<<cpu_build()    <<" "
      <<setprecision(4)<<setw(6)<<cpu_density()  <<" "
      <<setprecision(4)<<setw(7)<<cpu_force()    <<" "
      <<setprecision(4)<<setw(7)<<cpu_longstep() <<" "
      <<setprecision(6)<<setw(9)<<cpu_total()    <<endl;
  else
    to<<setprecision(4)<<setw(6)<<cpu_build()    <<" "
      <<setprecision(4)<<setw(7)<<cpu_force()    <<" "
      <<setprecision(4)<<setw(7)<<cpu_longstep() <<" "
      <<setprecision(6)<<setw(9)<<cpu_total()    <<endl;
}
//------------------------------------------------------------------------------
void BlockStepCode::short_stat_head(ostream& to) const
{
  to<<"    time   "
    <<"  energy    "
    <<"  -T/U  "
#if NDIM==3
    <<"   |L|     "
#else
    <<"   L_z     "
#endif
    <<"  |v_cm| ";
  if(shortest_step())
    for(register int i=0, h=-h0(); i<=shortest_step(); i++, h--)
           if(h>13)  to<<"  #2^" <<     h  <<" ";
      else if(h> 9)  to<<"  #"   << (1<<h) <<" ";
      else if(h> 6)  to<<"   #"  << (1<<h) <<" ";
      else if(h> 3)  to<<"    #" << (1<<h) <<" ";
      else if(h>=0)  to<<"     #"<< (1<<h) <<" ";
      else if(h==-1) to<<"   #1/2 ";
      else if(h==-2) to<<"   #1/4 ";
      else if(h==-3) to<<"   #1/8 ";
      else if(h==-4) to<<"  #1/16 ";
      else if(h==-5) to<<"  #1/32 ";
      else if(h==-6) to<<"  #1/64 ";
      else if(h>-10) to<<"  #2^" <<     h  <<" ";
      else           to<<" #2^"  <<     h  <<" ";
  if(MYSCHEME == rhos)
    to<<" build    rho   force    step     accum"<<endl;
  else
    to<<" build   force    step     accum"<<endl;
}
//------------------------------------------------------------------------------
void BlockStepCode::short_stat_line(ostream &to) const
{
  to<<" --------------------------------------------------";
  if(shortest_step())
    for(register int i=0; i<=shortest_step(); i++) to<<"--------";
  if(MYSCHEME == rhos &&  !is_leap_frog() )
    to<<"---------------------------------------"<<endl;
  else
    to<<"--------------------------------"<<endl;
}
