//-----------------------------------------------------------------------------+
//                                                                             |
// nmio.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2001                                               |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef    ALLOW_NEMO
#  define  ALLOW_NEMO
#endif

#include <nbdy/nmio.h>
#include <iostream>

extern "C" {
#include <stdinc.h>
#include <filestruct.h>
#include <history.h>
#include <snapshot/snapshot.h>
}

#ifdef TWODIMENSIONAL
#  define NDM 2
#else
#  define NDM 3
#endif

#define mystream static_cast<stream>(STREAM)

using std::cerr;
using namespace nbdy;

//------------------------------------------------------------------------------
inline char* tag(const nemo_io::Set X)
{
  switch(X) {
  case nemo_io::snap:   return SnapShotTag;
  case nemo_io::param:  return ParametersTag;
  case nemo_io::bodies: return ParticlesTag;
  case nemo_io::diags:  return DiagnosticsTag;
  default: cerr<<"Unknown nemo_io::Set\n"; exit(1);
  }
}
//------------------------------------------------------------------------------
inline int tag(const nemo_io::CoSys X)
{
  switch(X) {
  case nemo_io::cart:  return Cartesian;
  case nemo_io::spher: return Spherical;
  case nemo_io::scat:  return Scattering;
  default: cerr<<"FATAL ERROR: Unknown nemo_io::CoSys\n"; exit(1);
  }
}
//------------------------------------------------------------------------------
inline char* tag(const nemo_io::SingleScalar X)
{
  switch(X) {
  case nemo_io::time:    return TimeTag;
  case nemo_io::cputime: return "cputime";
  default: cerr<<"FATAL ERROR: Unknown nemo_io::SingleScalar\n"; exit(1);
  }
}
//------------------------------------------------------------------------------
inline char* tag(const nemo_io::SingleVector X)
{
  switch(X) {
  case nemo_io::energy:  return EnergyTag;
  default: cerr<<"FATAL ERROR: Unknown nemo_io::SingleVector\n"; exit(1);
  }
}
//------------------------------------------------------------------------------
inline char* tag(const nemo_io::SingleMatrix X)
{
  switch(X) {
  case nemo_io::KinT: return KETensorTag;
  case nemo_io::PotT: return PETensorTag;
  case nemo_io::AmT:  return AMTensorTag;
  default: cerr<<"FATAL ERROR: Unknown nemo_io::SingleMatrix\n"; exit(1);
  }
}
//------------------------------------------------------------------------------
inline char* tag(const nemo_io::SinglePhases X)
{
  switch(X) {
  case nemo_io::cofm: return CMPhaseSpaceTag;
  default: cerr<<"FATAL ERROR: Unknown nemo_io::SinglePhases\n"; exit(1);
  }
}
//------------------------------------------------------------------------------
inline char* tag(const nemo_io::BodiesScalar X)
{
  switch(X) {
  case nemo_io::mass: return MassTag;
  case nemo_io::pot:  return PotentialTag;
  case nemo_io::aux:  return AuxTag;
  case nemo_io::eps:  return "Eps";
  default: cerr<<"FATAL ERROR: Unknown nemo_io::BodiesScalar\n"; exit(1);
  }
}
//------------------------------------------------------------------------------
inline char* tag(const nemo_io::BodiesVector X)
{
  switch(X) {
  case nemo_io::acc:    return AccelerationTag;
  case nemo_io::pos:    return PosTag;
  case nemo_io::vel:    return VelTag;
  default: cerr<<"FATAL ERROR: Unknown nemo_io::BodiesVector\n"; exit(1);
  }
}
//------------------------------------------------------------------------------
inline char* tag(const nemo_io::BodiesPhases X)
{
  switch(X) {
  case nemo_io::posvel: return PhaseSpaceTag;
  default: cerr<<"FATAL ERROR: Unknown nemo_io::BodiesPhases\n"; exit(1);
  }
}
//------------------------------------------------------------------------------
inline void noallocerror() {
  cerr<<"FATAL ERROR: nemo_io: no output memory allocated\n";
  exit(1);
}
//------------------------------------------------------------------------------
void nemo_io::open(const char* file, char* control)
{
  if(STREAM) strclose(mystream);
  STREAM = static_cast<void*>(stropen(file,control));
}  
//------------------------------------------------------------------------------
void nemo_io::close()
{
  if(STREAM) strclose(mystream);
  STREAM = 0;
}  
//------------------------------------------------------------------------------
#define ISPRESENT(NAME)						\
bool nemo_io::is_present(const NAME X) const			\
{								\
  return get_tag_ok(mystream,tag(X));	\
}

ISPRESENT(Set)
ISPRESENT(SingleScalar)
ISPRESENT(SingleVector)
ISPRESENT(SingleMatrix)
ISPRESENT(SinglePhases)
ISPRESENT(BodiesScalar)
ISPRESENT(BodiesVector)
ISPRESENT(BodiesPhases)
#undef ISPRESENT
//------------------------------------------------------------------------------
nemo_io::nemo_io(const char* file, char* control)
{
  STREAM       = static_cast<void*>(stropen(file,control));
  N            = 0;
  BODIESSCALAR = 0;
  BODIESARRAYS = 0;
}
//------------------------------------------------------------------------------
nemo_io::~nemo_io()
{
  close();
  reset();
}
//------------------------------------------------------------------------------
void nemo_io::open_set(const Set S, const bool get) const
{
  if(get) get_set(mystream,tag(S));
  else    put_set(mystream,tag(S));
}
//------------------------------------------------------------------------------
void nemo_io::close_set(const Set S, const bool get) const
{
  if(get) get_tes(mystream,tag(S));
  else    put_tes(mystream,tag(S));
}
//------------------------------------------------------------------------------
int nemo_io::read_N() const
{
  reset();
  get_data(mystream,NobjTag,IntType,&N,0);
  return N;
}
//------------------------------------------------------------------------------
float nemo_io::read(const SingleScalar X, float*Y) const
{
  if(Y) {
    get_data_coerced(mystream,tag(X),FloatType,Y,0);
    return *Y;
  } else {
    float scal;
    get_data_coerced(mystream,tag(X),FloatType,&scal,0);
    return scal;
  }
}
//------------------------------------------------------------------------------
void nemo_io::read(const SingleVector X, float*Y) const
{
  if(Y) get_data_coerced(mystream,tag(X),FloatType,Y,NDM,0);
  else  get_data_coerced(mystream,tag(X),FloatType,SINGLEVECTOR,NDM,0);
}
//------------------------------------------------------------------------------
void nemo_io::read(const SinglePhases X, float*Y) const
{
  if(Y) get_data_coerced(mystream,tag(X),FloatType,Y,2,NDM,0);
  else  get_data_coerced(mystream,tag(X),FloatType,SINGLEPHASES,2,NDM,0);
}
//------------------------------------------------------------------------------
void nemo_io::read(const SingleMatrix X, float*Y) const
{
  if(Y) get_data_coerced(mystream,tag(X),FloatType,Y,NDM,NDM,0);
  else  get_data_coerced(mystream,tag(X),FloatType,SINGLEMATRIX,NDM,NDM,0);
}
//------------------------------------------------------------------------------
void nemo_io::read(const BodiesScalar X, float*Y) const
{
  if(Y)
    get_data_coerced(mystream,tag(X),FloatType,Y,N,0);
  else {
    allocscalar();
    get_data_coerced(mystream,tag(X),FloatType,BODIESSCALAR,N,0);
  }
}
//------------------------------------------------------------------------------
void nemo_io::read(const BodiesVector X, float*Y) const
{
  if(Y)
    get_data_coerced(mystream,tag(X),FloatType,Y,N,NDM,0);
  else {
    allocarrays();
    if(X==vel)
      get_data_coerced(mystream,tag(X),FloatType,BODIESARRAYS+N*NDM,N,NDM,0);
    else
      get_data_coerced(mystream,tag(X),FloatType,BODIESARRAYS,N,NDM,0);
  }
}
//------------------------------------------------------------------------------
void nemo_io::read(const BodiesPhases X, float*Y) const
{
  if(Y)
    get_data_coerced(mystream,tag(X),FloatType,Y,N,2,NDM,0);
  else {
    allocarrays();
    get_data_coerced(mystream,tag(X),FloatType,BODIESARRAYS,N,2,NDM,0);
  }
}
//------------------------------------------------------------------------------
void nemo_io::read_history() const
{
  get_history(mystream);
}
//------------------------------------------------------------------------------
void nemo_io::write_N(const int n) const
{
  reset();
  N = n;
  put_data(mystream,NobjTag,IntType,&N,0);
}
//------------------------------------------------------------------------------
void nemo_io::write(const CoSys X) const
{
  CS = CSCode(tag(X), NDM, 2);
  put_data(mystream,CoordSystemTag,IntType,&CS,0);
}
//------------------------------------------------------------------------------
void nemo_io::write(const SingleScalar X, const float scal) const
{
  float s = scal;
  put_data(mystream,tag(X),FloatType,&s,0);
}
//------------------------------------------------------------------------------
void nemo_io::write(const SingleVector X, float* Y) const
{
  if(Y) put_data(mystream,tag(X),FloatType,Y,NDM,0);
  else  put_data(mystream,tag(X),FloatType,SINGLEVECTOR,NDM,0);
}
//------------------------------------------------------------------------------
void nemo_io::write(const SinglePhases X, float* Y) const
{
  if(Y) put_data(mystream,tag(X),FloatType,Y,2,NDM,0);
  else  put_data(mystream,tag(X),FloatType,SINGLEPHASES,2,NDM,0);
}
//------------------------------------------------------------------------------
void nemo_io::write(const SingleMatrix X, float* Y) const
{
  if(Y) put_data(mystream,tag(X),FloatType,Y,NDM,NDM,0);
  else  put_data(mystream,tag(X),FloatType,SINGLEMATRIX,NDM,NDM,0);
}
//------------------------------------------------------------------------------
void nemo_io::write(const BodiesScalar X, float* Y) const
{
  if(Y) put_data(mystream,tag(X),FloatType,Y,N,0);
  else if(BODIESARRAYS==0) noallocerror();
  else  put_data(mystream,tag(X),FloatType,BODIESSCALAR,N,0);
}
//------------------------------------------------------------------------------
void nemo_io::write(const BodiesVector X, float* Y) const
{
  if(Y)           put_data(mystream,tag(X),FloatType,Y,N,NDM,0);
  else if(BODIESARRAYS==0) noallocerror();
  else if(X==vel) put_data(mystream,tag(X),FloatType,BODIESARRAYS+N*NDM,N,NDM,0);
  else            put_data(mystream,tag(X),FloatType,BODIESARRAYS,N,NDM,0);
}
//------------------------------------------------------------------------------
void nemo_io::write(const BodiesPhases X, float* Y) const
{
  if(Y) put_data(mystream,tag(X),FloatType,Y,N,2,NDM,0);
  else if(BODIESARRAYS==0) noallocerror();
  else  put_data(mystream,tag(X),FloatType,BODIESARRAYS,N,2,NDM,0);
}
//------------------------------------------------------------------------------
void nemo_io::write_history() const
{
  put_history(mystream);
}
#undef mystream
#undef NDM
//------------------------------------------------------------------------------
