//-----------------------------------------------------------------------------+
//                                                                             |
// stic.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// written by Walter Dehnen, 2001                                              |
// e-mail:   dehnen@mpia-hd.mpg.de                                             |
// address:  Max-Planck Institut fr Astronomie, Knigstuhl 69117 Heidelberg   |
//           Germany                                                           |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// support for sticky particles                                                |
//                                                                             |
//-----------------------------------------------------------------------------+

#include <nbdy/stic.h>    // definition of sticky tree and interactor           
#include <nbdy/tree.h>    // definition of tree                                 

using namespace nbdy;

////////////////////////////////////////////////////////////////////////////////
// class nbdy::basic_finder                                                     
////////////////////////////////////////////////////////////////////////////////
inline void basic_finder::add_pair(soul_type*A, soul_type*B) const
{
  if(N<MAX)
    if(mybody(A) < mybody(B)) {
      BL[N][0] = mybody(A);
      BL[N][1] = mybody(B);
    } else {
      BL[N][0] = mybody(B);
      BL[N][1] = mybody(A);
    }
  N++;
  if(N==MAX) warning("interaction list overflow");
}
////////////////////////////////////////////////////////////////////////////////
// class nbdy::sticky_finder                                                    
////////////////////////////////////////////////////////////////////////////////
void sticky_finder::single(soul_type*A, soul_type*B) const
{
  if(!is_sink(A) && !is_sink(B)) return;             // no sinks -> no job      
  register vect R = pos(A)-pos(B);                   // R   = distance vector   
  register real xq= square(size(A)+size(B));         // x^2 = (combined size)^2 
  if(norm(R) < xq) return add_pair(A,B);             // |R|<x: overlap at t=0   
  if(TAU==zero) return;                              // no times t>0            
  register vect V = vel(A)-vel(B);                   // V   = velocity diff     
  register real RV= R*V;                             // R*V = scalar product    
  if(RV>zero) return;                                // R*V>0: diverging orbits 
  register real t = min(TAU, -RV/norm(V));           // t of minimum distance   
  if(norm(R+t*V) < xq) return add_pair(A,B);         // d_min<x: overlapping    
}
//------------------------------------------------------------------------------
#define CHECK_PAIR								\
  R = pos(A)-pos(B);                                 /* distance vector     */	\
  xq= square(size(A)+size(B));                       /* (combined size)^2   */	\
  if(norm(R) < xq) { add_pair(A,B); continue; }      /* overlap at t=0      */	\
  if(TAU==zero) continue;                            /* no times t>0        */	\
  V = vel(A)-vel(B);                                 /* velocity difference */	\
  RV= R*V;                                           /* scalar product      */	\
  if(RV>zero) continue;                              /* diverging orbits    */	\
  t = min(TAU,-RV/norm(V));                          /* time of min distance*/	\
  if(norm(R+t*V) < xq) add_pair(A,B);                /* overlap: add to list*/

inline void sticky_finder::many(const bool all, soul_type*A,
				soul_type*B0, const soul_type*BN) const
{
  register vect   R,V;
  register real   xq,RV,t;
  register soul_type *B;
  if(all) for(B=B0; B<BN; B++)                { CHECK_PAIR }
  else    for(B=B0; B<BN; B++) if(is_sink(B)) { CHECK_PAIR }
}
#undef CHECK_PAIR
//------------------------------------------------------------------------------
inline bool sticky_finder::discard(cell_type*A, soul_type*B) const
  // true  if there cannot possibly be an interaction between any A and B       
  // false if an interaction between any A and B is possible                    
{
  register vect R = pos(A)-pos(B);                   // R   = distance vector   
  register real
    Rq = norm(R),                                    // R^2 = distance^2        
    x  = size(A)+size(B);                            // x   = combined size     
  if(Rq < x*x) return false;                         // overlap at t=0          
  if(TAU==zero) return true;                         // no times t>0            
  register vect V = vel(A)-vel(B);                   // V   = velocity diff     
  register real
    w  = vrad(A),                                    // w   = v-size            
    wq = w*w,                                        // w^2 = (v-size)^2        
    RV = R*V,                                        // R*V = scalar product    
    RVq= RV*RV;                                      // (R*V)^2                 
  if(RV>zero && RVq>wq*Rq) return true;              // R*V/|R| > w: diverging  
  register real
    Vq = norm(V),                                    // V^2 = (v-distance)^2    
    t  = (wq>=Vq)? TAU :                             // v-size too large        
    min(TAU, (w*SQRT((Rq*Vq-RVq)/(Vq-wq))-RV)/Vq );  // t of min dist           
  if(norm(R+t*V) < square(x+t*w)) return false;      // possible overlap        
  return true;                                       // no overlap -> discard   
}
//------------------------------------------------------------------------------
inline bool sticky_finder::discard(cell_type*A, cell_type*B) const
  // true  if there cannot possibly be an interaction between any A and B       
  // false if an interaction between any A and B is possible                    
{
  register vect R = pos(A)-pos(B);                   // R   = distance vector   
  register real
    Rq = norm(R),                                    // R^2 = distance^2        
    x  = size(A)+size(B);                            // x   = combined size     
  if(Rq < x*x) return false;                         // overlap at t=0          
  if(TAU==zero) return true;                         // no times t>0            
  register vect V = vel(A)-vel(B);                   // V   = velocity diff     
  register real
    w  = vrad(A)+vrad(B),                            // w   = combined v-size   
    wq = w*w,                                        // w^2 = (comb v-size)^2   
    RV = R*V,                                        // R*V = scalar product    
    RVq= RV*RV;                                      // (R*V)^2                 
  if(RV>zero && RVq>wq*Rq) return true;              // R*V/|R| > w: diverging  
  register real
    Vq = norm(V),                                    // V^2 = (v-distance)^2    
    t  = (wq>=Vq)? TAU :                             // v-size too large        
    min(TAU, (w*SQRT((Rq*Vq-RVq)/(Vq-wq))-RV)/Vq );  // t of min dist           
  if(norm(R+t*V) < square(x+t*w)) return false;      // possible overlap        
  return true;                                       // no overlap -> discard   
}
////////////////////////////////////////////////////////////////////////////////
// class nbdy::neighbour_finder                                                 
////////////////////////////////////////////////////////////////////////////////
void neighbour_finder::single(soul_type*A, soul_type*B) const
{
  if(!is_sink(A) && !is_sink(B)) return;             // no sinks -> no job      
  register real Rq = dist_sq(pos(A),pos(B));         // R^2 = distance^2        
  if(Rq < sizeq(A) || Rq < sizeq(B))                 // |R| < max(s_A,s_B) ?    
    return add_pair(A,B);
}
//------------------------------------------------------------------------------
#define CHECK_PAIR								\
  Rq = dist_sq(pos(A),pos(B));                        /* R^2 = distance^2    */ \
  if(Rq < sizeq(A) || Rq < sizeq(B)) add_pair(A,B);   /* |R| < max(s_A,s_B) ?*/

inline void neighbour_finder::many(const bool all, soul_type*A,
				   soul_type*B0, const soul_type*BN) const
{
  register real Rq;
  register soul_type *B;
  if(all) for(B=B0; B<BN; B++)                { CHECK_PAIR }
  else    for(B=B0; B<BN; B++) if(is_sink(B)) { CHECK_PAIR }
}
#undef CHECK_PAIR
//------------------------------------------------------------------------------
inline bool neighbour_finder::discard(cell_type*A, soul_type*B) const
{
  register real
    Rq = dist_sq(pos(A),pos(B)),                     // R^2 = distance^2        
    x  = max(rmax(A)+size(B),size(A));               // max size distance       
  if(Rq < square(x)) return false;                   // |R| < max(s_A,r_A+s_B)  
  return true;
}
//------------------------------------------------------------------------------
inline bool neighbour_finder::discard(cell_type*A, cell_type*B) const
{
  register real
    Rq = dist_sq(pos(A),pos(B)),                     // R^2 = distance^2        
    x  = max(rmax(A)+size(B),rmax(B)+size(A));       // max size distance       
  if(Rq < square(x)) return false;                   // |R| < max(rB+sA,rA+sB)  
  return true;
}
////////////////////////////////////////////////////////////////////////////////
// class nbdy::neighbour_counter                                                
////////////////////////////////////////////////////////////////////////////////
void neighbour_counter::single(soul_type*A, soul_type*B) const
{
  if(!is_sink(A) && !is_sink(B)) return;             // no sinks -> no job      
  register real Rq = dist_sq(pos(A),pos(B));         // R^2 = distance^2        
  if(is_sink(A) && Rq < sizeq(A)) A->inc();          // A is sink && |R|<siza(A)
  if(is_sink(B) && Rq < sizeq(B)) B->inc();          // B is sink && |R|<siza(B)
}
//------------------------------------------------------------------------------
inline void neighbour_counter::many(const bool all, soul_type*A,
				    soul_type*B0, const soul_type*BN) const
{
  register real      Rq;
  register soul_type*B;
  if(is_sink(A))
    if(all)
      for(B=B0; B<BN; B++) {
        Rq = dist_sq(pos(A),pos(B));
        if(Rq < sizeq(A)) A->inc();
        if(Rq < sizeq(B)) B->inc();
      }
    else
      for(B=B0; B<BN; B++) {
        Rq = dist_sq(pos(A),pos(B));
        if(              Rq < sizeq(A)) A->inc();
        if(is_sink(B) && Rq < sizeq(B)) B->inc();
      }
  else
    if(all)
      for(B=B0; B<BN; B++) {
        Rq = dist_sq(pos(A),pos(B));
        if(Rq < sizeq(B)) B->inc();
      }
    else
      for(B=B0; B<BN; B++) if(is_sink(B)) {
        Rq = dist_sq(pos(A),pos(B));
        if(Rq < sizeq(B)) B->inc();
      }
}
//------------------------------------------------------------------------------
inline bool neighbour_counter::discard(cell_type*A, soul_type*B) const
{
  register real
    Rq = dist_sq(pos(A),pos(B)),                     // R^2 = distance^2        
    x  = max(rmax(A)+size(B),size(A));               // max size distance       
  if(Rq < square(x)) return false;                   // |R| < max(s_A,r_A+s_B)  
  return true;
}
//------------------------------------------------------------------------------
inline bool neighbour_counter::discard(cell_type*A, cell_type*B) const
{
  register real
    Rq = dist_sq(pos(A),pos(B)),                     // R^2 = distance^2        
    x  = max(rmax(A)+size(B),rmax(B)+size(A));       // max size distance       
  if(Rq < square(x)) return false;                   // |R| < max(rb+sA,rA+sB)  
  return true;
}
////////////////////////////////////////////////////////////////////////////////
// class nbdy::basic_iactor                                                     
////////////////////////////////////////////////////////////////////////////////
const indx N_[3] = {100, 25, 100};
//------------------------------------------------------------------------------
inline bool basic_iactor::many(cell_type*A, soul_type*B) const
{
  many(is_sink(B) || al_sink(A), B,
       A->begin_soul_descendants(), A->end_soul_descendants());
  return true;
}
//------------------------------------------------------------------------------
inline bool basic_iactor::many(cell_type*A, cell_type*B) const
{
  LoopAllSouls(cell_type,B,Bi) many(A,Bi);
  return true;
}
//------------------------------------------------------------------------------
inline bool basic_iactor::many(cell_type*A) const
{
  LoopAllSouls(cell_type,A,Ai)
    many(al_sink(A),Ai,Ai+1,A->end_soul_descendants());
  return true;
}
//------------------------------------------------------------------------------
bool basic_iactor::interact(cell_type*A, soul_type*B) const
{
  if(!is_sink(A) && !is_sink(B)) return true;        // no sinks -> no job      
  if(discard(A,B))               return true;        // try to discard          
  if(number(A)<N_[0])            return many(A,B);   // perform many single     
                                 return false;       // must split              
}
//------------------------------------------------------------------------------
bool basic_iactor::interact(cell_type*A, cell_type*B) const
{
  if(!is_sink(A) && !is_sink(B)) return true;        // no sinks -> no job      
  if(discard(A,B))               return true;        // try to discard          
  if(number(A)*number(B)<N_[1])  return many(A,B);   // perform many single     
                                 return false;       // must split              
}
//------------------------------------------------------------------------------
bool basic_iactor::interact(cell_type*A) const
{
  if(!is_sink(A))                return true;        // no sinks -> no job      
  if(number(A)<N_[2])            return many(A);     // perform many single     
                                 return false;       // must split              
}
////////////////////////////////////////////////////////////////////////////////
// class nbdy::sticky_tree                                                      
////////////////////////////////////////////////////////////////////////////////
#define LoopMyCellsDown							\
  for(register cell_type*Ci= begin_cells(); Ci!= end_cells(); ++Ci)
#define LoopMyCellsUp							\
  for(register cell_type*Ci=rbegin_cells(); Ci!=rend_cells(); --Ci)
#define LoopMySouls							\
  for(register soul_type*Si= begin_souls(); Si!= end_souls(); ++Si)
//------------------------------------------------------------------------------
void sticky_tree::prepare_sticky()
{
  if(S0==0) return;                                // no souls -> no job        
  // 1. update souls with size and vel (body flag & pos known from tree souls)  
  if(use_bodies()) LoopMySouls Si->set_sticky(my_bodies());
  else             LoopMySouls Si->set_sticky(S,V);
  // 2. for all cells: compute Sum x_i and Sum v_i, pass sink flag up           
  register vect spos,svel;                         // vel & pos sum             
  LoopMyCellsUp {
    spos = zero;
    svel = zero;
    Ci->reset_flag();
    LoopSoulKids(cell_type,Ci,s) {
      spos+= pos(s);
      svel+= vel(s);
      Ci->add_body_flag(s);
    }
    LoopCellKids(cell_type,Ci,c) {
      spos+= pos(c);
      svel+= vel(c);
      Ci->add_body_flag(c);
    }
    Ci->pos() = spos;
    Ci->vel() = svel;
  }
  // 3. for all cells: normalize pos & vel; compute size & vrad                  
  register real iN,dmax,vmax;
  LoopMyCellsUp {
    iN = 1./real(number(Ci));
    Ci->pos() *= iN;
    Ci->vel() *= iN;
    dmax = zero;
    vmax = zero;
    LoopSoulKids(cell_type,Ci,s) {
      update_max(dmax,SQRT(dist_sq(pos(Ci),pos(s))) + size(s));
      update_max(vmax,SQRT(dist_sq(vel(Ci),vel(s))));
    }
    LoopCellKids(cell_type,Ci,c) {
      update_max(dmax,SQRT(dist_sq(pos(Ci),pos(c))) + size(c));
      update_max(vmax,SQRT(dist_sq(vel(Ci),vel(c))) + vrad(c));
    }
    Ci->size() = dmax;
    Ci->vrad() = vmax;
  }
}
//------------------------------------------------------------------------------
void sticky_tree::prepare_sph()
{
  if(S0==0) return;                                // no souls -> no job        
  // 1. update souls with size and vel (body flag & pos known from tree souls)  
  if(use_bodies()) LoopMySouls Si->set_sph(my_bodies());
  else             LoopMySouls Si->set_sph(S);
  // 2. for all cells: compute Sum x_i, pass sink flag up
  register vect spos;
  LoopMyCellsUp {
    spos = zero;
    Ci->reset_flag();
    LoopSoulKids(cell_type,Ci,s) { spos+= pos(s); Ci->add_body_flag(s); }
    LoopCellKids(cell_type,Ci,c) { spos+= pos(c); Ci->add_body_flag(c); }
    Ci->pos() = spos;
  }
  // 3. for all cells: normalize pos, compute size
  register real iN,dmax,smax,d;
  LoopMyCellsUp {
    iN = 1./real(number(Ci));
    Ci->pos() *= iN;
    dmax = zero;
    smax = zero;
    LoopSoulKids(cell_type,Ci,s) {
      d = SQRT(dist_sq(pos(Ci),pos(s)));
      update_max(dmax, d + size(s));
      update_max(smax, d);
    }
    LoopCellKids(cell_type,Ci,c) {
      d = SQRT(dist_sq(pos(Ci),pos(c)));
      update_max(dmax, d + size(c));
      update_max(smax, d + rmax(c));
    }
    Ci->size() = dmax;
    Ci->rmax() = smax;
  }
}
//------------------------------------------------------------------------------
void sticky_tree::set_num(soul*S0) const
{
  LoopMySouls if(is_sink(Si)) my_bodies()->rho(mybody(Si)) = num(Si);
}
//------------------------------------------------------------------------------
void sticky_tree::set_num(areal*NUM) const
{
  LoopMySouls if(is_sink(Si)) NUM[mybody(Si)] = num(Si);
}
//------------------------------------------------------------------------------
#undef LoopMyCellsDown
#undef LoopMyCellsUp
#undef LoopMySouls
////////////////////////////////////////////////////////////////////////////////
