// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// tree.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
#include <nbdy/tree.h>
#include <nbdy/bman.h>
#include <nbdy/bcell.h>

// #include <time.h>

////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
#define LoopDims for(register int d=0; d<NDIM; d++)
  //////////////////////////////////////////////////////////////////////////////
  // auxiliarty constants                                                       
  //////////////////////////////////////////////////////////////////////////////
  const int BIT[8] = {1,2,4,8,16,32,64,128};
  const flag_type SUBTREECELL   = 256;             // cell = cell of subtree    
  const flag_type SUBTREEZOMBIE = 512;             // cell with 1 subtree subcel
  const flag_type SUBTREE_FLAGS = 896;             // sum of all subtree flags  
  //////////////////////////////////////////////////////////////////////////////
  // auxiliarty functions                                                       
  //////////////////////////////////////////////////////////////////////////////
#define CBFCF const basic_flag*const&F
  inline void flag_as_subtreecell  (basic_flag*F) { F->add_flag(SUBTREECELL); }
  inline void flag_as_subtreezombie(basic_flag*F) { F->add_flag(SUBTREEZOMBIE);}
  inline void unflag_subtreecell   (basic_flag*F) { F->un_flag (SUBTREECELL); }
  inline void unflag_subtreezombie (basic_flag*F) { F->un_flag (SUBTREEZOMBIE);}
  inline void unflag_subtree_flags (basic_flag*F) { F->un_flag (SUBTREE_FLAGS);}
  inline bool in_subtree           (CBFCF) { return F->is_set(SUBTREE); }
  inline bool is_subtreecell       (CBFCF) { return F->is_set(SUBTREECELL); }
  inline bool is_subtreezombie     (CBFCF) { return F->is_set(SUBTREEZOMBIE);}
#undef CBFCF
  //----------------------------------------------------------------------------
  inline vect Integer(const vect& x)
  {
    // This routines returns, in each dimension, the nearest integer to x       
    register vect c = zero;                        // reset return position     
    LoopDims c[d]=int(x[d]+half);                  // find center position      
    return c;                                      // and return it             
  }
  //----------------------------------------------------------------------------
  template<class LEAF>
  inline vect RootCenter(LEAF*L0, const LEAF*LN)
  {
    // This routines returns the root center nearest to the mean position       
    register vect x = zero;
    for(register LEAF*L=L0; L!=LN; ++L)
      x += L->pos();
    x /= real(int(LN-L0));
    return Integer(x);
  }
  //----------------------------------------------------------------------------
  inline real RootRadius(const vect& X, const vect& xmin, const vect& xmax)
  {
    // This routines returns the half-size R of the smallest cube, centered on X
    // that contains the points xmin and xmax.                                  
    register real R,D=zero;                        // R, distance to leaf       
    LoopDims {                                     // loop dimensions       >   
      R = max(abs(xmax[d]-X[d]),abs(xmin[d]-X[d]));//   distance to xmin, xmax  
      update_max(D,R);                             //   update maximum distance 
    }                                              // <                         
    return pow(two,int(one+LOG(D)/LOG(two)));      // return R                  
  }
  //----------------------------------------------------------------------------
  inline uint octant(const vect& cen, const vect& pos)
  {
    register uint 
      oct=(pos[0] > cen[0])? 1 : 0;
    if(pos[1] > cen[1]) oct |= 2;
#if NDIM==3
    if(pos[2] > cen[2]) oct |= 4;
#endif
    return oct;                                    // return octant             
  }
  //----------------------------------------------------------------------------
  inline bool contains(const vect& cen,
		       const real& rad,
		       const vect& pos)
  {
#if NDIM==2
    return abs(cen[0]-pos[0]) < rad
      &&   abs(cen[1]-pos[1]) < rad
#else
    return abs(cen[0]-pos[0]) < rad
      &&   abs(cen[1]-pos[1]) < rad
      &&   abs(cen[2]-pos[2]) < rad;
#endif
  }
  //////////////////////////////////////////////////////////////////////////////
  // auxiliarty types                                                           
  //////////////////////////////////////////////////////////////////////////////
  //----------------------------------------------------------------------------
  // class nbdy::sub_tree_builder<SUB_CELL,PARENT_PCELL>                        
  //----------------------------------------------------------------------------
  template<class SUB_CELL, class PARENT_CELL>
  class sub_tree_builder :
    protected cell_access<SUB_CELL>                // access to subtree cells   
  {
#define PCELL static_cast<PARENT_CELL*>            /* cast: parent-tree cell* */
    //--------------------------------------------------------------------------
    // some types                                                               
    //--------------------------------------------------------------------------
    typedef typename PARENT_CELL::soul_type parent_soul;
    //--------------------------------------------------------------------------
    // static methods                                                           
    //--------------------------------------------------------------------------
    static parent_soul* find_soul(const PARENT_CELL*C)
      // parent cell with one soul in subtree: find that one                    
    {
      LoopCellKids(typename PARENT_CELL,C,c) if(in_subtree(c))
	return find_soul(PCELL(c));
      LoopSoulKids(typename PARENT_CELL,C,s) if(in_subtree(s)) return s;
      std::cerr<<" blind ending in subtree cutting 1\n";
      return 0;
    }
    //--------------------------------------------------------------------------
    static PARENT_CELL* find_cell(const PARENT_CELL*C)
      // parent cell with one cell descendant in subtree: find that one         
    {
      LoopCellKids(typename PARENT_CELL,C,c) if(in_subtree(c)) {
	if(is_subtreecell  (c)) return c;
	if(is_subtreezombie(c)) return find_cell(PCELL(c));
      }
      std::cerr<<" blind ending in subtree cutting 2\n";
      return 0;
    }
    //--------------------------------------------------------------------------
  public:
    static uint link(const PARENT_CELL*P,          // parent cell to link to    
			   cell_type  *C,          // cell to link with         
			   cell_type  *&Cf,        // next free cell            
			   soul_type  *&Sf)        // next free soul            
    {
      register uint   dep=0;                       // depth                     
      radius_(C) = radius(P);                      // copy radius               
      center_(C) = center(P);                      // copy center               
      number_(C) = 0;                              // reset number              
      nsouls_(C) = 0;                              // reset cell: # soul kids   
      ncells_(C) = 0;                              // reset cell: # cell kids   
      csouls_(C) = Sf;                             // set cell: sub-souls       
      LoopSoulKids(typename PARENT_CELL,P,ps)      // loop soul kids of Pcell  >
	if(in_subtree(ps)) {                       //   if soul == subt soul   >
	  (Sf++)->copy(ps);                        //     copy link to body etc 
	  nsouls_(C)++;                            //     increment # soul kids 
	}                                          // < <                       
      LoopCellKids(typename PARENT_CELL,P,pc)      // loop cell kids of Pcell  >
	if(is_subtreecell(pc) ||                   //   IF cell == subt cell   >
	   is_subtreezombie(pc) ) ncells_(C)++;    //     count # cell kids     
	else if(in_subtree(pc)) {                  //   < ELSE >                
	  (Sf++)->copy(find_soul(PCELL(pc)));      //     find soul & copy it   
	  nsouls_(C)++;                            //     increment # soul kids 
	}                                          // < <                       
      number_(C) += nsouls_(C);                    // # souls <= # soul kids    
      if(ncells_(C)) {                             // IF cell has cell kids >   
	register uint      de;                     //   depth of sub-cell       
	register cell_type*Ci=Cf;                  //   remember free cells     
	ccells_(C) = Ci;                           //   set cell children       
	Cf += ncells_(C);                          //   reserve children cells  
	LoopCellKids(typename PARENT_CELL,P,pc)    //   loop cell kids of Pcell 
	  if(is_subtreecell(pc)) {                 //     IF subt cell          
	    de = link(PCELL(pc),Ci,Cf,Sf);         //       link sub cells      
	    if(de>dep) dep=de;                     //       update depth        
	    number_(C) += number_(Ci++);           //       count soul descends 
	  } else if(is_subtreezombie(pc)) {        //     ELSE if zombie cell   
	    de = link(find_cell(PCELL(pc)),Ci,Cf,Sf); //    link sub cells      
	    if(de>dep) dep=de;                     //       update depth        
	    number_(C) += number_(Ci++);           //       count soul descends 
	  }                                        //     <                     
      } else                                       // < ELSE >                  
	ccells_(C) = 0;                            //   set pter to cell kids   
      return dep+1;                                // return cells depth        
    }
#undef PCELL
  };
  //----------------------------------------------------------------------------
  // class nbdy::node                                                           
  //----------------------------------------------------------------------------
  class node {
  private:
    vect POS;                                      // position                  
  public:
    node(const node&);                             // not implemented           
    node() {}                                      // default constructor       
    node(const vect&X) : POS(X) {}                 // constructor from position 
    vect&       pos()                  { return POS; }
    vect const &pos()            const { return POS; }
    real const &pos(const int i) const { return POS[i]; }
    friend const vect& pos(const node*N) { return N->POS; }
  };
  //----------------------------------------------------------------------------
  // class nbdy::leaf<SOUL>                                                     
  //----------------------------------------------------------------------------
  template<class SOUL>
  struct leaf :
    public node
  {
    typedef SOUL                     soul_type;    // type of associated soul   
    typedef const soul_type         *soul_pter;    // const pointer to soul     
    typedef typename SOUL::bodies_type bodies_type;// type of bodies            
    typedef typename bodies_type::body_type body_type;// type of body           
    //--------------------------------------------------------------------------
    // data of class leaf<SOUL>                                                 
    //--------------------------------------------------------------------------
    mutable leaf*NEXT;
    unsigned     LINK;
    //--------------------------------------------------------------------------
    // non-const methods                                                        
    //--------------------------------------------------------------------------
    void  set_up  (const body_type& B) {
      pos() = nbdy::pos(B);
      LINK  = index(B);
    }
    //--------------------------------------------------------------------------
    void  set_up  (soul_pter const& S) {
      pos()  = nbdy::cofm(S);
      MYLINK = mylink(S);
    }
    //--------------------------------------------------------------------------
    void  set_up  (const areal*x[NDIM], const uint i) {
      LINK = i;
      pos()[0] = x[0][i];
      pos()[1] = x[1][i];
#if NDIM==3
      pos()[2] = x[2][i];
#endif
    }
    //--------------------------------------------------------------------------
    void  set_soul(soul_type*S, const bodies_type*B) {
      B->set_soul(S,LINK);
      S->link_and_set_up(B,LINK);
    }
    //--------------------------------------------------------------------------
    void  set_soul(soul_type*S) {
      S->link_and_set_up(pos(),LINK);
    }
  };
  //----------------------------------------------------------------------------
  // class nbdy::box<CELL>                                                      
  //----------------------------------------------------------------------------
  template<class CELL>
  struct box :
    public node
  {
    typedef CELL                     cell_type;    // type of associated cell   
    typedef typename CELL::soul_type soul_type;    // type of associated soul   
    typedef leaf<soul_type>          leaf_type;    // type of associated leaf   
    //--------------------------------------------------------------------------
    // data of class box<CELL>                                                  
    //--------------------------------------------------------------------------
    real       RADIUS;                             // half size of the cube     
    indx       TYPE;                               // bitfield: 1=cell, 0=leaf  
    bool       IS_ZOMBIE;                          // are we a zombie cell?     
    node      *OCT[NSUB];                          // octants                   
    uint       NUMBER;                             // number of leafs           
    leaf_type *LEAFS;                              // linked list of leafs      
    //--------------------------------------------------------------------------
    // constructor                                                              
    //--------------------------------------------------------------------------
    box() {}
    //--------------------------------------------------------------------------
    // const methods                                                            
    //--------------------------------------------------------------------------
    bool marked_as_box  (const int i) const {return TYPE & BIT[i]; }
    bool marked_as_leaf (const int i) const {return !marked_as_box(i); }
    bool octant_is_empty(const int i) const {return OCT[i]==0; }
    bool octant_is_box  (const int i) const {return OCT[i] && marked_as_box(i);}
    bool octant_is_leaf (const int i) const {return OCT[i] && marked_as_leaf(i);}
    vect const&center   ()            const {return pos(); }
    real const&center   (const int i) const {return pos(i); }
    //--------------------------------------------------------------------------
    uint octant(const leaf_type*L) const {         // I: leaf (must be in box)  
      return nbdy::octant(pos(),L->pos());
    }
    //--------------------------------------------------------------------------
    uint octant(const box*P) const {                // I: box (must be in box)  
      return nbdy::octant(pos(),P->pos());
    }
    //--------------------------------------------------------------------------
    bool contains(const leaf_type*L) const {
      return nbdy::contains(center(),RADIUS,L->pos());
    }
    //--------------------------------------------------------------------------
    // non-const methods                                                        
    //--------------------------------------------------------------------------
    void set_up(const cell_type*C) {
      pos()  = center(C);
      RADIUS = radius(C);
      TYPE      = 0;
      NUMBER    = 0;
      IS_ZOMBIE = true;
      LEAFS     = 0;
      for(register int b=0; b!=NSUB; ++b) OCT[b] = 0;
    }
    //--------------------------------------------------------------------------
    box& operator= (const box& P) {
      pos()     = P.pos();
      RADIUS    = P.RADIUS;
      TYPE      = P.TYPE;
      NUMBER    = P.NUMBER;
      IS_ZOMBIE = P.IS_ZOMBIE;
      LEAFS     = P.LEAFS;
      for(register int b=0; b!=NSUB; ++b) OCT[b] = P.OCT[b];
    }
    //--------------------------------------------------------------------------
    void mark_as_box    (const int i) { TYPE |=  BIT[i]; }
    void mark_as_leaf   (const int i) { TYPE &= ~BIT[i]; }
    vect&      center   ()            { return pos(); }
    //--------------------------------------------------------------------------
    box& reset_octants() {
      for(register node**P=OCT; P!=OCT+NSUB; ++P) *P = 0;
      return *this;
    }
    //--------------------------------------------------------------------------
    box& reset() {
      TYPE      = 0;
      IS_ZOMBIE = true;
      NUMBER    = 0;
      LEAFS     = 0;
      reset_octants();
      return *this;
    }
    //--------------------------------------------------------------------------
    void addleaf_to_list(leaf_type*L) {            // add leaf L to linked list 
      L->NEXT   = LEAFS;                           // add to linked list        
      LEAFS     = L;                               // update head of list       
      NUMBER ++;                                   // increment number          
    }
    //--------------------------------------------------------------------------
    void addleaf_to_octs(leaf_type*L) {            // add leaf L to octants     
      register uint b = octant(L);                 // find appropriate octant   
      OCT[b] = L;                                  // fill into octant          
      mark_as_leaf(b);                             // mark octant as leaf       
      NUMBER ++;                                   // increment number          
    }
    //--------------------------------------------------------------------------
    void addbox_to_octs(box*P) {                   // add box P to octants      
      register uint b = octant(P);                 // find appropriate octant   
      OCT[b] = P;                                  // fill into octant          
      mark_as_box(b);                              // mark octant as box        
      NUMBER += P->NUMBER;                         // increment number          
    }
    //--------------------------------------------------------------------------
    void shrink_to_octant(const uint i) {          // shrink to octant i        
      RADIUS *= half;                              // reduce radius             
      if(i&1) center()[0]+= RADIUS; else center()[0]-= RADIUS;
      if(i&2) center()[1]+= RADIUS; else center()[1]-= RADIUS;
#if NDIM==3
      if(i&4) center()[2]+= RADIUS; else center()[2]-= RADIUS;
#endif
    }
    //--------------------------------------------------------------------------
    uint  N_subboxes() {                           // count sub-boxes           
      register uint n=0;                           // index, counter            
      for(register int b=0; b!=NSUB; ++b)          // loop octants          >   
	if(TYPE & BIT[b]) n++;                     //   count boxes         <   
      return n;                                    // return counter            
    }
    //--------------------------------------------------------------------------
    // const friends                                                            
    //--------------------------------------------------------------------------
    friend vect      &center (      box*B) {return B->center();  }
    friend vect const&center (const box*B) {return B->center();  }
  };
#define PLEAF static_cast<leaf_type*>
#define PBOX  static_cast<box_type*>
  //----------------------------------------------------------------------------
  // class nbdy::builder_tools<CELL>                                            
  //----------------------------------------------------------------------------
  template<class CELL>
  class builder_tools :
    protected cell_access<CELL>
  {
    builder_tools           (const builder_tools&);// not implemented           
    builder_tools& operator=(const builder_tools&);// not implemented           
    //--------------------------------------------------------------------------
    // types of class builder_tools                                             
    //--------------------------------------------------------------------------
  protected:
    typedef box<cell_type>           box_type;     // type of associated box    
    typedef leaf<soul_type>          leaf_type;    // type of associated leaf   
    //--------------------------------------------------------------------------
    // data of class builder_tools                                              
    //--------------------------------------------------------------------------
    uint               NCRIT, NZOMBIES;            // Ncrit, # zombies          
    uint               DMAX,DEPTH;                 // max/actual tree depth     
    manager<box_type> *BM;                         // manager of boxes          
    //--------------------------------------------------------------------------
    // const methods                                                            
    //--------------------------------------------------------------------------
  public:
    uint const &N_zombies  () const { return NZOMBIES; }
    uint const &N_allocated() const { return BM->N_allocated(); }
    uint const &N_used     () const { return BM->N_used(); }
    uint        N_needed   () const { return N_used() - N_zombies(); }
    uint        N_free     () const { return N_allocated()-N_used(); }
    uint const &depth      () const { return DEPTH; }
    uint const &Ncrit      () const { return NCRIT; }
    //--------------------------------------------------------------------------
    // non-const methods                                                        
    //--------------------------------------------------------------------------
    // 1. protected non-const methods                                           
    //--------------------------------------------------------------------------
  protected:
    box_type* new_box    (const uint nl) {
      return &(BM->new_box(nl)->reset());          // get new box from manager  
    }
    //--------------------------------------------------------------------------
    box_type* make_subbox(const box_type*B, const uint i, const uint nl)
      // This routines provides a new box, whose center and radius are that of  
      // the ith octant of box B.                                               
    {
      register box_type*P = new_box(nl);           // get box off the stock     
      P->RADIUS   = B->RADIUS;                     // copy radius of parent     
      P->center() = B->center();                   // copy center of parent     
      P->shrink_to_octant(i);                      // shrink to correct octant  
      return P;                                    // return new box            
    }
    //--------------------------------------------------------------------------
    box_type* make_subbox(const box_type* B, const uint i, leaf_type* L,
			  const uint nl)
      // This routines provides a new box, whose center and radius are that of  
      // the ith octant of box B and which contains leaf L as first entry.      
    {
      register box_type*P = make_subbox(B,i,nl);   // make new sub-box          
      if(NCRIT>1) P->addleaf_to_list(L);           // add old leaf to list      
      else        P->addleaf_to_octs(L);           // or to appropriate octant  
      return P;                                    // return new box            
    }
    //--------------------------------------------------------------------------
    box_type* make_cellbox(const cell_type*C, const uint nl)
      // This routine makes a box and sets its center and radius to that of cell
    {
      register box_type* P = BM->new_box(nl);
      P->set_up(C);
      return P;
    }
    //--------------------------------------------------------------------------
    box_type* make_copybox(const box_type&Po, const uint nl) {
      // This routine makes a box and sets its data to that of box Po           
      return &((BM->new_box(nl))->operator= (Po));
    }
    //--------------------------------------------------------------------------
    void de_zombie(box_type* &B) const
      // If B is zombie, it is replaced by its first non-zombie descendant.     
    {
      register uint  i,s;                          // index, counter            
      register node *N,**P;                        // sub-node, octant          
      for(;;) {                                    // for ever                > 
	for(s=0,i=0,N=0,P=B->OCT; i<NSUB; i++,P++) //   loop octants of box B > 
	  if(*P) {                                 //     IF octant active    > 
	    if(s) return;                          //       IF not 1st: DONE    
	    s = 1;                                 //       ELSE: set s=1       
	    N = *P;                                //       and N=octant        
	  }                                        //   < <                     
	B = PBOX(N);                               //   set B to only sub-node  
      }                                            // <                         
    }
    //--------------------------------------------------------------------------
    // 2. constructors of class builder_tools                                   
    //--------------------------------------------------------------------------
    builder_tools(const uint nc, const uint dm) :
      NCRIT(nc), NZOMBIES(0), DMAX(dm) {}
  };
  //----------------------------------------------------------------------------
  // class nbdy::tree_builder<CELL>                                             
  //----------------------------------------------------------------------------
  template<class CELL>
  class tree_builder :
    public builder_tools<CELL>
  {
    tree_builder           (const tree_builder&);  // not implemented           
    tree_builder& operator=(const tree_builder&);  // not implemented           
    //--------------------------------------------------------------------------
    // basic types                                                              
    //--------------------------------------------------------------------------
    typedef typename CELL::bodies_type bodies_type;
    //--------------------------------------------------------------------------
    // data of class tree_builder                                               
    //--------------------------------------------------------------------------
    const bodies_type *BODIES;                     // bodies                    
    uint               NLEAFS;                     // # leafs                   
    leaf_type         *L0, *LN;                    // begin/end of leafs        
    vect               XMIN, XMAX;                 // extreme positions         
    //--------------------------------------------------------------------------
    // const methods                                                            
    //--------------------------------------------------------------------------
    bool is_twig(const box_type* B) const{
      return B->NUMBER <= NCRIT;                   // twig if not N > Ncrit     
    }
    //--------------------------------------------------------------------------
  public:
    uint const &N_leafs    () const { return NLEAFS; }
    //--------------------------------------------------------------------------
    // non-const methods                                                        
    //--------------------------------------------------------------------------
    // 1. private non-const methods                                             
    //--------------------------------------------------------------------------
  private:
    void split_box(      box_type*P,               // I: box to be splitted     
		   const uint     l,               // I: level                  
		   const uint     nl)              // I: # leafs added so far   
      // This routines splits a box:                                            
      // The leafs in the linked list are sorted into octants. Octants with one 
      // leaf will just hold that leaf, octants with many leaf will be boxes    
      // with the leaf in the linked list.                                      
      // If all leafs happen to be in just one octant, the process is repeated  
      // recursively on the box of this octant.                                 
    {
      if(l>DMAX) error("exceeding max depth");     // IF (level>DMAX) ERROR     
      register uint      NUM[NSUB];                // array with number of leafs
      register uint      b,ne;                     // octant & counters         
      register box_type *sub=0;                    // current sub-box           
      register leaf_type*Li,*Ln;                   // current & next leaf       
      do {                                         // DO until no zombie >      
	for(b=0; b<NSUB; b++) NUM[b] = 0;          //   reset counters          
	for(Li=P->LEAFS; Li; Li=Ln) {              //   loop linked list >      
	  Ln = Li->NEXT;                           //     next leaf in list     
	  b  = P->octant(Li);                      //     octant of current leaf
	  Li->NEXT = PLEAF(P->OCT[b]);             //     add leaf to list      
	  P->OCT[b]= Li;                           //     update head of list   
	  NUM[b] ++;                               //     increase number       
	}                                          //   <                       
	P->LEAFS = 0;                              //   reset list of sub-leafs 
	for(ne=b=0; b<NSUB; b++) if(NUM[b]) {      //   loop non-empty octants >
	  ne++;                                    //     count them            
	  if(NUM[b]>1) {                           //     IF(many leafs) >      
	    sub = make_subbox(P,b,nl);             //       make sub-box        
	    sub->LEAFS = PLEAF(P->OCT[b]);         //      assign sub-box's     
	    sub->NUMBER= NUM[b];                   //       leaf list & number  
	    P->OCT[b]= sub;                        //       set octant=sub-box  
	    P->mark_as_box(b);                     //       mark octant as box  
	  }                                        //     <                     
	}                                          //   <                       
	if(ne!=1) P->IS_ZOMBIE = false;            //   mark P as non-zombie    
	P = sub;                                   //   set current box=sub-box 
      } while(ne==1);                              // < while only 1 octant     
    }
    //--------------------------------------------------------------------------
    void addleaf(box_type  *base,                  // I:   add to this box      
		 leaf_type *Li,                    // I:   this leaf            
		 const uint nl,                    // I:   leafs added sofar    
		 const uint l0)                    // I:   level of base        
    {
      register uint       b,l;                     // octant, level             
      register leaf_type *Lo;                      // actual leaf loaded        
      register box_type  *P=base;;                 // actual box                
      register node     **oc;                      // actual sub-node           
      for(l=l0; l<DMAX; l++) {                     //   loop over boxes         
	if(NCRIT>1 && is_twig(P)) {                //   IF(box == twig) >       
	  P->addleaf_to_list(Li);                  //     add leaf to list      
	  if(P->NUMBER>NCRIT) split_box(P,l,nl);   //     IF(N > NCRIT) split   
	  return;                                  //     DONE with this leaf   
	} else {                                   //   < ELSE (box==branch) >  
	  b  = P->octant(Li);                      //     leaf's octant         
	  oc = P->OCT+b;                           //     pointer to octant     
	  P->NUMBER++;                             //     increment number      
	  if((*oc)==0) {                           //     IF octant empty >     
	    *oc = Li;                              //       assign leaf to it   
	    P->mark_as_leaf(b);                    //       mark octant as leaf 
	    P->IS_ZOMBIE = false;                  //       box cannot be zombie
	    return;                                //       DONE with this leaf 
	  } else if(P->marked_as_leaf(b)) {        //     < ELSE (octant=leaf) >
	    Lo = PLEAF(*oc);                       //       get old leaf        
	    P->mark_as_box(b);                     //       mark octant as box  
	    P = make_subbox(P,b,Lo,nl);            //       create sub-box      
	    *oc = P;                               //       assign sub-box to oc
	  } else                                   //     < ELSE (octant=box) > 
	    P  = PBOX(*oc);                        //       set current box     
	}                                          //     <                     
      }                                            //   <                       
      if(l>=DMAX) error("exceeding max depth");    //   IF (level>DMAX) ERROR   
    }
    //--------------------------------------------------------------------------
    void addleafs(box_type*base)                   // add to this               
    {
      register uint       nl=0;                    // counter                   
      register leaf_type *Li;                      // actual leaf loaded        
      for(Li=L0; Li<LN; Li++,nl++)                 // loop over all leafs >     
	addleaf(base,Li,nl,0);
    }
    //--------------------------------------------------------------------------
    template<bool U>
    uint link_cells(                               // link cells & souls        
		    box_type * P,                  // current box               
		    cell_type* C,                  // current cell              
		    cell_type*&Cf,                 // free cells                
		    soul_type*&Sf) const           // free souls                
      // RECURSIVE                                                              
      // This routines transforms the box-leaf tree into the cell-soul tree,    
      // such that all the cells that are contained within some cell are conti- 
      // guous in memory, as are the souls.                                     
    {
      register uint dep=0;                         // depth of cell             
      radius_(C) = P->RADIUS;                      // copy radius               
      center_(C) = P->center();                    // copy center               
      number_(C) = P->NUMBER;                      // copy number               
      csouls_(C) = Sf;                             // set cell: soul kids       
      nsouls_(C) = 0;                              // reset cell: # soul kids   
      if(is_twig(P)) {                             // IF(box==twig)         >   
	ccells_(C) = 0;                            //   set cell: sub-cells     
	ncells_(C) = 0;                            //   set cell: # sub-cells   
	register leaf_type*Li=P->LEAFS;            //   sub-leaf pointer        
	for(; Li; Li=Li->NEXT) {                   //   loop sub-leafs      >   
	  if(U) Li->set_soul(Sf++,BODIES);         //     set soul              
	  else  Li->set_soul(Sf++);                //     set soul              
	  nsouls_(C)++;                            //     increment # sub-souls 
	}                                          //   <                       
      } else {                                     // < ELSE (box==branch)  >   
	register uint   i;                         //   index to octant         
	register node **N;                         //   pointer to current sub  
	de_zombie(P);                              //   de-zombie               
	for(i=0,N=P->OCT; i!=NSUB; ++i,++N)        //   loop octants     >      
	  if(P->octant_is_leaf(i)) {               //     IF sub-leaf    >      
	    if(U) PLEAF(*N)->set_soul(Sf++,BODIES);//       set soul            
	    else  PLEAF(*N)->set_soul(Sf++);       //       set soul            
	    nsouls_(C)++;                          //       inc # sub-souls     
	  }                                        //   < <                     
	register uint de,nsub=P->N_subboxes();     //   number of sub-boxes     
	if(nsub) {                                 //   IF sub-boxes         >  
	  register box_type *Pi;                   //     sub-box pointer       
	  register cell_type*Ci=Cf;                //     remember free cells   
	  ncells_(C) = nsub;                       //     set cell: sub-cells   
	  ccells_(C) = Cf;                         //     set cell: sub-cells   
	  Cf += nsub;                              //     reserve nsub cells    
	  for(i=0,N=P->OCT; i!=NSUB; ++i,++N)      //     loop octants     >    
	    if(*N && P->marked_as_box(i)) {        //       IF sub-box     >    
	      Pi = PBOX(*N);                       //         sub-box           
	      de = link_cells<U>(Pi,Ci++,Cf,Sf);   //         recursive call    
	      if(de>dep) dep=de;                   //         update depth      
	    }                                      //     < <                   
	} else {                                   //   < ELSE (no sub-boxes) > 
	  ncells_(C) = 0;                          //     set cell: sub-cells   
	  ccells_(C) = 0;                          //     set cell: sub-cells   
	}                                          //   <                       
      }                                            // <                         
      dep++;                                       // increment depth           
      return dep;                                  // return cells depth        
    }
    //--------------------------------------------------------------------------
    // 2. public non-const methods                                              
    //--------------------------------------------------------------------------
  public:
    void build(const vect*pc)                      // I: center for root        
      // This routine simply calls the above with the root box as argument.     
    {
      register box_type *P= new_box(1);            // get a new box: root       
      P->center()= pc? *pc:RootCenter(L0,LN);      // set root: center          
      P->RADIUS  = RootRadius(P->center(),XMIN,XMAX);// set root: radius        
      addleafs(P);                                 // call above, box=root      
      register uint nzom=0u;                       // zombie counter            
      for(register typename manager<box_type>::iterator // loop boxes           
	    I=BM->begin(); I!=BM->end(); ++I)      //   (using iterator)   >    
	if(!is_twig(I) && I->IS_ZOMBIE) ++nzom;    //   count zombie cells <    
      NZOMBIES = nzom;                             // reset # zombie cells      
    }
    //--------------------------------------------------------------------------
    void link(cell_type*C0, soul_type*Sf) {
      register cell_type* Cf=C0+1;
      if(BODIES) DEPTH = link_cells<true> (BM->root(),C0,Cf,Sf);
      else       DEPTH = link_cells<false>(BM->root(),C0,Cf,Sf);
    }
    //--------------------------------------------------------------------------
    // constructors of class tree_builder                                       
    //--------------------------------------------------------------------------
    // 1 for the use with bodies                                                
    //--------------------------------------------------------------------------
    tree_builder(const uint nc, const uint dm, const bodies_type* const& b)
      : builder_tools<CELL>(nc,dm), BODIES(b)
    {
      L0 = new leaf_type[BODIES->N_body()];        // allocate leafs            
      register leaf_type*Li=L0;                    // current leaf              
      XMAX = XMIN = pos(BODIES->begin_bodies());   // reset x_min & x_max       
      LoopBodies(typename bodies_type,BODIES,Bi)   // loop bodies               
	if(is_in_tree(Bi)) {                       //   IF(body is active)   >  
	  Li->set_up(Bi);                          //     initialize leaf       
	  if     (Li->pos(0) > XMAX[0]) XMAX[0]=Li->pos(0);
	  else if(Li->pos(0) < XMIN[0]) XMIN[0]=Li->pos(0);
	  if     (Li->pos(1) > XMAX[1]) XMAX[1]=Li->pos(1);
	  else if(Li->pos(1) < XMIN[1]) XMIN[1]=Li->pos(1);
#if NDIM==3
	  if     (Li->pos(2) > XMAX[2]) XMAX[2]=Li->pos(2);
	  else if(Li->pos(2) < XMIN[2]) XMIN[2]=Li->pos(2);
#endif
	  Li++;                                    //     increment current leaf
	}                                          //   <                       
      LN     = Li;                                 // set: beyond last leaf     
      NLEAFS = int(LN-L0);                         // set: # active leafs       
      BM     = new manager<box_type>(1+NLEAFS/4,   // initialize box manager    
				     NLEAFS);      // first allocate N/4 boxes  
    }
    //--------------------------------------------------------------------------
    // 2 for the use with arrays                                                
    //--------------------------------------------------------------------------
    tree_builder(const uint nc,const uint dm,
		 const areal*x[NDIM], const int*f, const uint nb)
      : builder_tools<CELL>(nc,dm), BODIES(0)
    {
      register uint i;                             // index                     
      L0 = new leaf_type[nb];                      // allocate leafs            
      register leaf_type*Li=L0;                    // current          leaf     
      LoopDims XMAX[d] = XMIN[d] = x[d][0];        //   reset x_min & x_max <   
      for(i=0; i<nb; i++) if(is_in_tree(f[i])) {   // loop active bodies    >   
	Li->set_up(x,i);                           //   initialize leaf         
	if     (Li->pos(0) > XMAX[0]) XMAX[0]=Li->pos(0);
	else if(Li->pos(0) < XMIN[0]) XMIN[0]=Li->pos(0);
	if     (Li->pos(1) > XMAX[1]) XMAX[1]=Li->pos(1);
	else if(Li->pos(1) < XMIN[1]) XMIN[1]=Li->pos(1);
#if NDIM==3
	if     (Li->pos(2) > XMAX[2]) XMAX[2]=Li->pos(2);
	else if(Li->pos(2) < XMIN[2]) XMIN[2]=Li->pos(2);
#endif
	Li++;                                      //   increment current leaf  
      }                                            // <                         
      LN     = Li;                                 // set: beyond last leaf     
      NLEAFS = int(LN-L0);                         // set: # active leafs       
      BM     = new manager<box_type>(1+NLEAFS/4,   // initialize box manager    
				     NLEAFS);      // first allocate N/4 boxes  
    }
    //--------------------------------------------------------------------------
    // destructor                                                               
    //--------------------------------------------------------------------------
    ~tree_builder()  {
      delete[] L0;                                 // de-allocate leafs         
      delete   BM;                                 // de-allocate boxes         
    }
    //--------------------------------------------------------------------------
  };
#undef LoopDims
#undef PLEAF
#undef PBOX
}
using namespace nbdy;
////////////////////////////////////////////////////////////////////////////////
// class nbdy::basic_tree<CELL>                                                 
////////////////////////////////////////////////////////////////////////////////
#define LoopMyCellsDown							\
  for(register cell_type*Ci= begin_cells(); Ci!= end_cells(); ++Ci)
#define LoopMyCellsUp							\
  for(register cell_type*Ci=rbegin_cells(); Ci!=rend_cells(); --Ci)
//------------------------------------------------------------------------------
template<class CELL> void
basic_tree<CELL>::mark_for_subtree(                // sub-tree mark my nodes    
				   uint &Nsubc,    // O: # subtree cells        
				   uint &Nsubs)    // O: # subtree souls        
{
  register int  nsun,nsuc;                         // # subt nodes per cell     
  register uint ns=0u, nc=0u;                      // # subt nodes in total     
  LoopMyCellsUp {                                  // loop tree cells backward  
    unflag_subtree_flags(Ci);                      //   reset subtree flags     
    nsun=0;                                        //   counter: subt subnodes  
    nsuc=0;                                        //   counter: subt subcells  
    LoopSoulKids(typename cell_type,Ci,s)          //   loop child souls       >
      if(in_subtree(s)) ++nsun;                    //     count subt souls     <
    ns += nsun;                                    //   count total subt souls  
    LoopCellKids(typename cell_type,Ci,c)          //   loop child cells       >
      if(in_subtree(c)) {                          //     IF cell is in subt   >
	++nsun;                                    //       count subt nodes    
	if(is_subtreecell(c) ||                    //       IF is subt cell     
	   is_subtreezombie(c) )                   //       OR is subt zombie  >
	  ++nsuc;                                  //         count subcells   <
      }                                            //   < <                     
    if(nsun > 0) {                                 //   IF cell is subt node   >
      flag_for_subtree(Ci);                        //     flag it as such       
      if(nsun > 1) {                               //     IF tcell subt cell   >
	flag_as_subtreecell(Ci);                   //       flag it as such     
	++nc;                                      //       count subt cells    
      } else if(nsuc > 0)                          //     < ELSE tcell=zombie  >
	flag_as_subtreezombie(Ci);                 //       so flag it as such <
    }                                              //   <                       
  }                                                // <                         
  Nsubs = ns;                                      // # subt souls              
  Nsubc = nc;                                      // # subt cells              
}
#undef LoopMyCellsDown
#undef LoopMyCellsUp
////////////////////////////////////////////////////////////////////////////////
// class nbdy::tree<CELL>                                                       
////////////////////////////////////////////////////////////////////////////////
template<class CELL>
tree<CELL>::tree(const bodies_type*BB,             // I: array with bodies      
		 const vect       *pc,             // [I: root center]          
		 const unsigned    nc,             // [I: N_crit]               
		 const unsigned    dm) :           // [I: max tree depth]       
  basic_tree<CELL>(BB)
{
  tree_builder<CELL> PM(nc,dm,BB);                 // initiliaze tree_builder   
  if(PM.N_leafs() == 0) {                          // IF(no leaf in tree) >     
    warning(" no bodies in tree");                 //   issue a warning         
    return;                                        //   and return undone       
  }                                                // <                         
  PM.build(pc);                                    // build box-leaf tree       
  Nc = PM.N_needed();                              // # cells = # active boxes  
  C0 = new cell_type[Nc];                          // allocate cells            
  CN = C0 + Nc;                                    // pointer past last cell    
  Ns = PM.N_leafs();                               // # souls = # active bodies 
  S0 = new soul_type[Ns];                          // allocate souls            
  SN = S0+Ns;                                      // pointer past last soul    
  PM.link(C0,S0);                                  // link box-leaf <> cell-soul
  Nz = PM.N_zombies();                             // # zombie cells            
  Dp = PM.depth();                                 // depth of tree             
}  
//------------------------------------------------------------------------------
template<class CELL>
tree<CELL>::tree(const flag_type*FL,               // I: array: flags           
		 const areal    *X[NDIM],          // I: arrays: x,y,z          
		 const unsigned  Nb,               // I: size of arrays         
		 const vect     *pc,               // [I: root center]          
		 const unsigned  nc,               // [I: N_crit]               
		 const unsigned  dm) :             // [I: max tree depth]       
  basic_tree<CELL>(0)
{
  tree_builder<CELL> PM(nc,dm,X,FL,Nb);            // initiliaze tree_builder   
  if(PM.N_leafs() == 0) {                          // IF(no leaf in tree) >     
    warning(" no bodies in tree");                 //   issue a warning         
    return;                                        //   and return undone       
  }                                                // <                         
  PM.build(pc);                                    // build box-leaf tree       
  Nc = PM.N_needed();                              // # cells = # active boxes  
  C0 = new cell_type[Nc];                          // allocate cells            
  CN = C0 + Nc;                                    // pointer past last cell    
  Ns = PM.N_leafs();                               // # souls = # active bodies 
  S0 = new soul_type[Ns];                          // allocate souls            
  SN = S0+Ns;                                      // pointer past last soul    
  PM.link(C0,S0);                                  // link box-leaf <> cell-soul
  Nz = PM.N_zombies();                             // # zombie cells            
  Dp = PM.depth();                                 // depth of tree             
}
////////////////////////////////////////////////////////////////////////////////
// class nbdy::sub_tree<CELL,SOUL,PCELL,PSOUL>                                  
////////////////////////////////////////////////////////////////////////////////
template<class SUB_CELL, class PARENT_CELL>        // types for parent tree     
sub_tree<SUB_CELL,PARENT_CELL>::
sub_tree(basic_tree<PARENT_CELL> *parent) :
  basic_tree<SUB_CELL>(parent->my_bodies())
{
  parent->mark_for_subtree(Nc,Ns);                 // mark parent tree          
  if(Ns==0 || Nc==0) {                             // IF no nodes marked    >   
    warning("empty subtree");                      //   issue warning and       
    return;                                        //   return                  
  }                                                // <                         
  S0 = new soul_type[Ns];                          // allocate subtree cells    
  SN = S0+Ns;                                      // set end of cells          
  C0 = new cell_type[Nc];                          // allocate subtree souls    
  CN = C0+Nc;                                      // set end of souls          
  register soul_type *Sf=S0;                       // pter to 1st free soul     
  register cell_type *Cf=C0+1;                     // pter to 1st free cell     
  Dp = sub_tree_builder<SUB_CELL,PARENT_CELL>::
    link(parent->root(),C0,Cf,Sf);                 // link sub-tree             
}
////////////////////////////////////////////////////////////////////////////////
// specialiazations                                                             
////////////////////////////////////////////////////////////////////////////////
#include <nbdy/soul.h>
#include <nbdy/cell.h>
#include <nbdy/stic.h>

template tree<cell>;
template sub_tree<sticky_cell,cell>;
////////////////////////////////////////////////////////////////////////////////
