//-----------------------------------------------------------------------------+
//                                                                             |
// yanc.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//                        YET ANOTHER N-BODY CODE                              |
//                                                                             |
//-----------------------------------------------------------------------------+

#include <yanc.h>
#include <iostream>
#include <iomanip>

extern "C" {
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <pwd.h>
}

#include <nbdy.h>
#include <nbdy/nmio.h>
#include <util/ionl.h>
using util::open;
using util::SwallowRestofLine;

using std::string;
using std::ios;
using std::endl;
using std::cerr;
using std::istream;
using std::ostream;
using std::ifstream;
using std::ofstream;
using std::setw;
using std::setfill;
using std::setprecision;

using namespace nbdy;
////////////////////////////////////////////////////////////////////////////////
namespace nbdy{
  //===========================================================================+
  //                                                                           |
  // auxiliary constants and functions for the installation of class ndata     |
  //                                                                           |
  //============================================================================
  inline basic_nbody::soft_type soften(const indx SOFTEN) {
    switch(SOFTEN) {
    case 1:  return basic_nbody::individual_fixed;
    case 2:  return basic_nbody::global_fixed;
    default: return basic_nbody::global_fixed; 
    }
  }
  //----------------------------------------------------------------------------
  inline nbdy::kern_type kern(const indx KERN) { 
    switch(KERN) { 
    case  0: return nbdy::f0;
    case  1: return nbdy::f1;
    case  2: return nbdy::f2;
    case  3: return nbdy::f3;
    case 10: return nbdy::k0;
    case 11: return nbdy::k1;
    case 12: return nbdy::k2;
    case 13: return nbdy::k3;
    case 20: return nbdy::p0;
    case 21: return nbdy::p1;
    case 22: return nbdy::p2;
    case 23: return nbdy::p3;
    default: warning(" kernel unknown, defaulting to Newtonian\n");
      return nbdy::newton;
    }
  }
  //----------------------------------------------------------------------------
  const char* opts[] = {"start_data",      //  0
			"theta",           //  1
			"Nreuse",          //  2
			"Ncrit",           //  3
			"root_center",     //  4
			"reuse",           //  5
			"softening",       //  6
			"Nsoft",           //  7
			"eps",             //  8
			"eps_given",       //  9
			"kernel",          // 10
			"scheme",          // 11
			"Nref",            // 12
			"fac",             // 13
			"tq",              // 14
			"time",            // 15
			"hmin",            // 16
			"Nsteps",          // 17
			"give_pot",        // 18
			"give_rho",        // 19
			"data_format",     // 20
			"N",               // 21
			"##","#",          // 22, 23
			"data_file"};      // 24
  const unsigned nopt=25;
  //----------------------------------------------------------------------------
  inline void yancwarning(const char*msgs) {
    cerr<<"YANC: WARNING: "<<msgs<<"\n";
  }
  //----------------------------------------------------------------------------
  inline void yancerror  (const char*msgs) {
    cerr<<"YANC: ERROR: "<<msgs<<"\n";
    exit(1);
  }
  //----------------------------------------------------------------------------
  inline void nemowarning(const char*msgs) {
    cerr<<"YancNemo: WARNING: "<<msgs<<"\n";
  }
  //----------------------------------------------------------------------------
  inline void nemoerror  (const char*msgs) {
    cerr<<"YancNemo: ERROR: "<<msgs<<"\n";
    exit(1);
  }
  //----------------------------------------------------------------------------
  inline istream& skip_char(istream& i, const char c) {
    char x;
    i>>x;
    if(x==c) return i;
    return i.putback(x);
  }
  //----------------------------------------------------------------------------
  inline istream& skip_line(istream& i, const char c) {
    char x;
    i>>x;
    if(x==c) { SwallowRestofLine(i); return i; }
    else       return i.putback(x);
  }
  //----------------------------------------------------------------------------
  inline nbdy::uint check_eps_range(const real EPS,
				    nbdy::bodies *BODIES)
  {
    register nbdy::uint alright=0u;
    LoopBodies(bodies,BODIES,Bi)
      if(nbdy::eps(Bi) > EPS) {                   // eps out of range ?
	Bi.eps() = EPS;                               // put it in range
	alright++;
      }
    return alright;
  }
  //===========================================================================+
  //                                                                           |
  // class nbdy::ndata                                                         |
  //                                                                           |
  // auxiliary for the installation of class yanc                              |
  //                                                                           |
  //============================================================================
  class ndata {
#ifdef ALLOW_NEMO
    nemo_output *NEMO_O;                     // NEMO output device              
  public:
    real         CPU_I;                      // initial CPU time                
#endif
  public:
    string  FILE;                            // input file                      
    real    THETA;                           // opening angle                   
    uint    NREUSE;                          // max # subsequent falcON::reuse()
    uint    NCRIT;                           // max # bodies in unsplit cell    
    vect    OR, *POR;                        // center of root                  
    uint    NSOFT;                           // # bodies in softening spheres   
    uint    NREF;                            // # bodies for density estimate   
    real    EPS;                             // eps / eps_max                   
    indx    SOFTEN;                          // 0/1/2 for globally fixed,       
                                             // indiv fixed, and indiv adaptive 
    indx    KERN;                            // (0/10/20)+(0/1/2/3) soft kernel 
    real    FAC;                             // factor in time-step scheme      
    real    TINI;                            // initial simulation time         
    real    TIME;                            // simulation time                 
    bool    EPS_GIVEN;                       // eps_i are given on input        
    add_out ADD_OUT;                         // what to give on output          
    int     HMIN;                            // tau_min = 1/2^h0                
    indx    NSTEPS;                          // number of tau                   
    indx    FORMAT;                          // data format: 0/1 : ascii/binary 
    bodies *BODIES;                          // bodies themselves               
  private:
    ndata();                                 // not implemented                 
    ndata(const ndata&);                     // not implemented                 
    //--------------------------------------------------------------------------
  private:
    void read_yanc        ()                     // read input using yanc format
    {
      ifstream in;
      if(!open(in,FILE)) exit(1);
      char option[100];
      register bool tmp_bool;
      register int  tmp_int;
      register uint iopt = 0;
      register uint Ns,Nbody;
      do {
	if(in)
	  skip_char(in,'#') >> option;
	else {
	  cerr<<"# YANC: end of input seen before \"start_data\"\n";
	  return;
	}
	for(iopt=0; iopt<nopt && strcmp(option,opts[iopt]); iopt++);
	switch(iopt) {
	case  0:                         SwallowRestofLine(in); break;
	case  1: in >> THETA;            SwallowRestofLine(in); break;
	case  2: in >> NREUSE;           SwallowRestofLine(in); break;
	case  3: in >> NCRIT;            SwallowRestofLine(in); break;
	case  4: in >> OR; POR = &OR;    SwallowRestofLine(in); break;
	case  5: cerr<<"# YANC: parameter \"reuse\" currently not supported\n";
	  SwallowRestofLine(in); break;
	case  6: in >> SOFTEN;           SwallowRestofLine(in); break;
	case  7: in >> NSOFT;            SwallowRestofLine(in); break;
	case  8: in >> EPS;              SwallowRestofLine(in); break;
	case  9: EPS_GIVEN = true;       SwallowRestofLine(in); break;
	case 10: in >> KERN;             SwallowRestofLine(in); break;
	case 11: cerr<<"# YANC: parameter \"scheme\" currently not supported\n";
	  SwallowRestofLine(in); break;
	case 12: in >> NREF;             SwallowRestofLine(in); break;
	case 13: in >> FAC;              SwallowRestofLine(in); break;
	case 14: cerr<<"# YANC: parameter \"tq\" currently not supported\n";
	  SwallowRestofLine(in); break;
	case 15: in >> TIME;             SwallowRestofLine(in); break;
	case 16: in >> HMIN;             SwallowRestofLine(in); break;
	case 17: in >> Ns;               SwallowRestofLine(in);
	         if(Ns>=1) NSTEPS = Ns;
	         else cerr<<"# YANC: Nsteps>0 required\n";      break;
	case 18: in >> tmp_bool;
	         if(tmp_bool) ADD_OUT |= add_out(pots);
	                                 SwallowRestofLine(in); break;
	case 19: in >> tmp_int;
	         switch(tmp_int) {
		 case 0: break;
		 case 1: ADD_OUT |= rhos; break;
		 case 2: ADD_OUT |= surf; break;
		 default: cerr<<"# YANC: unregocnized value for give_rho\n";
		 }                       SwallowRestofLine(in); break;
	case 20: in >> FORMAT;           SwallowRestofLine(in); break;
	case 21: in >> Nbody;            SwallowRestofLine(in); break;
	case 22: case 23:                SwallowRestofLine(in); break;
	case 24: in >> FILE;             read_yanc();           return;
	default: cerr<<"# YANC: unknown parameter \""<<option<<"\" ignored\n";
	  SwallowRestofLine(in); break;
	}
      } while(iopt);
      if(BODIES) BODIES->reset(Nbody);
      else       BODIES = new bodies(Nbody);
      if(FORMAT == 0) BODIES->read_ascii (in,EPS_GIVEN);
      else            BODIES->read_binary(in);
      if(! in) {
	cerr<<"# YANC: too few bodies in input\n";
	delete BODIES;
	BODIES = 0;
	return;
      }
      if(SOFTEN && EPS_GIVEN) {
	uint a = check_eps_range(EPS,BODIES);
	if(NSOFT>0u && a>0) cerr<<"# YANC: "<<a<<" eps_i > eps_max\n";
      }
    }
    //--------------------------------------------------------------------------
  private:
    void write_yanc_header(                       // write header in yanc format
			   ostream   &out,        // I/O:  stream to write to   
			   const char*file,       // I:    name of file written 
			   const bool eps_i,      // I:    add param eps_given  
			   const char*prog,       // I:    name of calling main 
			   const indx format)     // I:    format type          
      const 
    {
      char host[100]; gethostname(host,100);
      time_t now = ::time(0);
      out.setf(ios::left);
      out<<setfill(' ');
      out<<"## -------------------------------"
	"-----------------------------------\n"
	 <<"##  file          \""<<file<<"\"\n"
	 <<"##  generated on  "  <<ctime(&now);
      if(prog) out<<"##  with program  \""<<prog<<"\"\n";
      out<<"##  by user       \""<<(getpwuid(geteuid())->pw_name)<<"\"\n"
	 <<"##  on machine    \""<<host<<"\"\n";
      if(!FILE.empty())
	out<<"##\n"
	   <<"##  original data read from file \""<<FILE<<"\"\n";
      out<<"##  initial time = "<<TINI<<"\n"
	 <<"## -------------------------------"
	"-----------------------------------\n";
      // tree issues
      out<<"# theta          "<<setw(15)<<THETA;
      if(THETA<zero) out<<" constant opening angle\n";
      else           out<<" mass-dependent opening angle; theta_min\n";
      out<<" Nreuse          "<<setw(15)<<NREUSE
	 <<" max # of subsequend tree reuses\n"
	 <<" Ncrit           "<<setw(15)<<NCRIT
	 <<" max # of bodies in un-splitted cell\n";
      if(POR) out<<"# root_center    "<<setw(4)<<OR[0]<<" "<<setw(4)<<OR[1]<<" "
		 <<setw(4)<<OR[2]<<"  use fixed root center\n";
      // softening issues
      out<<"# softening      "<<setw(15)<<SOFTEN;
      switch(SOFTEN) {
      case 1:
	out<<" individually fixed softening lengths\n";
	break;
      case 2:
	out<<"individually adaptive softening lengths with\n"
	   <<"# Nsoft          "<<setw(15)<<NSOFT
	   <<" # bodies in softening sphere\n"
	   <<"# eps            "<<setw(15)<<EPS
	   <<" maximum eps_i\n";
	break;
      default:
	out<<"globally fixed softening length:\n"
	   <<"# eps            "<<setw(16)<<EPS<<"\n";
      }
      // time-stepping issues
      if(NSTEPS==1) 
	out<<"# Nsteps         1                 (ordinary leap-frog)\n"
	   <<"# hmin           "<<setw(15)<<HMIN<<" with step tau "
	   <<tau_min()<<"\n";
      else {
	out<<"# fac            "<<setw(15)<<FAC<<" tau < "<<FAC<<"  / rho\n";
	out<<"# hmin           "<<setw(15)<<HMIN<<" tau_min = 2^"<<(-HMIN)
	   <<" = "<<tau_min()<<"\n";
	out<<"# Nsteps         "<<setw(15)<<NSTEPS<<" tau_max = 2^"
	   <<(NSTEPS-HMIN-1)
	   <<" = "<<tau_max()<<"\n";
      }
      // data issues
      out<<"# time           "<<setw(15)<<TIME<<" simulation time\n"
	 <<"# N              "<<setw(15)<<BODIES->N_body()<<" number of bodies\n"
	 <<"# data_format    "<<setw(15)<<format;
      if(format==0) out<<" bodies in yanc ascii format\n";
      else          out<<" bodies in yanc binary format\n";
      if(format==0) {
	if(ADD_OUT & pots)
	  out<<"# give_pot       1                 potential given on output\n";
	if(ADD_OUT & rhos)
	  out<<"# give_rho       1                 density given on output\n";
	else if(ADD_OUT & surf)
	  out<<"# give_rho       2                 surface density given on output\n";
      }
      if(eps_i)
	out<<"# eps_given                        eps_i given with data below\n";
      out<<"## -----------------------------------"
	"--------------------------------\n"
	 <<"# start_data"<<endl;
    }
    //--------------------------------------------------------------------------
#ifdef ALLOW_NEMO
    //--------------------------------------------------------------------------
  public: void open_nemo(const char *file, const bool resume=false)
    {
      if(NEMO_O) { delete NEMO_O; NEMO_O = 0; }
      if(resume) {
	if(file==0 || FILE == string(file))
	  NEMO_O = new nemo_output(FILE.c_str(),"a");
	else {
	  NEMO_O = new nemo_output(file);
	  NEMO_O->write_history();
	}
      } else if(file) {
	if(FILE == string(file)) nemoerror("out==in; use option resume instead");
	NEMO_O = new nemo_output(file);
	NEMO_O->write_history();
      } else
	nemowarning("cannot open nemo output");
    }
    //--------------------------------------------------------------------------
  public: void close_nemo() {
      delete NEMO_O;
      NEMO_O = 0;
    }
    //--------------------------------------------------------------------------
  public: bool nemo_is_open() const { return NEMO_O!=0 && NEMO_O->is_open(); }
    //--------------------------------------------------------------------------
  public: void read_nemo(const bool resume=false) {// read from NEMO snapshot   
      if(!BODIES) BODIES = new bodies();           // initialize bodies         
      nemo_input NEMO(FILE.c_str());               // open nemo input           
      NEMO.read_history();                         // read history              
      if(resume) {                                 // IF (resuming old sim)    >
	do {                                       //   do: read particles     >
	  NEMO.open_set(nemo_io::snap);            //     open nemo snapshot    
	  BODIES->read_nemo_particles(NEMO,&TIME); //     read particles        
	  if(NEMO.is_present(nemo_io::diags)) {    //     IF(diags set)      >  
	    NEMO.open_set(nemo_io::diags);         //       open diags set      
	    if(NEMO.is_present(nemo_io::cputime))  //       IF(cpu time there)  
	      CPU_I=NEMO.read(nemo_io::cputime)*60;//         read CPU [min]    
	    NEMO.close_set(nemo_io::diags);        //       close diags set     
	  }                                        //     <                     
	  NEMO.close_set(nemo_io::snap);           //     close nemo snapshot   
	} while(NEMO.is_present(nemo_io::snap));   //   < while more to be read 
      } else {                                     // < ELSE (not resuming) >   
	  NEMO.open_set(nemo_io::snap);            //   open nemo snapshot      
	  BODIES->read_nemo_particles(NEMO,&TIME); //     read particles        
	  NEMO.close_set(nemo_io::snap);           //   close nemo snapshot     
      }
    }
    //--------------------------------------------------------------------------
  public:
    void write_nemo(const basic_nbody* NBODY=0)    // [I]: nbody for diags O    
    {
      if(!NEMO_O) {
	nemoerror("no nemo output stream opened; cannot write snapshot");
	return;
      }
      if(!BODIES) {
	nemowarning("no body exist; cannot write snapshot");
	return;
      }
      register uint        i,j;
      NEMO_O->open_set(nemo_io::snap);              // open a new nemo snapshot 
      BODIES->write_nemo_particles(*NEMO_O,         //   write out particles    
				   SOFTEN? ADD_OUT | epsi : ADD_OUT, &TIME);
      if(NBODY) {                                   //   IF diags output wantd >
	NEMO_O->open_set(nemo_io::diags);           //     open diagnostics set 
	NEMO_O->single_vec(0)=NBODY->total_energy();//       copy total energy  
	NEMO_O->single_vec(1)=NBODY->kin_energy();  //       copy kin energy    
	NEMO_O->single_vec(2)=NBODY->pot_energy();  //       copy pot energy    
	NEMO_O->write(nemo_io::energy);             //       write energies     
	for(i=0;i!=NDIM;++i) for(j=0;j!=NDIM;++j)   //     loop dims twice  > > 
	  NEMO_O->single_mat(i,j) = NBODY->kin_energy(i,j);//  copy K_ij    < < 
	NEMO_O->write(nemo_io::KinT);               //     write K_ij           
	for(i=0;i!=NDIM;++i) for(j=0;j!=NDIM;++j)   //     loop dims twice  > > 
	  NEMO_O->single_mat(i,j) = NBODY->pot_energy(i,j);//  copy W_ij    < < 
	NEMO_O->write(nemo_io::PotT);               //     write W_ij           
	for(i=0;i!=NDIM;++i) for(j=0;j!=NDIM;++j)   //     loop dimensions  > > 
	  NEMO_O->single_mat(i,j) = NBODY->as_angmom(i,j); //  copy A_ij    < < 
	NEMO_O->write(nemo_io::AmT);                //     write A_ij           
	for(i=0;i!=NDIM;++i) {                      //     loop dims           <
	  NEMO_O->single_phs(0,i)=NBODY->center_of_mass()[i];// copy c-of-m pos 
	  NEMO_O->single_phs(1,i)=NBODY->total_momentum()[i];// copy c-of-m vel 
	}                                           //     <                    
	NEMO_O->write(nemo_io::cofm);               //     write c-of-m (x,v)   
	NEMO_O->write(nemo_io::cputime,             //     write accum CPU [min]
		      NBODY->cpu_total()/60.);      //                          
	NEMO_O->close_set(nemo_io::diags);          //     close diagnostics set
      }                                             //   <                      
      NEMO_O->close_set(nemo_io::snap);             // close nemo snapshot      
      NEMO_O->reset();                              // reset nemo output        
    }
    //--------------------------------------------------------------------------
  public:
    ndata(const char   *file,
	  const bool    nemo,
	  const double  th,
	  const int     nu,
	  const int     nc,
	  const double  ep,
	  const int     kn,
	  const int     hm,
	  const add_out ao,
	  const int     ns,
	  const int     nr,
	  const double  fc,
	  const bool    resume=false)     	    // only applies if nemo=true
      :
      NEMO_O    (0),
      CPU_I     (zero),
      FILE      (file),
      THETA     (real(th)),
      NREUSE    (uint(nu)),
      NCRIT     (uint(nc)),
      POR       (0),
      NSOFT     (0u),
      NREF      (uint(nr)),
      EPS       (real(ep)),
      SOFTEN    (0),
      KERN      (indx(kn)),
      FAC       (real(fc)),
      TIME      (zero),
      EPS_GIVEN (false),
      ADD_OUT   (ao),
      HMIN      (hm),
      NSTEPS    (indx(ns)),
      FORMAT    (0),
      BODIES    (0)
    {
      if(nemo) read_nemo(resume);
      else     read_yanc();
      TINI = TIME;
    }
#endif
    //--------------------------------------------------------------------------
  public:
    ndata(const char *file) :
      FILE      (file),
      THETA     (half),
      NREUSE    (3u),
      NCRIT     (6u),
      POR       (0),
      NSOFT     (0u),
      NREF      (16),
      EPS       (zero),
      SOFTEN    (0),
      KERN      (21),
      FAC       (0.001),
      TIME      (zero),
      EPS_GIVEN (false),
      ADD_OUT   (none),
      HMIN      (6),
      NSTEPS    (indx(1)),
      FORMAT    (0),
      BODIES    (0)
    {
      read_yanc();
      TINI = TIME;
    }
    //--------------------------------------------------------------------------
  public:  ~ndata()
    {
#ifdef ALLOW_NEMO
      close_nemo();
#endif
      if(BODIES) delete BODIES;
    }
    //--------------------------------------------------------------------------
  public:
    void describe(                                 // describe parameter        
		  ostream   & out,                 // I: output stream          
		  const char* name,                // I: name of main           
		  const char* output)              // I: base name: output files
    {
      if(!okay()) {
	out<<"#  ERROR occured during input of parameters and bodies\n";
	return;
      }
      out<<
	"# =======================================================================";
      if(name) out<<"\n# \""<<name<<"\"";
      out<<"\n#  parameters:"
	// tree issues
	 <<"\n#   theta                "<<THETA
	 <<"\n#   Nreuse               "<<NREUSE
	 <<"\n#   Ncrit                "<<NCRIT;
      if(POR) out<<"\n#   center of root       "<<(*POR);
      // softening issues
      switch(SOFTEN) {
      case 1:
	out<<"\n#   softening            individually fixed";
	if(EPS_GIVEN) out<<"\n#   eps_i                given with data";
	else          out<<"\n#   eps_i                to be initialized using"
			 <<"\n#   Nsoft                "<<NSOFT
			 <<"\n#   eps_max              "<<EPS;
	break;
      case 2:
	out<<"\n#   softening            individually adaptive"
	   <<"\n#   Nsoft                "<<NSOFT
	   <<"\n#   eps_max              "<<EPS;
	break;
      default:
	out<<"\n#   global eps           "<<EPS;
	break;
      }
      out<<"\n#   kernel               "<<falcON::describe(kern(KERN));
      // time-stepping issues
      if(NSTEPS==1)
	out<<"\n#   time step            "<<tau_min();
      else
	out<<"\n#   adaptive time steps: tau < "<<FAC<<" / rho"
	   <<"\n#   tau_min              "<<tau_min()
	   <<"\n#   tau_max              "<<tau_max();
      // data issues
      out<<"\n#   time                 "<<TIME
	 <<"\n#   N                    "<<BODIES->N_body()
	 <<"\n#   input file           "<<FILE;
      if(output)
	out<<"\n#   output files         "<<output<<".XXXX";
      if(FORMAT)  out<<"\n#   data format          yanc binary";
      else        out<<"\n#   data format          yanc ascii";
      if(ADD_OUT & pots) out<<"\n#   potential            written on output";
      if(ADD_OUT & rhos) out<<"\n#   density              written on output";
      if(ADD_OUT & surf) out<<"\n#   surface density      written on output";
      out<<endl;
    }
    //--------------------------------------------------------------------------
  public:
    static void describe_yanc_format(ostream& o)  // descripe yanc format
    {
      o<<
	" Format of YANC I/O files\n"
	" ========================\n"
	" The file starts with a header with one line per parameter.  The syntax\n"
	" of these lines is \"[#] PARAMETER VALUE(S) COMMENT\", where PARAMETER is\n"
	" a character string, see below, while  VALUE is the  numerical value of\n"
	" of the associated parameter. COMMENT is anything between VALUE and EOL\n"
	" and is ignored, as is an optional \"#\" at  the beginning of the  header\n"
	" lines. The meaning of the various parameters is as follows.\n\n"

	" PARAMETER   VALUE default meaning\n"
	" ---------------------------------------------------------------------\n"
	" ##                      this line is (quietly) ignored\n"
	" data_file     FILE      ignore rest of file and use FILE instead\n"
	" theta           th 0.5  opening angle  < 0 -> th=|VALUE| = const\n"
	"                                        > 0 -> th=th(M), th_min=TH\n"
	" Nreuse          Nr   3  max # subsequent tree re-uses\n"
	" Ncrit           Nc   6  max # bodies in un-splitted cell\n"
	" root_center  X,Y,Z      if specified: center of root cell\n"
	" softening        S   0  determines usage of softening lengths\n"
	"                          0 -> globally fixed\n"
	"                          1 -> individually fixed (NOT TESTED)\n"
	"                          2 -> individually adapted (NOT TESTED)\n"
	" Nsoft           Ns   0  # bodies in softening spheres (for S>0)\n"
	" eps             ep   0  fixed / maximum softening length\n"
	" eps_given               individual eps_i are given on input\n"
	" kernel          kr  21  n/(10+n)/(20+n) for Fn/Kn/Pn softening kernel\n"
	" Nref            Nf  16  # bodies in density estimate with scheme=1\n"
	" fac             fc 0.1  factor in time-step criterion  tau < fc / rho\n"
	" hmin            hm   6  tau_min = (1/2)^hm\n"
	" Nsteps          Nt   1  # time steps (1 -> leap-frog)\n"
	" give_pot                give potential with output\n"
	" give_rho         g   0  0/1/2: output density (1) or surface density (2)\n"
	" time             t   0  initial simulation time\n"
	" N                n   0  number of bodies to be read\n"
	" data_format    for   0  0/1 -> body data are in ascii/binary\n"
	" start_data              this terminates the header\n\n"
	" A parameter already  set may be superseeded  by another line.  A line\n"
	" with unknown PARAMETER is notified but otherwise ignored.\n"
	" Body data in ASCII format are N lines of the format\n"
	"         mass x y z vx vy vz [eps] [pot] [rho],\n"
	" where eps must be given if indicated by the option \"eps_given\".\n"
	" Body data in binary format are N times the binary sequence of\n"
	"         mass x y z vx vy vz eps,\n"
	" where eps will be used only if the option \"eps_given\" has been set.\n";
    }
    //--------------------------------------------------------------------------
  public:
    bool write_yanc_ascii (                        // write yanc format, ascii  
			   const char* file,       // I:  name: output file     
			   const char* prog) const // I:  name: calling program 
    {
      ofstream out;
      if(!open(out,file)) {
	yancwarning("cannot output");
	return false;
      }
      if(!okay()) {
	yancwarning("not okay(); cannot output");
	return false;
      }
      write_yanc_header(out,file,SOFTEN,prog,0);
      BODIES->write_ascii(out,SOFTEN? ADD_OUT | epsi : ADD_OUT);
      if(out) return true;
      else    return false;
    }
    //--------------------------------------------------------------------------
  public:
    bool write_yanc_binary(                        // write yanc format, binary 
			   const char* file,       // I:  name: output file     
			   const char* prog) const // I:  name: calling program 
    {
      ofstream out;
      if(!open(out,file)) {
	cerr<<"YANC: write_yanc_binary(): cannot output\n";
	return false;
      }
      if(!okay()) {
	out<<" YANC: not okay(), cannot write_yanc_binary()\n";
	return false;
      }
      write_yanc_header(out,file,SOFTEN,prog,1);
      BODIES->write_binary(out);
      if(out) return true;
      else    return false;
    }
    //------------------------------------------------------------------------
    // informators
    bool          okay ()       const { return BODIES!=0; }
    const indx&   Nsteps()      const { return NSTEPS; }
    const uint&   Nbodies()     const { return BODIES->N_body(); }
    const uint&   Nref()        const { return NREF; }
    const uint&   Nreuse()      const { return NREUSE; }
    const uint&   Ncrit()       const { return NCRIT; }
    const string& data_file()   const { return FILE; }
    const real&   eps()         const { return EPS; }
    const real    tau_min()     const { return pow(half,HMIN); }
    const real    tau_max()     const { return pow(half,HMIN+1-NSTEPS); }
    indx          softening()   const { return SOFTEN; }
    indx          kernel()      const { return KERN; }
    const real&   time  ()      const { return TIME; }
    const real&   initial_time()const { return TINI; }
    const real&   theta()       const { return THETA; }
    const real&   fac()         const { return FAC; }
  };
  }
//=============================================================================+
//                                                                             |
// class nbdy::yanc                                                            |
//                                                                             |
//==============================================================================
inline void check(const ndata* NDATA, const char* m)
{
  if(! NDATA->okay() ) {
    cerr<<"YANC: not okay. cannot "<<m<<"\n";
    exit(1);
  }
}
//------------------------------------------------------------------------------
void yanc::describe(ostream   & out,
		    const char* name,
		    const char* outf)
{
  MYNDATA->describe(out,name,outf);
  out<<"#   total energy:        "<<MYNBODY->total_energy()
     <<
    "\n# ======================================================================="
    "\n#\n";
}
//------------------------------------------------------------------------------
bool yanc::write_yanc_ascii(const char* file, const char* prog)
{
  if(MYNDATA->ADD_OUT & rhos)      MYNBODY->estimate_mass_densities();
  else if(MYNDATA->ADD_OUT & surf) MYNBODY->estimate_surf_densities();
  return MYNDATA->write_yanc_ascii(file,prog);
}
//------------------------------------------------------------------------------
bool yanc::write_yanc_binary(const char* file, const char* prog) const
{
  return MYNDATA->write_yanc_binary(file,prog);
}
//------------------------------------------------------------------------------
void yanc::describe_yanc_format(ostream& o)
{
  return ndata::describe_yanc_format(o);
}
//------------------------------------------------------------------------------
void yanc::full_step() 
{
  check(MYNDATA, "long_step()");
  MYNBODY->full_step(MYNDATA->POR);
  MYNDATA->TIME = MYNBODY->time();
}
//------------------------------------------------------------------------------
void yanc::long_stats  (ostream&o) const
{
  MYNBODY->long_stats(o);
}
//------------------------------------------------------------------------------
void yanc::short_stats (ostream&o) const
{
  o<<" "; MYNBODY->short_stats(o);
}
//------------------------------------------------------------------------------
void yanc::short_head  (ostream&o) const
{
  o<<"#"; MYNBODY->short_stat_head(o);
  o<<"#"; MYNBODY->short_stat_line(o);
}
//------------------------------------------------------------------------------
#define FUNC(TYPE,NAME) TYPE yanc::NAME() const { return TYPE(MYNDATA->NAME()); }
FUNC(bool, okay)
FUNC(int,  Nsteps)
FUNC(int,  Nbodies)
FUNC(int,  Nref)
FUNC(int,  Nreuse)
FUNC(int,  Ncrit)
FUNC(int,  softening)
FUNC(int,  kernel)
FUNC(float,time)
FUNC(float,initial_time)
FUNC(float,theta)
FUNC(float,fac)
FUNC(float,eps)
FUNC(float,tau_min)
FUNC(float,tau_max)
#ifdef ALLOW_NEMO
FUNC(bool,nemo_is_open)
#endif
#undef FUNC
float       yanc::cpu_total() const { return MYNBODY->cpu_total(); }
const char* yanc::data_file() const { return MYNDATA->data_file().c_str(); }
// const body* yanc::bodies   () const { return MYNDATA->bodies(); }
//------------------------------------------------------------------------------
#ifdef ALLOW_NEMO
//------------------------------------------------------------------------------
void yanc::open_nemo (const char *file, const bool resume) {
  MYNDATA->open_nemo(file,resume);
}
//------------------------------------------------------------------------------
void yanc::close_nemo()                 { MYNDATA->close_nemo(); }
//------------------------------------------------------------------------------
void yanc::write_nemo()
{
  if(MYNDATA->ADD_OUT & rhos)      MYNBODY->estimate_mass_densities();
  else if(MYNDATA->ADD_OUT & surf) MYNBODY->estimate_surf_densities();
  MYNDATA->write_nemo(MYNBODY);
}
//------------------------------------------------------------------------------
void yanc::describe_nemo(ostream   & out,
			 const char* com)
{
  if(!okay()) {
    out<<"ERROR something wrong with N-body data, perhaps not read yet?\n";
    exit(1);
  }
  out<<"#"; MYNBODY->short_stat_line(out);
  out<<"# \""<<com<<"\"\n#"<<endl;
}
//------------------------------------------------------------------------------
yanc::yanc(const char     *in,                      // I: input file            
	   const bool      nemo,                    // I: do nemo I/O?          
	   const double    th,                      // I: theta                 
	   const int       nu,                      // I: Nreuse                
	   const int       nc,                      // I: Ncrit                 
	   const double    ep,                      // I: eps                   
	   const int       kn,                      // I: kern                  
	   const int       hm,                      // I: hmin                  
	   const add_out   ao,                      // I: what to give on output
	   const int       ns,                      // I: Nsteps                
	   const int       nr,                      // I: Nref                  
	   const double    fc,                      // I: fac                   
	   const bool      resume,                  //[I: resume old sim (nemo)]
	   const potential*pt) :                    //[I: external potential]   
  MYNDATA(new ndata(in,nemo,th,nu,nc,ep,kn,hm,ao,ns,nr,fc,resume)),
  MYNBODY(MYNDATA->NSTEPS > 1 ?
	  new BlockStepCode(MYNDATA->BODIES,
			    MYNDATA->NSOFT,
			    MYNDATA->EPS,
			    MYNDATA->TIME,
			    MYNDATA->HMIN+1-MYNDATA->NSTEPS,
			    MYNDATA->NSTEPS,
			    MYNDATA->NREF,
			    abs(MYNDATA->FAC),
			    MYNDATA->POR,
			    4.0,
			    soften(MYNDATA->SOFTEN),
			    MYNDATA->THETA,
			    MYNDATA->NCRIT,
			    kern(MYNDATA->KERN),
			    pt) :
	  new LeapFrogCode (MYNDATA->BODIES,
			    MYNDATA->NSOFT,
			    MYNDATA->EPS,
			    MYNDATA->TIME,
			    MYNDATA->HMIN,
			    MYNDATA->NREUSE,
			    MYNDATA->NREF,
			    MYNDATA->POR,
			    soften(MYNDATA->SOFTEN),
			    MYNDATA->THETA,
			    MYNDATA->NCRIT,
			    kern(MYNDATA->KERN),
			    pt) )
{
  if(resume) MYNBODY->reset_cpu_total(MYNDATA->CPU_I);
}
#endif
//------------------------------------------------------------------------------
yanc::yanc(const char  *in) :                       //  I:  input file          
  MYNDATA(new ndata(in)),
  MYNBODY( MYNDATA->NSTEPS > 1 ?
	   new BlockStepCode(MYNDATA->BODIES,
			     MYNDATA->NSOFT,
			     MYNDATA->EPS,
			     MYNDATA->TIME,
			     MYNDATA->HMIN+1-MYNDATA->NSTEPS,
			     MYNDATA->NSTEPS,
			     MYNDATA->NREF,
			     abs(MYNDATA->FAC),
			     MYNDATA->POR,
			     4.0,
			     soften(MYNDATA->SOFTEN),
			     MYNDATA->THETA,
			     MYNDATA->NCRIT,
			     kern(MYNDATA->KERN)) :
	   new LeapFrogCode (MYNDATA->BODIES,
			     MYNDATA->NSOFT,
			     MYNDATA->EPS,
			     MYNDATA->TIME,
			     MYNDATA->HMIN,
			     MYNDATA->NREUSE,
			     MYNDATA->NREF,
			     MYNDATA->POR,
			     soften(MYNDATA->SOFTEN),
			     MYNDATA->THETA,
			     MYNDATA->NCRIT,
			     kern(MYNDATA->KERN)) ) {}
//------------------------------------------------------------------------------
yanc::~yanc()
{
  delete MYNBODY;
  delete MYNDATA;
}
//------------------------------------------------------------------------------
