// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// YancNemo.cc                                                                 |
//                                                                             |
// copyright Walter Dehnen, 2001                                               |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie, Knigstuhl 69117 Heidelberg   |
//           Germany                                                           |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//           YancNemo -- Yet Another N-body Code, for                          |
//                       Not Everybody Must Observe                            |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// history:                                                                    |
//                                                                             |
// v 1.0.0   28/05/2001  WD created with the help of NEMO caption PJT          |
// v 1.0.1   01/06/2001  WD bug in this file removed, added history output     |
// v 1.0.2   13/06/2001  WD Ncrit option added (before Ncrit==1)               |
// v 1.0.3   20/06/2001  WD tree::make() back to adding-leafs algorithm        |
//                          Coordinatesystem given with snapshot               |
// v 1.0.4   25/06/2001  WD bug (builder::link_tree()) removed (thanks to PJT) |
// v 1.0.5   28/06/2001  WD changes in tree::grow(). Avoid box overflow.       |
//                          bug (in builder) removed                           |
// v 1.0.6   01/08/2001  WD substantial code upgrade in the tree (falcON)      |
// v 1.0.7   02/08/2001  WD Nreuse option added (before Nreuse==0)             |
// v 1.0.8   08/08/2001  WD new data lay-out for bodies, enabling special      |
//                          treatment for leap-frog, saves memory & time (?)   |
// v 1.0.9   17/09/2001  WD added support for dynamically linked external      |
//                          potential (potential.h)                            |
// v 1.0.10  20/09.2001  WD added option ("resume") for resuming an old or     |
//                          interrupted simulation. Appends to input file      |
// v 1.0.11  21/09/2001  WD improved handling of nemo I/O; can now also read   |
//                          PosTag & VelTag instead of PhaseSpaceTag           |
// v 1.0.12  21/09/2001  WD added parameter give_acc                           |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef ALLOW_NEMO
#error You need "ALLOW_NEMO" to compile YancNemo
#endif

#include <yanc.h>                              // the N-body code               
#include <nbdy/pext.h>                         // external potential            
using nbdy::yanc;
using nbdy::add_out;
using nbdy::nemo_pot;

#include <iostream>                            // C++ I/O                       
using std::cout;

extern "C" {
#include <nemo.h>                              // NEMO basics                   
#include <potential.h>                         // NEMO potentials               
}
//------------------------------------------------------------------------------
string defv[] = {
  "in=???\n            input file                                    ",
  "out=\n              output file   must be given unless resume=t   ",
  "tstop=10\n          integration time                              ",
  "step=1\n            time between outputs                          ",
  "theta=0.6\n         tolerance parameter at M=M_tot                ",
  "Nreuse=0\n          max # subsequent tree re-uses                 ",
  "Ncrit=6\n           max # bodies in un-split cells                ",
  "eps=0.05\n          softening length                              ",
  "kernel=21\n         softening kernel                              ",
  "hmin=6\n            tau_min = (1/2)^hmin                          ",
  "Nlev=1\n            # time-step levels                            ",
  "Nref=16\n           # bodies in cell for density estimate         ",
  "fac=\n              factor in tau < fac / rho                     ",
  "resume=f\n          resume old simulation?  that implies:\n"
  "                   - read last snapshot from input file\n"
  "                   - append to input file (unless out given)     ",
  "give_pot=t\n        t/f:  output potentials                       ",
  "give_acc=f\n        t/f:  output accelerations                    ",
  "give_rho=0\n        0/1/2 output none/density/surface density     ",
  "potname=\n          name of external potential                    ",
  "potpars=\n          parameters of external potential              ",
  "potfile=\n          file required by external potential           ",
  "VERSION=1.0.12\n    21/September/2001 WD                          ",
  NULL};
string usage = "YancNemo -- Yet Another N-body Code, for\n"
               "            Not Everybody Must Observe    ";
//------------------------------------------------------------------------------
#ifdef DOUBLEPRECISION
# define get_potential_real get_potential_double
#else
# define get_potential_real get_potential_float
#endif // DOUBLEPRECISION
//------------------------------------------------------------------------------
void nemo_main()
{
  register double dTou,Tout,Tend;
  Tend   = getdparam("tstop");
  dTou   = getdparam("step");
  register int Nlev = getiparam("Nlev");
  if(Nlev>1 && !hasvalue("fac")) error("fac required if Nlev > 1");
  nemo_pot *pot = hasvalue("potname")?
    new nemo_pot(get_potential_real(getparam("potname"),
				    getparam("potpars"),
				    getparam("potfile"))) : 0;
  bool resume = getbparam("resume");
  if(!resume && !hasvalue("out")) error("out required if resume=f");
  add_out ao = nbdy::none;
  if     (getbparam("give_pot"))    ao |= nbdy::pots;
  if     (getbparam("give_acc"))    ao |= nbdy::accs;
  if     (getiparam("give_rho")==1) ao |= nbdy::rhos;
  else if(getiparam("give_rho")==2) ao |= nbdy::surf;
  yanc YANC(getparam ("in"),
	    true,
	    getdparam("theta"),
	    getiparam("Nreuse"),
	    getiparam("Ncrit"),
	    getdparam("eps"),
	    getiparam("kernel"),
	    getiparam("hmin"),
	    ao,
	    Nlev,
	    getiparam("Nref"),
	    hasvalue ("fac")? getdparam("fac") : nbdy::zero,
	    resume,
	    pot);
  if(!YANC.okay()) error("initialization error\n");
  if(Tend <  YANC.initial_time()) {
    warning("tstop < t_ini: nothing to be done\n");
    return;
  }
  Tout = YANC.initial_time() + 0.999999*dTou;
  YANC.describe_nemo(cout,*(ask_history()));
  YANC.short_head   (cout);
  YANC.short_stats  (cout);
  if(hasvalue("out")) {
    YANC.open_nemo    (getparam ("out"),resume);
    YANC.write_nemo   ();
  } else
    YANC.open_nemo    (0,resume);
  if(!YANC.nemo_is_open()) error("opening for output failed");
  while(YANC.time() < Tend) {
    YANC.full_step();
    YANC.short_stats(cout);
    if(YANC.time() >= Tout) {
      YANC.write_nemo();
      Tout += dTou;
    }
  }
  YANC.close_nemo();
  if(pot) delete pot;
}
//---------------------end-of-YancNemo.cc------that's-it-!-----------------------
