// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// mkking.cc                                                                   |
//                                                                             |
// copyright Walter Dehnen, 2001                                               |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie, Knigstuhl 69117 Heidelberg   |
//           Germany                                                           |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// creates N-body initial conditions from a King model                         |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// history:                                                                    |
//                                                                             |
// v 1.0    11/09/2001  WD  created for nemo 3.0.4                             |
// v 1.0.1  17/09/2001  WD  changed to init_xrandom (nemo 3.0.8)               |
// v 1.0.2  19/09/2001  WD  bug fixed; centrate center of mass (option zerocm) |
// v 1.0.3  21/09/2001  WD  changes in body & nmio                             |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef ALLOW_NEMO
#error You need "ALLOW_NEMO" to compile mkking
#endif

#include <util/Pi.h>
#include <king.h>
#include <body.h>
#include <nbdy/nmio.h>

extern "C" {
#include <nemo.h>
  double xrandom(double,double);
  int    init_xrandom(string);
  double sqrt(double);
  double sin(double);
  double cos(double);
}
////////////////////////////////////////////////////////////////////////////////
string defv[] = {
  "out=???\n                output file                             ",
  "nbody=???\n              number of bodies                        ",
  "W0=???\n                 W0 = Psi0/sigma                         ",
  "seed=0\n                 seed for the randum number generator    ",
  "time=0\n                 simulation time of snapshot             ",
  "mass=1\n                 total mass                              ",
  "core_radius=1\n          core radius                             ",
  "tidal_radius=\n          tidal radius (don't use core radius)    ",
  "zerocm=t\n               centrate snapshot?                      ",
  "VERSION=1.0.2\n          19/September/2001 WD                    ",
  NULL};
string usage = "mkking -- construct a King model";
////////////////////////////////////////////////////////////////////////////////
void nemo_main()
{
  //----------------------------------------------------------------------------
  // 1. create king model, i.e. integrate to get Phi(r) etc.                    
  //----------------------------------------------------------------------------
  nbdy::king_model KM(getdparam("W0"));
  //----------------------------------------------------------------------------
  // 2. rescale model to mass and core/tidal radius wanted                      
  //----------------------------------------------------------------------------
  if(hasvalue("tidal_radius"))
    KM.reset_scales_tidal(getdparam("mass"), getdparam("tidal_radius"));
  else
    KM.reset_scales_core (getdparam("mass"), getdparam("core_radius"));
  //----------------------------------------------------------------------------
  // 3. create initial conditions from King model                               
  //----------------------------------------------------------------------------
  int N = getiparam("nbody");
  nbdy::bodies BB(N);
  init_xrandom(getparam("seed"));
  const    double     m = 1./double(N);
  register double     rad,vel,cth,R,phi;
  for(register nbdy::body Bi = BB.begin_bodies(); Bi!=BB.end_bodies(); ++Bi) {
    //                                                                          
    // 3.1 set mass and get r and v                                             
    //                                                                          
    Bi.mass() = m;
    KM.random(xrandom(0.,1.), xrandom(0.,1.), rad, vel);
    //                                                                          
    // 3.2 get x,y,z                                                            
    //                                                                          
    cth       = xrandom(-1.,1.);
    R         = rad * sqrt(1.-cth*cth);
    phi       = xrandom(0.,util::TPi);
    Bi.pos(0) = R * cos(phi);
    Bi.pos(1) = R * sin(phi);
    Bi.pos(2) = rad * cth;
    //                                                                          
    // 3.3 get vx,vy,vz                                                         
    //                                                                          
    cth       = xrandom(-1.,1.);
    R         = vel * sqrt(1.-cth*cth);
    phi       = xrandom(0.,util::TPi);
    Bi.vel(0) = R * cos(phi);
    Bi.vel(1) = R * sin(phi);
    Bi.vel(2) = vel * cth;
  }
  //----------------------------------------------------------------------------
  // 4. centrate position and velocity on origin                                
  //----------------------------------------------------------------------------
  if(getbparam("zerocm")) {
    register nbdy::vect cpos(nbdy::zero), cvel(nbdy::zero);
    for(register nbdy::body Bi = BB.begin_bodies(); Bi!=BB.end_bodies(); ++Bi) {
      cpos += nbdy::pos(Bi);
      cvel += nbdy::vel(Bi);
    }
    cpos *= real(m);
    cvel *= real(m);
    for(register nbdy::body Bi = BB.begin_bodies(); Bi!=BB.end_bodies(); ++Bi) {
      Bi.pos() -= cpos;
      Bi.vel() -= cvel;
    }
  }
  //----------------------------------------------------------------------------
  // 5. output of snapshot                                                      
  //----------------------------------------------------------------------------
  nbdy::nemo_output out(getparam("out"));
  nbdy::real time = getdparam("time");
  BB.write_nemo_snapshot(out,nbdy::none,&time);
};
